options
{
	language="Cpp";
}

class CMMParser extends Parser;

options
{
	k = 2;
	importVocab=CMM;
  // analyzerDebug = true;
}

/* --------------------------------------------------------------------- */	

identifier:
  IDENT ;

qualified_name:
  (SCOPE)?
  identifier
  (
	  options { warnWhenFollowAmbig = false; }: // kvuli pocatecnimu SCOPE
    SCOPE
    (BNOT)?
    identifier
  )* ;

complete_class_name:
  qualified_name ;

type_name:
  qualified_name ;

/* --------------------------------------------------------------------- */	

// template_class_name: "template_class_name";
// template_declaration: "template_declaration";
// try_stat: "try_stat";

/* --------------------------------------------------------------------- */	

expr :
	assignment_expr
	( COMMA assignment_expr	)* ;

assignment_expr :
 	conditional_expr // !?
	(	(	ASSIGN
    | PLUS_ASSIGN
    | MINUS_ASSIGN
    | STAR_ASSIGN
    | DIV_ASSIGN
    | MOD_ASSIGN
    | SR_ASSIGN
    | SL_ASSIGN
    | BAND_ASSIGN
    | BXOR_ASSIGN
    | BOR_ASSIGN )
		assignment_expr
	)?
	;

conditional_expr :
  logical_or_expr
  ( QUESTION expr COLON conditional_expr )? ;

logical_or_expr :
	logical_and_expr (LOR logical_and_expr)* ;

logical_and_expr :
  inclusive_or_expr (LAND inclusive_or_expr)* ;

inclusive_or_expr :
	exclusive_or_expr (BOR exclusive_or_expr)* ;

exclusive_or_expr :
	and_expr (BXOR and_expr)* ;

and_expr :
	equality_expr (BAND equality_expr)* ;

equality_expr :
	relational_expr
	((NOT_EQUAL | EQUAL) relational_expr)* ;

relational_expr :
	shift_expr
	(
    // options { warnWhenFollowAmbig = false; }:
	  ( LT_ | GT | LE |	GE )
	  shift_expr
	)* ;

shift_expr :
	additive_expr
	((SL | SR) additive_expr)* ;

additive_expr :
	multiplicative_expr
	((PLUS | MINUS) multiplicative_expr)* ;

multiplicative_expr :
	unary_expr
	((STAR | DIV | MOD ) unary_expr)* ;

pm_expr:
  cast_expr
  ((DOT_STAR | ARROW_STAR) cast_expr)* ;

cast_expr:
  unary_expr ;

unary_expr :
  	postfix_expr
 	| INC unary_expr
 	|	DEC unary_expr
 	|	(STAR | BAND | PLUS| MINUS | BNOT | LNOT) cast_expr
  | sizeof_expr
  | allocation_expr
  | deallocation_expr ;

sizeof_expr:
  "sizeof" unary_expr;

allocation_expr :
  // (SCOPE)?
  "new"
  (type_name | LPAREN type_description RPAREN)
  (new_initializer)? ;

new_initializer :
  LPAREN (initializer_list)? RPAREN ;

deallocation_expr :
  // (SCOPE)?
  "delete" (LBRACK RBRACK)? cast_expr ;

cast_spec:
  LT_ type_description GT LPAREN expr RPAREN ;

postfix_expr :
  (	primary_expr
  | "dynamic_cast" cast_spec
  | "static_cast" cast_spec
  | "const_cast" cast_spec
  | "reinterpreter_cast" cast_spec
  | "typeid" LPAREN expr RPAREN )

 	(	LBRACK expr RBRACK
  |	LPAREN expr_list RPAREN
 	| DOT identifier
 	|	ARROW identifier
 	| INC
 	| DEC
 	)* ;

expr_list:
	assignment_expr (COMMA assignment_expr)* ;

primary_expr :	
    literal
  | "this"
	// | LPAREN expr RPAREN
  | qualified_name ;

literal:
    NUM_INT
  | NUM_FLOAT
  | CHAR_LITERAL
  | STRING_LITERAL;

/* ---------------------------------------------------------------------- */

declaration_list:
   declaration (declaration)* ;

declaration :
    common_declaration
  | enum_declaration
  | class_declaration
  | typedef_declaration
  | asm_declaration
  | template_declaration
  | linkage_specification
  | SEMI ;

decl_specifiers:
  (decl_start)*
  type_description ;

type_description:
  (decl_simple (decl_simple)* | decl_stop )
  (ptr_specifier)*
  ;

decl_start:
    storage_class_specifier
  | fct_specifier
  | cv_specifier ;

storage_class_specifier:
    "auto"
  | "register"
  | "static"
  | "extern" ;

fct_specifier:
    "inline"
  | "virtual"
  | "explicit" ;

cv_specifier:
    "const"
  | "volatile" ;

decl_simple:
    "char"
  | "short"
  | "int"
  | "long"
  | "signed"
  | "unsigned"
  | "float"
  | "double"
  | "void" ;

decl_stop:
  qualified_name (template_arg_list)? ;

/* enum */

enum_declaration:
  "enum" identifier LCURLY (enum_list)? RCURLY SEMI ;

enum_list:
  enumerator ( COMMA enumerator )? ;

enumerator:
  identifier ( ASSIGN const_expr )? ;

const_expr:
  conditional_expr ;

/* typedef */

typedef_declaration:
  "typedef" decl_specifiers declarator;

/* extern "..." */

linkage_specification:
  "extern"
  STRING_LITERAL
  (
    LCURLY (declaration_list)? RCURLY
  |
    declaration
  );

/* asm */

asm_declaration:
  "asm" LPAREN STRING_LITERAL RPAREN ;

/* ---------------------------------------------------------------------- */

common_declaration :
  decl_specifiers qualified_name (common_tail | function_tail);

common_tail:
  (initializer)? (COMMA qualified_name (initializer)? )* SEMI;

function_tail :
  // dname
  LPAREN argument_declaration_list RPAREN
  (
    (ctor_initializer)? fct_body
  |
    SEMI
  );


/* ---------------------------------------------------------------------- */

// declarator_list:
//   init_declarator (COMMA init_declarator)* ;

// init_declarator:
//   declarator (initializer)? ;

declarator:
  (ptr_specifier)*

  (
    dname
  |
    LPAREN declarator RPAREN
  )

  (
	  // options { warnWhenFollowAmbig = false; }: // !?
    LPAREN argument_declaration_list RPAREN
    // (cv_specifier_list)? // !?
  |
    LBRACK (expr)? RBRACK
  )*
  ;

ptr_specifier:
  ( STAR
  | BAND
  // | (complete_class_name SCOPE STAR) => complete_class_name SCOPE STAR // !?
  )
  // (cv_specifier_list)? // !?
  ;

cv_specifier_list:
  cv_specifier (cv_specifier)* ;

dname:
    BNOT identifier
  | qualified_name ;

argument_declaration_list:
  argument_declaration (COMMA argument_declaration)* ;

argument_declaration:
  type_description
  (identifier)?
  (ASSIGN assignment_expr)?
  ;

fct_body:
  compound_stat ;

/* initializer */

initializer:
  ASSIGN initializer_item ;

initializer_item:
    assignment_expr
  |
    LCURLY
    initializer_list
    // (COMMA)? // !?
    RCURLY ;

initializer_list:
    initializer_item
    (COMMA initializer_item)*
    ;

/* ---------------------------------------------------------------------- */

class_declaration :
  class_key
  identifier
  (
    (base_spec)?
    LCURLY member_list RCURLY
  )?
  SEMI ;

class_key:
     "class"
   | "struct"
   | "union" ;

member_list:
  (access_specifier COLON | member_declaration )* ;

member_declaration:
    (qualified_name SEMI) => qualified_name SEMI
  | declaration ;

base_spec:
  COLON base_list;

base_list:
  base_specifier ( COMMA base_specifier )* ;

base_specifier:
  (
  | "virtual" (access_specifier)?
  | access_specifier ("virtual")? )
  complete_class_name ;

access_specifier:
    "private"
  | "protected"
  | "public" ;

ctor_initializer:
  COLON mem_initializer_list ;

mem_initializer_list:
  mem_initializer (COMMA mem_initializer)* ;

mem_initializer:
   complete_class_name LPAREN (expr_list)? RPAREN ;

/* --------------------------------------------------------------------- */	

stat :
  // (declaration) => declaration_stat |
  labeled_stat |
  case_stat |
  default_stat |
  expression_stat |
  empty_stat |
  compound_stat |
  if_stat |
  switch_stat |
  while_stat |
  do_stat |
  for_stat |
  break_stat |
  continue_stat |
  return_stat |
  goto_stat |
  try_stat ;

labeled_stat:
  identifier COLON stat;

case_stat:
  "case" expr COLON stat;

default_stat:
  "default" COLON stat;

expression_stat:
   expr SEMI;

empty_stat:
	SEMI;

compound_stat :	
  LCURLY (stat)* RCURLY;

if_stat:
	"if" LPAREN expr RPAREN stat
	 (
 		 options { warnWhenFollowAmbig = false; } :
  	"else" stat
	 )? ;

switch_stat:
	"switch" LPAREN expr RPAREN stat;

while_stat:
	"while" LPAREN expr RPAREN stat;

do_stat:
	"do" stat "while" LPAREN expr RPAREN SEMI;

for_stat:
	"for"
	LPAREN
	for_init SEMI
	(expr)?	SEMI
	(expr_list)?
	RPAREN
	stat;

for_init :
  (	
    (declaration) => declaration
	|	
     expr
	) ;

break_stat:
	"break" SEMI;

continue_stat:
	"continue" SEMI;

return_stat:
	"return" (expr)? SEMI;

goto_stat:
	"goto" identifier SEMI;

declaration_stat:
	 declaration SEMI;

/* ---------------------------------------------------------------------- */

template_declaration:
  "template" LT_ template_argument_list GT declaration ;

template_argument_list:
  template_argument (COMMA template_argument)* ;

template_argument:
    type_argument
  | value_argument  ;

type_argument:
  "class" identifier ;

value_argument:
  type_description
  identifier
  (ASSIGN shift_expr)? ; // !?

template_arg_list:
   LT_ template_arg (COMMA template_arg)* GT;

template_arg:
   shift_expr; // !?
	
/* ---------------------------------------------------------------------- */

try_stat:
   "try" compound_stat handler_list;

handler_list:
	 handler (handler)* ;

handler:	
  "catch" LPAREN exception_declaration RPAREN compound_stat ;

exception_declaration:
    type_description (identifier)?
  | DOTS ;

throw_expr: /* !? */
	"throw" (expr)?;

exception_specification:  /* !? */
  "throw" LPAREN (type_list)? RPAREN ;

type_list:
   type_description (COMMA type_description)* ;
