
/* cmm.g */

options
{
   product_name = cmm_product; // output file name prefix (CMM.hpp, CMM.cpp)

   common_name  = cmm_common; // output file name prefix
   common_header  = cmm_common; // common header (enumerated types, forward class declarations)
   // public_header  = cmm_data; // public header (data structures)
   // private_header = cmm_private; // private header

   lexer_input = "lex.g"; // input file with lexer
   lexer_name = cmm_lexer; // output file name prefix

   parser_name = cmm_parser; // output file name prefix

   // product_input = "ext.h"; // input file with additional product declarations

   lexer_class = CmmLexer;
   parser_class = CmmParser;

   // product_namespace = finch;
   product_class = CmmProduct;
   product_super_class = Writer;

   antlr_mode = true; // use ANTLR C++ libraries

   // switchable_literals = true;
   // tree_code = true;

   // notes = true; // duplicate comments and spaces
   // only_start_notes = true;

   /*
   rule_prefix = rule_ ;
   type_prefix = type_ ;

   field_prefix = field_ ;
   list_prefix = queue_ ;
   internal_prefix = internal_ ;

   send_func_prefix = send_fce_ ;
   send_internal_prefix = original_ ;
   open_func_prefix = open_fce_ ;
   close_func_prefix = close_fce_ ;

   param_prefix = parameter_ ;
   temp_prefix = temporary_ ;
   local_prefix = local_ ;
   variable_prefix = variable_ ;

   conv_func_prefix = convert_ ;
   util_func_prefix = utility_ ;
   */

   basic_type = CmmBasic;
   expr_type = CmmExpr;

   binary_expr_type     = CmmBinaryExpr;
   binary_expr_left     = left;
   binary_expr_right    = right;
   binary_expr_selector = kind;

   unary_expr_type     = CmmUnaryExpr;
   unary_expr_param    = param;
   unary_expr_selector = kind;
}

/* --------------------------------------------------------------------- */

extension product_include
{@@
#include "writer.h"
@@}

/* --------------------------------------------------------------------- */

grammar_options
{
  language=Cpp;
  namespaceAntlr = "antlr::";
  namespaceStd = "std::";
  genHashLines = false;
}

lexer_options
{
  // testLiterals = false;
  k = 3;
  caseSensitive = true;
  caseSensitiveLiterals = true;
}

parser_options
{
  k = 2;
  // defaultErrorHandler = false;
  importVocab = cmm_lexer; // important
}

/* --------------------------------------------------------------------- */

// CmmAccess type shared by CmmMemberSect and CmmBaseItem

enum CmmAccess
{
  PrivateAccess,
  ProtectedAccess,
  PublicAccess
};

struct CmmMemberVisibility
{
  CmmAccess access;
};

struct CmmBaseItem
{
  CmmAccess access;
};

/* --------------------------------------------------------------------- */

parser ;

/* --------------------------------------------------------------------- */

identifier <primitive> <identifier>:
  IDENT ;

cont_name <CmmContName>:
  <no_space> "::" <no_space>
  (
     '~' <no_space> <set destructor_name = true>
  |
     "template" <set template_name = true>
  )?
  id:identifier ;

id_expr <CmmNameExpr:CmmExpr>:
   // ( "::" <set global=true> )? // !? @@
   ( id: identifier | special:special_function )
   ( <add> cont_name )*
   // ( template_params:template_arg_list )? // !? @@
   ;

/* --------------------------------------------------------------------- */

/* primary expression */

primary_expr <select CmmExpr>:
    numeric_literal_expr
  | char_literal_expr
  | string_literal_expr
  | this_expr
  | id_expr
  | subexpr_expr
  ;

numeric_literal_expr <CmmNumValueExpr:CmmExpr>:
  <set kind = NumValueExp>
  value:numeric_literal;

char_literal_expr <CmmCharValueExpr:CmmExpr>:
  <set kind = CharValueExp>
  value:char_literal;

string_literal_expr <CmmStringValueExpr:CmmExpr>:
  <set kind = StringValueExp>
  value:string_literal;

numeric_literal <primitive>:
  NUMBER ;

char_literal <primitive>:
  CHAR_LITERAL ;

string_literal <primitive>:
  STRING_LITERAL ;

this_expr <CmmThisExpr:CmmExpr>:
  "this";

subexpr_expr <CmmSubexprExpr:CmmExpr>:
  '(' param:expr ')';

/* postfix expression */

postfix_start <select CmmExpr>:
    modern_cast_expr
  | typeid_expr
  | primary_expr ; // last - important !?

modern_cast_expr <CmmModernCastExpr:CmmExpr>:
  ( "dynamic_cast"       <set kind = DynamicCast>
  | "static_cast"        <set kind = StaticCast>
  | "const_cast"         <set kind = ConstCast>
  | "reinterpret_cast"   <set kind = ReinterpreterCast> )
  '<' <no_space> type:type_id <no_space> '>'
  '(' param:expr ')'
  ;

typeid_expr <CmmTypeidExpr:CmmExpr>:
  "typeid"
  '('
  (
     (expr) => param1:expr
  |
     param2:type_id
  )
  ')'
  ;

postfix_expr <choose CmmExpr>:
  postfix_start
  (
    index_expr |
    call_expr |
    field_expr |
    ptr_field_expr |
    post_inc_expr |
    post_dec_expr
  )* ;

index_expr <CmmIndexExpr:CmmExpr>:
  <store left:CmmExpr>
  '[' right:expr ']' ;

call_expr <CmmCallExpr:CmmExpr>:
  <store param:CmmExpr>
  '(' list:expr_list ')' ;

field_expr <CmmFieldExpr:CmmExpr>:
  <store param:CmmExpr>
  <no_space> '.' <no_space>
  ( '~' <no_space> <set destructor_name = true> |
    "template" <set template_name = true> )?
  name:id_expr ;

ptr_field_expr <CmmPtrFieldExpr:CmmExpr>:
  <store param:CmmExpr>
  <no_space> "->" <no_space>
  ( '~' <no_space> <set destructor_name = true> |
    "template" <set template_name = true> )?
  name:id_expr ;

post_inc_expr <CmmPostIncExpr:CmmExpr>:
  <store param:CmmExpr>
  "++" ;

post_dec_expr <CmmPostDecExpr:CmmExpr>:
  <store param:CmmExpr>
  "--" ;

expr_list <CmmExprList>:
  ( <add> assignment_expr (',' <add> assignment_expr)* )? ;

/* unary expression */

unary_expr <select CmmExpr>:
    inc_expr
  | dec_expr
  | deref_expr
  | addr_expr
  | plus_expr
  | minus_expr
  | bit_not_expr
  | log_not_expr
  | sizeof_expr
  | allocation_expr
  | deallocation_expr
  | postfix_expr ; // last - important !?

inc_expr <CmmIncExpr:CmmExpr>:
  "++" param:cast_expr;

dec_expr <CmmDecExpr:CmmExpr>:
  "--" param:cast_expr;

deref_expr <CmmDerefExpr:CmmExpr>:
  '*' param:cast_expr;

addr_expr <CmmAddrExpr:CmmExpr>:
  '&' param:cast_expr;

plus_expr <CmmPlusExpr:CmmExpr>:
  '+' param:cast_expr;

minus_expr  <CmmMinusExpr:CmmExpr>:
  '-' param:cast_expr;

bit_not_expr <CmmBitNotExpr:CmmExpr>:
  '~' param:cast_expr;

log_not_expr <CmmLogNotExpr:CmmExpr>:
  '!' param:cast_expr;

sizeof_expr <CmmSizeofExp:CmmExpr>:
  "sizeof"
  (
     (unary_expr) => param1:unary_expr
  |
     '(' param2:type_id ')'
  );

allocation_expr <CmmNewExpr:CmmExpr>:
  // ( "::" <set a_global = true> )? // !? @@
  "new"
  (
     spec:alloc_type_specifier_list
     // ptr:ptr_specifier_list // !? @@
     ( <add> allocation_array_limit )*
  |
    '(' type2:type_id ')'
  )
  ( '(' init:expr_list ')' )?
  ;

allocation_array_limit <CmmNewArrayLimit>:
  '[' value:expr ']'  ;

deallocation_expr <CmmDeleteExpr:CmmExpr>:
  // ( "::" <set a_global = true> )? // !? @@
  "delete"
  ( '[' ']' <set a_array = true> )?
  param:cast_expr ;

/* cast expression */

// cast_expr <return CmmExpr>:
//    unary_expr ;

cast_expr <select CmmExpr>:
    ( cast_formula ) => cast_formula // !?
  | unary_expr ;

cast_formula <CmmCastFormula: CmmExpr>:
  '(' type:type_id ')' param:cast_expr ;

/* pm expression */

pm_expr <binary>:
  cast_expr
  ".*"  DotMemberExp ,
  "->*" ArrowMemverExp ;

/* arithmetic and logical expressions */

multiplicative_expr <binary>:
  pm_expr
  '*' MulExp ,
  '/' DivExp ,
  '%' ModExp ;

additive_expr <binary>:
  multiplicative_expr
  '+' AddExp ,
  '-' SubExp ;

shift_expr <binary>:
  additive_expr
  "<<" ShlExp ,
  ">>" ShrExp ;

relational_expr <binary>:
  shift_expr
  '<' LtExp  ,
  '>' GtExp  ,
  "<=" LeExp ,
  ">=" GeExp ;

equality_expr <binary>:
  relational_expr
  "==" EqExp ,
  "!=" NeExp ;

and_expr <binary>:
  equality_expr
  '&' BitAndExp ;

exclusive_or_expr <binary>:
  and_expr
  '^' BitXorExp ;

inclusive_or_expr <binary>:
  exclusive_or_expr
  '|' BitOrExp ;

logical_and_expr <binary>:
  inclusive_or_expr
  "&&" LogAndExp ;

logical_or_expr <binary>:
  logical_and_expr
  "||" LogOrExp ;

/* conditional expression */

conditional_expr <choose CmmExpr>:
  logical_or_expr
  ( conditional_cont_expr )? ;

conditional_cont_expr <new CmmCondExpr:CmmExpr>:
  <store cond_expr:CmmExpr>
  '?' then_expr:expr
  ':' else_expr:conditional_expr ; // !?

/* assignment expression */

assignment_expr <select CmmExpr>:
   throw_expr |
   simple_assignment_expr; // last - important

simple_assignment_expr <binary>:
  conditional_expr // !?
  '='   AssignExp ,
  "+="  AddAssignExp ,
  "-="  SubAssignExp ,
  "*="  MulAssignExp ,
  "/="  DivAssignExp ,
  "%="  ModAssignExp ,
  "<<=" ShlAssignExp ,
  ">>=" ShrAssignExp ,
  "&="  AndAssignExp ,
  "^="  XorAssignExp ,
  "|="  OrAssignExp ;

/* expression */

expr <binary>:
  assignment_expr
  ',' CommaExp ;

/* constant expression */

const_expr <return CmmExpr>:
  conditional_expr ;

/* --------------------------------------------------------------------- */

stat_list <CmmStatSect>:
   ( <new_line> <add> stat )* ;

stat <select CmmStat>:

     ( expression_stat ) => expression_stat |
     ( declaration_stat ) => declaration_stat |

     compound_stat |
     labeled_stat |
     case_stat |
     default_stat |
     empty_stat |
     if_stat |
     switch_stat |
     while_stat |
     do_stat |
     for_stat |
     break_stat |
     continue_stat |
     return_stat |
     goto_stat |
     try_stat ;

/* inner_stat and tail_stat - statements with different output indentation */

inner_stat <return CmmStat>:
   stat;

tail_stat <return CmmStat>:
   stat;

inner_stat_list <return CmmStatSect>:
   stat_list;

/* --------------------------------------------------------------------- */

declaration_stat <CmmDeclStat:CmmStat>:
  decl:block_declaration ;

labeled_stat <CmmLabeledStat:CmmStat>:
  lab:identifier ':' body:stat ;

case_stat <CmmCaseStat:CmmStat>:
  "case" case_expr:expr ':' body:inner_stat;

default_stat <CmmDefaultStat:CmmStat>:
  "default" ':' body:inner_stat;

expression_stat <CmmExprStat:CmmStat>:
  inner_expr:expr ';';

empty_stat <CmmEmptyStat:CmmStat>:
  ';' ;

compound_stat <CmmCompoundStat:CmmStat>:
  '{'
  <no_empty_line>
  body:inner_stat_list
  <no_empty_line>
  '}';

condition <CmmCondition>:
  (simple_declaration) => cond_decl:simple_declaration | // !? @@
  cond_expr:expr ;

if_stat <CmmIfStat:CmmStat>:
  "if" '(' cond:condition ')' then_stat:inner_stat
  (
    <silent>
    <new_line>
    "else" else_stat:tail_stat
  )? ;

switch_stat <CmmSwitchStat:CmmStat>:
  "switch" '(' cond:condition ')' body:inner_stat;

while_stat  <CmmWhileStat:CmmStat>:
  "while" '(' cond:condition ')' body:inner_stat;

do_stat <CmmDoStat:CmmStat>:
  "do" body:inner_stat "while" '(' cond_expr:expr ')' ';';

for_stat <CmmForStat:CmmStat>:
  "for"
  '('
  (
     (simple_declaration) => from_decl:simple_declaration
  |
     from_expr:expr ';'
  |
     ';' <set empty_param = true>
  )
  (to_expr:expr)?
  ';'
  (cond_expr:expr)?
  ')'
  body:inner_stat;

break_stat <CmmBreakStat:CmmStat>:
  "break" ';';

continue_stat <CmmContinueStat:CmmStat>:
  "continue" ';';

return_stat <CmmReturnStat:CmmStat>:
  "return" (return_expr:expr)? ';';

goto_stat <CmmGotoStat:CmmStat>:
  "goto" goto_lab:identifier ';';

/* ---------------------------------------------------------------------- */

declaration_list <CmmDeclSect>:
  ( <add> declaration <new_line> )* ;

declaration <select CmmDecl>:

    ( constructor_head ) => constructor_declaration
  | ( function_head ) =>  function_declaration
  | simple_declaration
  | empty_declaration

  | namespace_definition
  | using_declaration
  | asm_declaration

  | template_declaration
  | linkage_specification ;

block_declaration <select CmmDecl>:

    simple_declaration
  | empty_declaration
  | namespace_definition
  | using_declaration
  | asm_declaration ;

/* constructor declaration */

constructor_head <CmmCtorHead>:
  spec:constructor_decl_specifiers
  ( '~' <set destructor=true> )?

  name:declarator_id
  cont:cont_function_declarator

  ( ':' init: ctor_initializer )?
  ( '{' <set body=true>
  | "try" <set try_block=true>
  | ';' <set empty=true> )
  ;

constructor_declaration <CmmCtorDecl:CmmDecl>:
  spec:constructor_decl_specifiers
  ( '~' <set destructor=true> )?
  decl:declarator
  (
     ( ':' init:ctor_initializer )?
     <new_line>
     statements:compound_stat
  |
     <new_line>
     "try"
     ( ':' init:ctor_initializer )?
     <new_line>
     statements:compound_stat
     handlers:handler_list
  |
     ';' <set empty=true>
  );

constructor_decl_specifiers <CmmSpecSect>:
  ( <add> constructor_decl_specifier )* ;

constructor_decl_specifier <CmmSpec>:
  "inline"   <set kind = a_inline>   |
  "virtual"  <set kind = a_virtual>  |
  "explicit" <set kind = a_explicit> ;

/* function declaration */

function_head <CmmFuncHead>:
  spec:decl_specifiers

  // decl:declarator

  ptr:ptr_specifier_list
  name:declarator_id // !?
  params:cont_function_declarator
  cont:cont_declarator_list

  (
     <new_line>
     '{' <set body=true>
  |
     <new_line>
     "try" <set try_block = true>
  |
     ';' <set empty=true>
  |
     '=' zero:numeric_literal ';' <set abstract=true>
  );

function_declaration <CmmFuncDecl:CmmDecl>:
  spec:decl_specifiers
  decl:declarator
  (
     <new_line>
     statements:compound_stat
  |
     <new_line>
     "try"
     statements:compound_stat
     handlers:handler_list
  |
     ';' <set empty=true>
  |
     '=' zero:numeric_literal ';' <set abstract=true>
  );

/* simple declaration */

simple_declaration <CmmSimpleDecl:CmmDecl>:
  spec:decl_specifiers
  (
     <add> init_declarator
     ( ',' <add> init_declarator )*
  )?
  ';' ;

init_declarator <CmmInitDeclarator>:
  decl:declarator
  ( init:initializer )?
  // ( '=' init:conditional_expr )? // !? @@
  ;

/* empty declaration */

empty_declaration <CmmEmptyDecl:CmmDecl>:
  ';' ;

/* ---------------------------------------------------------------------- */

/* declaration specifiers */

decl_specifiers <CmmDeclSpec>: // !?
  mode:simple_decl_specifiers

  ( name:id_expr
    ( template_args:template_arg_list )? // !?
  |
    class_spec:class_specifier
  |
    enum_spec:enum_specifier
  |
    type:simple_type_specifiers
  )

  mode2:simple_decl_specifiers ;

simple_decl_specifiers <CmmSpecSect>:
  ( <add> simple_decl_specifier )* ;

simple_decl_specifier <CmmSpec>:
  "auto"     <set kind = a_auto>     |
  "register" <set kind = a_register> |
  "static"   <set kind = a_static>   |
  "extern"   <set kind = a_extern>   |
  "mutable"  <set kind = a_mutable>  |

  "inline"   <set kind = a_inline>   |
  "virtual"  <set kind = a_virtual>  |
  "explicit" <set kind = a_explicit> |

  "friend"   <set kind = a_friend>   |
  "typedef"  <set kind = a_typedef>  |

  "const"    <set kind = a_const>    |
  "volatile" <set kind = a_volatile> ;

simple_type_specifiers <CmmTypeSect>:
  <add> simple_type_specifier
  ( <add> simple_type_specifier )* ;

simple_type_specifier <CmmType>:
  "char"     <set kind = a_char>     |
  "wchar_t"  <set kind = a_wchar>    |
  "bool"     <set kind = a_bool>     |
  "short"    <set kind = a_short>    |
  "int"      <set kind = a_int>      |
  "long"     <set kind = a_long>     |
  "signed"   <set kind = a_signed>   |
  "unsigned" <set kind = a_unsigned> |
  "float"    <set kind = a_float>    |
  "double"   <set kind = a_double>   |
  "void"     <set kind = a_void>     ;

/* ---------------------------------------------------------------------- */

alloc_type_specifier_list <CmmAllocTypeSpecSect>:
  <add> alloc_type_specifier
  ( <add> alloc_type_specifier )* ; // !? @@

alloc_type_specifier <CmmAllocTypeSpec>:
  "const" <set a_const = true> |
  "volatile" <set a_volatile = true> |

  name:id_expr | // !?
  class_spec:class_specifier |
  enum_spec:enum_specifier |

  simple:simple_type_specifier ;

/* ---------------------------------------------------------------------- */

/* enum */

enum_specifier <CmmEnumDecl:CmmDecl>:
  "enum"
  ( name:id_expr )? // !?
  <new_line> '{' <indent> items:enum_list <unindent> '}' ;

enum_list <CmmEnumSect>:
  ( <add> enumerator ( ',' <add> enumerator )* )?;

enumerator <CmmEnumItem>:
  <new_line> name:identifier ( '=' init:const_expr )? ;

/* namespace */

namespace_definition <CmmNamespace:CmmDecl> :
  "namespace"
  (
     <new_line>
     '{' <indent> unnamed_body:declaration_list <unindent> '}'
  |
     namespace_name:identifier // !?
     (
        '=' value:id_expr ';'
     |
        <new_line>
        '{' <indent> body:declaration_list <unindent> '}'
     )
  );

/* using */

using_declaration <CmmUsingDecl:CmmDecl>:
  "using"
  (
     "namespace" <set a_namespace = true>
  |
     "typename"  <set a_typename  = true>
  )?
  name: id_expr
  ';'
  ;

/* extern "..." */

linkage_specification <CmmExternDecl:CmmDecl>:
  "extern"
  language:string_literal
  (
    '{' decl_list:declaration_list '}'
  |
    decl:declaration
  );

/* asm declaration */

asm_declaration <CmmAsmDecl:CmmDecl>:
  "asm" '(' instruction:string_literal ')' ';' ;

/* ---------------------------------------------------------------------- */

/* type id */

type_id <CmmTypeId>:
  spec:decl_specifiers
  decl:abstract_declarator
  ;

common_type_id <CmmCommonTypeId>:
  spec:decl_specifiers
  decl:common_declarator
  ;

/* declarator */

declarator <CmmDeclarator>:
  ptr:ptr_specifier_list
  (
     name:declarator_id
  |
     '(' inner:declarator ')'
  )
  cont:cont_declarator_list
  ;

abstract_declarator <CmmAbstractDeclarator>:
  ptr:ptr_specifier_list
  (
     '(' inner:abstract_declarator ')' // !? @@
      cont:cont_declarator_list
  )?
  ;

common_declarator <CmmCommonDeclarator>:
  ptr:ptr_specifier_list
  (
     (
       name:declarator_id
     |
       '(' inner:common_declarator ')'
     )
     cont:cont_declarator_list
  )?
  ;

cont_declarator_list <CmmContDeclaratorSect>:
  ( <add> cont_declarator )*
  ;

cont_declarator <select CmmContDeclarator>:
  cont_function_declarator
  | cont_array_declarator ;

cont_function_declarator <CmmContFunctionDeclarator:CmmContDeclarator>:
  '(' parameters:parameter_declaration_list ')'
  cv_spec:cv_specifier_list
  ( exception_spec:exception_specification )? ;

cont_array_declarator <CmmContArrayDeclarator:CmmContDeclarator>:
  '[' (lim:expr)? ']' ;

declarator_id <return CmmNameExpr>:
  id_expr ;

ptr_specifier_list <CmmPtrSpecifierSect>:
  ( <add> ptr_specifier )* ;

ptr_specifier <CmmPtrSpecifier>:
    '*' cv_spec:cv_specifier_list <set pointer=true>
  | '&' <set reference=true>
  // |  (id_expr "::" '*') =>
  //    name2:id_expr "::" '*'
  //    cv_spec2:cv_specifier_list // !? @@
  ;

cv_specifier_list <CmmCvSpecifierSect>:
  ( <add> cv_specifier )*;

cv_specifier <CmmCvSpecifier>:
  "const" <set a_const = true> |
  "volatile" <set a_volatile = true> ;

/* parameter declarations */

parameter_declaration_list <CmmParamSect>:
  (
     <add> parameter_declaration
     ( ',' <add> parameter_declaration )*
  )? ;

parameter_declaration <CmmParamItem>:
  (
     type:common_type_id
     ( '=' init:assignment_expr )?
  |
     "..." <set dots=true> // !? @@
  );

/* initializer */

initializer <CmmInitializer>:
   '=' value:initializer_item ;

initializer_item <select CmmInitItem>:
     simple_initializer
   | initializer_list ;

simple_initializer <CmmInitSimple:CmmInitItem>:
    expr:assignment_expr ;

initializer_list <CmmInitList:CmmInitItem>:
    '{'
    <indent>
    (
      <add> initializer_item
      (
        ','
        <new_line>
        <add> initializer_item
      )*
      ( ',' <set additional_comma=true> )? // !? @@
    )?
    <unindent>
    '}';

/* ---------------------------------------------------------------------- */

/* class */

class_specifier <CmmClassDecl:CmmDecl> :
  ( "class"  <set style = ClassStyle>
  | "struct" <set style = StructStyle>
  | "union"  <set style = UnionStyle> )
  (
    ( ':' base_list:base_specifier_list )?
     <new_line>
    '{'
     <indent>
     members:member_list
     <unindent>
    '}'
    <set anonymous=true>
  |
     name:id_expr // !?
     // ( <silent> template_args:template_arg_list )? // !? @@
     (
        (':') =>
           ':' base_list:base_specifier_list
            <new_line>
           '{'
            <indent>
            members:member_list
            <unindent>
           '}'
     |
        <new_line>
       '{'
        <indent>
        members:member_list
        <unindent>
       '}'
     )?
  );

/* members */

member_list <CmmDeclSect>:
  ( <add> member_item <new_line> )* ;

member_item <select CmmDecl>:
  member_visibility |
  ( constructor_head ) => <indent> constructor_declaration <unindent> |
  ( function_head ) => <indent> function_declaration <unindent> |
  <indent> member_declaration <unindent> |
  <indent> empty_declaration <unindent> |
  <indent> using_declaration <unindent> |
  <indent> template_declaration <unindent> ;

member_visibility <CmmMemberVisibility:CmmDecl>:
  ( "private"   <set access = PrivateAccess>
  | "protected" <set access = ProtectedAccess>
  | "public"    <set access = PublicAccess> )
  <no_space> ':' ;

member_declaration <CmmSimpleDecl:CmmDecl>:
  spec:decl_specifiers
  (
     <add> member_declarator
     ( ',' <add> member_declarator )*
  )?
  ';' ;

member_declarator <CmmInitDeclarator>:
  decl:declarator
  ( init:initializer | ':' width:conditional_expr )? // !?
  ;

/* base specification */

base_specifier_list <CmmBaseSect>:
  <add> base_specifier ( ',' <add> base_specifier )* ;

base_specifier <CmmBaseItem>:
  ("virtual" <set a_virtual = true> )?

  ( "private"   <set access = PrivateAccess>
  | "protected" <set access = ProtectedAccess>
  | "public"    <set access = PublicAccess>
  |             <set access = NoAccess> )

  // ("virtual" <set a_virtual2=true> )?

  from:id_expr
  ( template_args:template_arg_list )? // !?
  ;

/* constructor initializer */

ctor_initializer <CmmCtorInitializer>:
  <indent>
  <add> mem_initializer
  ( ',' <new_line> <add> mem_initializer )*
  <unindent> ;

mem_initializer <CmmMemberInitializer>:
   name:id_expr '(' params:expr_list ')';

/* ---------------------------------------------------------------------- */

/* special member functions */

special_function <CmmSpecialFuction>:
   "operator"
   (
     // !? @@
     // spec:type_specifier_list
     // ptr:ptr_specifier_list

     "new"    <set spec_new=true>    // ( "[" "]" <set spec_new_array=true> )?
   | "delete" <set spec_delete=true> // ( "[" "]" <set spec_delete_array=true> )?

   |  "+" <set spec_add=true>
   |  "-" <set spec_sub=true>
   |  "*" <set spec_mul=true>
   |  "/" <set spec_div=true>
   |  "%" <set spec_mod=true>
   |  "^" <set spec_xor=true>
   |  "&" <set spec_and=true>
   |  "|" <set spec_or=true>
   |  "~" <set spec_not=true>

   |  "!"  <set spec_log_not=true>
   |  "="  <set spec_eq=true>
   |  "<"  <set spec_lt=true>
   |  ">"  <set spec_gt=true>
   |  "+=" <set spec_add_assign=true>
   |  "-=" <set spec_sub_assign=true>
   |  "*=" <set spec_mul_assign=true>
   |  "/=" <set spec_div_assign=true>
   |  "%=" <set spec_mod_assign=true>

   |  "^="  <set spec_xor_assign=true>
   |  "&="  <set spec_and_assign=true>
   |  "|="  <set spec_or_assign=true>
   |  "<<"  <set spec_shl=true>
   |  ">>"  <set spec_shr=true>
   |  ">>=" <set spec_shl_assign=true>
   |  "<<=" <set spec_shr_assign=true>
   |  "=="  <set spec_eq=true>
   |  "!="  <set spec_ne=true>

   |  "<="  <set spec_le=true>
   |  ">="  <set spec_ge=true>
   |  "&&"  <set spec_log_and=true>
   |  "||"  <set spec_log_or=true>
   |  "++"  <set spec_inc=true>
   |  "--"  <set spec_dec=true>
   |  ","   <set spec_comma=true>
   |  "->*" <set spec_mebber_deref=true>
   |  "->"  <set spec_deref=true>

   |  "(" ")" <set spec_call=true>
   |  "[" "]" <set spec_index=true>
   ) ;

/* ---------------------------------------------------------------------- */

/* template declaration */

template_declaration <CmmTemplateDecl:CmmDecl>:
  ( "export" <set export_template = true> )?
  "template"
  (
     '<'
     ( params:template_param_list | <set explicit_specialization = true>)
     '>'
     <set parameter_block = true> // !?
  |
     <set explicit_instatiation = true>
  )
  body:declaration ;

template_param_list <CmmTemplateParamSect>:
  <add> template_param ( ',' <add> template_param )* ;

template_param <select CmmTemplateParam>:
    (template_extra_param) => template_extra_param // !?
  | (template_type_param) => template_type_param
  | template_value_param ;

template_type_param <CmmTemplateTypeParam:CmmTemplateParam>:
  ( "class" <set a_class = true> |
    "typename" <set a_typename = true> )
  ( name:identifier )?
  ( '=' value:type_id )? ;

template_extra_param <CmmTemplateExtraParam:CmmTemplateParam>:
  "template" '<' params:template_param_list '>' "class"
  ( name:identifier )?
  ( '=' value:id_expr )? ;

template_value_param <CmmTemplateValueParam:CmmTemplateParam>:
  type:common_type_id
  ( '=' init:shift_expr )? ; // !?

/* template arguments */

template_arg_list <CmmTemplateArgSect>:
   '<' ( <add> template_arg ( ',' <add> template_arg )* )? '>';

template_arg <select CmmTemplateArg>:
   (type_id) => template_type_arg |
   template_value_arg;

template_type_arg <CmmTemplateTypeArg:CmmTemplateArg>:
   type:type_id;

template_value_arg <CmmTemplateValueArg:CmmTemplateArg>:
   value:shift_expr; // !?

/* ---------------------------------------------------------------------- */

/* try */

try_stat <CmmTryStat:CmmStat>:
  "try"
  body:compound_stat
  handlers:handler_list;

handler_list <CmmHandlerSect>:
  <add> handler (<add> handler)* ;

handler <CmmHandlerItem>:
  "catch"
  '('
  exception:exception_declaration
  ')'
  body:compound_stat ;

exception_declaration <CmmExceptionDecl>:
    type:common_type_id
  | '...' <set dots=true>
  ;

throw_expr <CmmThrowExpr:CmmExpr>:
  "throw" (expr:assignment_expr) ? ;

exception_specification <CmmExceptionSpec>:
  "throw"
  '('
  ( <add> type_id (',' <add> type_id)* )?
  ')' ;

/* ---------------------------------------------------------------------- */
