/* input.cc */

/* Common parser functions */

#include "input.h"

#include "str.h"
#include "conv.h"
#include "inf.h"

#include "pas_parser.hpp" // class PasParser

OPEN_NAMESPACE

/********************************* LEXER **********************************/

void PasLexer::translate ()
{
   actual = NONE; // !?

   switch (TokenKind)
   {
      case IdentToken:
         actual = keyword (LowerCase (TokenTxt));
         if (actual == NONE)
            actual = IDENT;
         break;

      case NumberToken:
         actual = NUM_INT;
         break;

      case CharToken:
         actual = STRING_LITERAL;
         break;

      case StringToken:
         actual = STRING_LITERAL;
         break;

      case SeparatorToken:
         actual = delimiter ();
         if (actual == NONE)
            Error ("Unknown separator");
         TokenSym = 0;
         break;

      default: ;
   }
}


/**************************************************************************/

string PasLexer::recall_identifier ()
{
   if (actual != IDENT)
      Error ("Name expected");
   string s = TokenTxt;
   next ();
   return s;
}

string PasLexer::recall_number ()
{
   if (actual != NUM_INT)
      Error ("Numeric literal expected");
   string s = TokenTxt;
   next ();
   return s;
}

string PasLexer::recall_real_number ()
{
   if (actual != NUM_REAL)
      Error ("Floating point numeric literal expected");
   string s = TokenTxt;
   next ();
   return s;
}

string PasLexer::recall_string_literal ()
{
   if (actual != STRING_LITERAL)
      Error ("String value expected");
   string s = TokenTxt; // !?
   next ();
   return s;
}

/* ---------------------------------------------------------------------- */

CLOSE_NAMESPACE
