
/* main.cc */

#include <iostream>
#include <fstream>

#include "std.h"
#include "conv.h"
#include "inf.h"
// #include "support.h"
// #include "connection.h"

#include "pas_lexer.hpp"
#include "pas_parser.hpp"
#include "pas_product.hpp"

OPEN_NAMESPACE

/* ---------------------------------------------------------------------- */

// global variables
string inp_name = "";
string out_name = "";

/* ---------------------------------------------------------------------- */

string ident_subst (string value) { return ""; }

/* ---------------------------------------------------------------------- */

CLOSE_NAMESPACE

USE_NAMESPACE

extern "C" int main (int argc, char * argv [])
{
   int err_code = 0;
   set_prog_name ("scan");

   set_error_handlers ();
   enable_error_trace ();

   // debug
   #ifndef NDEBUG
      if (argc == 1)
      {
         // #define NAME "examples"
         #define NAME "idedemo"

         static char * tab [] =
         {
          "",
          "./" NAME ".pas",
          "-o",
          NAME ".dump",
         };

         argc = sizeof (tab) / sizeof (tab[0]);
         argv = tab;
      }
   #endif

   bool o_opt = false;

   for (int i = 1; i < argc; i++)
   {
      string txt = argv [i];

      if (o_opt)
      {
         if (txt[0] == '-')
            error ("Output file name expected");
         if (out_name != "")
            error ("Output file name already defined");
         out_name = txt;
         o_opt = false;
      }

      else if (txt == "-o" || txt == "--output")
         o_opt = true;

      else
      {
         if (txt[0] == '-')
            error ("Unknown option");
         if (inp_name != "")
            error ("Input file name already defined");
         inp_name = txt;
      }
   }

   if (inp_name == "") error ("Missing input file name");
   if (out_name == "") error ("Missing output file name");

   // if (line_numbers)
   //    comments = false;

   // if (out_dir != "" && out_dir [out_dir.length()] != "/")
   //    out_dir = out_dir + '/';

   try
   {
      std::cout << get_prog_name () << ": " << inp_name << " -> " << out_name << std::endl;

      // std::fstream inp (inp_name.c_str (), std::ios::in);
      // if (! inp.good ())
      //    error ("Cannot open input file: " + inp_name);
      // PasParser * parser = PasParser::open_parser (inp, inp_name, 0);

      try
      {
         PasParser * parser = new PasParser;

         parser->opt.LowerCaseIdentifiers = true;
         parser->opt.PascalNumbers = true;
         parser->opt.PascalStrings = true;
         parser->opt.PascalComments = true;

         TTextEnv * env = new TTextEnv (); // !? delete
         parser->SetEnv (env);

         parser->Open (inp_name);
         parser->next ();

         TModule * m = parser->module_decl ();
         // parser->CheckEnd ();

         if (out_name != "")
         {
            PasProduct product;
            product.print_open (out_name);
            product.send_module_decl (m);
            product.print_close ();
         }
      }
      /*
      catch (antlr::RecognitionException & e)
      {
          show_error (e.getMessage (), e.getFilename (), e.getLine (), e.getColumn ());
          err_code = 1;
      }
      catch (antlr::ANTLRException & e)
      {
         antlr::LexerSharedInputState state = parser->lexer->getInputState ();
         show_error (e.getMessage (), state->filename, state->line, state->column);
         err_code = 1;
      }
      */
      catch (std::exception & e)
      {
         show_error (e.what());
         err_code = 1;
      }
   }
   catch (std::exception & e)
   {
      show_error (string ("exception: ") + e.what ());
      err_code = 1;
   }

   if (err_code == 0)
      show_info ("O.K.");

   return err_code;
}
