/* writer.cc */

#include "writer.h"

// #include <fstream>
#include "memstream.h"

#include "str.h"
#include "conv.h" // function IntToStr

// OPEN_NAMESPACE

/***************************** INITIALIZATION *****************************/

// class variables

Writer::Writer () :

   input_file_name (""),
   output_file_name (""),
   print_stream (NULL),

   indentation (0),

   space (false),
   start_line (true),

   add_empty_line (false),
   ignore_empty_line (false),

   skip_spaces (false)
{
   // nothing
}

/******************************* INPUT NAME *******************************/

void Writer::set_input_name (const string file_name)
{
    // store file name
    input_file_name = file_name;
}

string Writer::get_input_name ()
{
    return input_file_name;
}

/********************************* PRINT **********************************/

void Writer::print_open (const string file_name)
{
   assert (print_stream == NULL);

   // open output file
   output_file_name = file_name;

   // print_stream = new ofstream (file_name.c_str());
   print_stream = new MemStream;
   print_stream -> open (file_name.c_str());

   start_line = true;
   space = false;

   indentation = 0;

   add_empty_line = false;
   ignore_empty_line = false;
}

void Writer::print_close ()
{
   // print_stream->close ();
   print_stream->store ();

   delete print_stream;
   print_stream = NULL;
}

/* ---------------------------------------------------------------------- */

void Writer::print_open_string ()
{
   print_open ("");
}

string Writer::print_close_string ()
{
   string result = print_stream->get_text ();

   delete print_stream;
   print_stream = NULL;

   return result;
}

/* ---------------------------------------------------------------------- */

void Writer::printeol ()
{
   bool empty = start_line; // store value from start_line

   // end of line
   (*print_stream) << endl;

   start_line = true;
   space = false;

   add_empty_line = false;
   ignore_empty_line = false;

   if (empty)
      ignore_empty_line = true;
}

void Writer::print_cond_eol ()
{
   // finish not empty line
   if (! start_line)
      printeol ();
}

void Writer::print (const string s)
{
   if (add_empty_line && ! ignore_empty_line)
      printeol (); // print empty line

   if (start_line)
   {
      // indentation at line beginning
      for (int i=1; i<=indentation; i++)
         (*print_stream) << "   ";
      start_line = false;
      space = false;
   }

   (*print_stream) << s;
}

void Writer::println (const string s)
{
   // print string and finish line

   if (s != "")
      print (s);

   printeol ();
}

/* ---------------------------------------------------------------------- */

void Writer::indent ()
{
   indentation ++;
}

void Writer::unindent ()
{
   indentation --;
}

/********************************** SEND **********************************/

void Writer::send (string s)
{
   // add optional space and print string

   char c = ' ';
   if (s.length () >= 1) c = s[0];

   if (c == ',' || c == ';' || c == ')' || c == ']' ) space = false;

   if (space && ! skip_spaces) print (" ");
   print (s);

   space = true;
   if (c == '(' || c == '[' || c == '#') space = false;
}

void Writer::sendeol ()
{
   printeol ();
}

void Writer::sendln (string s)
{
   send (s);
   sendeol ();
}

/********************************* STYLE **********************************/

void Writer::style_space ()
{
   space = true;
}

void Writer::style_no_space ()
{
   space = false;
}

/* ---------------------------------------------------------------------- */

void Writer::style_new_line ()
{
   // finish line and start new one
   print_cond_eol ();
}

void Writer::style_empty_line ()
{
   // add one empty line
   print_cond_eol (); // quoted stensils - use send, not print
   add_empty_line = true;
}

void Writer::style_no_empty_line ()
{
   // ignore style_empty_line ()
   ignore_empty_line = true;
}

/* ---------------------------------------------------------------------- */

void Writer::style_indent ()
{
   // indent recomendation
   style_new_line ();
   indent ();
}

void Writer::style_unindent ()
{
   unindent ();
   style_new_line ();
}

/* ---------------------------------------------------------------------- */

// CLOSE_NAMESPACE
