
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

class CodeOutput extends Tool
{

  /* enum */

  private void send_enumerations ()
  {
     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof EnumInfo)
        {
           EnumInfo e = (EnumInfo) t;

           println ("enum " + info.type_prefix + e.ident);
           println ("{");
           indent ++;

           Enumeration items = e.getItems ();
           while (items.hasMoreElements ())
           {
              DictItem item = (DictItem) items.nextElement ();
              print (item.ident);

              if (items.hasMoreElements ()) print (", ");
              println ("");
           }

           indent --;
           println ("};");
           println ("");
        }
     }
  }

  /* structures */

  private void send_func_header (String func_name, String type_name, String field_name)
  {
     String pref = info.internal_prefix;

     StructInfo s = info.findStruct (type_name);
     FieldInfo f = s.findField (field_name);

     print ("virtual bool " + pref + func_name +
             " (" + f.type.type_name () + " " + pref + "param) ");
  }

  private void send_basic_func (String func_name, String type_name, String field_name)
  {
     send_func_header (func_name, type_name, field_name);
     println ("{ return false; }");
  }

  private void send_func (String func_name, String type_name, String field_name)
  {
     send_func_header (func_name, type_name, field_name);
     println ("{ return " + info.field_prefix + field_name + " == " +
              info.internal_prefix + "param; }");
  }

  private void send_basic_type ()
  {
     println ("class " + info.basic_type);
     println ("{");
     indent ++;
     println ("public:");
     indent ++;

     String pref = info.internal_prefix;
     if (info.tree_code)
        println ("virtual string " + pref + "getTypeName () { return " + '"' + info.basic_type + '"' + "; }");

     if (info.copy_method)
        println ("virtual " + info.basic_type + " * " + pref + "copy ();");

     if (info.tree_code)
     {
        String location_type = info.location_type;

        println ("virtual void " + pref + "find_subst (" + info.writer_decl + ") {}");

        println ("virtual void " + pref + "recursive_find_subst (" + info.writer_decl + ") " +
                 "{ " + pref + "find_subst "+ " (" + info.writer_param + "); }");

        println ("virtual string " + pref + "produce (" + info.writer_decl + ");");
        println ("virtual string " + pref + "produce_subst (" + info.writer_decl + ") { return " + '"' + '"' + "; }");
        println ("virtual string " + pref + "produce_list_subst (" + info.writer_decl + ", string & item_name) { return " + '"' + '"' + "; }");
        println ("virtual void " + pref + "produce_attributes (" + info.writer_decl + ") {}");
        println ("virtual void " + pref + "produce_sequence (" + info.writer_decl + ") {}");
        if (info.locations)
           println ("virtual " + location_type + " " + pref + "getFirstLocation () { return 0; }");

        println ("virtual bool " + pref + "isCustomized () { return false; }");
        println ("virtual bool " + pref + "isSimple () { return true; }");
        println ("virtual bool " + pref + "haveSubitems () { return false; }");
     }

     if (info.getStruct (info.unary_expr_type) != null)
        send_basic_func ("isUnary", info.unary_expr_type, info.unary_expr_selector);

     if (info.getStruct (info.binary_expr_type) != null)
        send_basic_func ("isBinary", info.binary_expr_type, info.binary_expr_selector);

     // methods
     Enumeration methods = info.getBasicTypeMethods ();
     while (methods.hasMoreElements ())
     {
        MethodInfo m = (MethodInfo) methods.nextElement ();
        send_additional_method (m);
     }

     if (! empty (info.subst_type))
        println (info.subst_type + " * " + info.internal_prefix + "subst_obj;");

     if (! empty (info.replace_type))
        println (info.replace_type + " * " + info.internal_prefix + "replace_obj;");

     if (! empty (info.subst_type) || ! empty (info.replace_type))
     {
        // constructor
        print (info.basic_type + " () : ");
        if (! empty (info.subst_type))
        {
           print (info.internal_prefix + "subst_obj (NULL)");
           if (! empty (info.replace_type))
              print (",");
           print (" ");
        }
        if (! empty (info.replace_type))
           print (info.internal_prefix + "replace_obj (NULL) ");
        println ("{}");
     }

     println ("virtual ~ " + info.basic_type + " () {}");
     indent --;

     // add declarations from extensions
     print_extension ("common_class");

     indent --;
     println ("};");
     println ("");
  }

  private void declare_variable (String type,
                                 String name,
                                 int count,
                                 boolean fixed)
  {
     if (fixed)
        for (int i = 1; i <= count; i++)
           println (type + " " + name + '_' + i + ";");
     else
        println (type + " " + name + " [" + count + "];");
  }

  private void send_additional_method (MethodInfo m)
  {
     if (m.virtual_func)
        print ("virtual ");

     print (m.result_decl + " " + m.ident + " (" + m.params_decl + ")");

     if (! empty (m.return_value))
        print (" { return " + m.return_value + "; }");
     else
        print (";");

     println ("");
  }

  private void send_structure (StructInfo s)
  {
     if (! s.declared)
     {
        if (s.from != null)
           send_structure (s.from);
        s.declared = true;

        print ("class " + s.type_name () + " : public " );
        if (s.from != null)
           print (s.from.type_name ());
        else
           print (info.basic_type);
        println ("");

        println ("{");
        indent ++;

        println ("public:");
        indent ++;

        // parameters
        if (s.from != null)
          s.copyParams (s.from);

        boolean any = false;

        // fields
        Enumeration fields = s.getFields ();
        while (fields.hasMoreElements ())
        {
           FieldInfo f = (FieldInfo) fields.nextElement ();
           if (f.is_const)
           {
              println ("static const " + f.type.inner_decl() + " " +
                       info.field_prefix + f.ident + " = " +
                       f.const_value + ";");
           }
           else
           {
              println (f.type.inner_decl() + " " + info.field_prefix + f.ident + ";");
              s.addParam (f);
           }
           any = true;
        }

        if (any) println ("");
        any = false;

        // places
        Enumeration places = s.getPlaces ();
        while (places.hasMoreElements ())
        {
           FieldInfo f = (FieldInfo) places.nextElement ();
           println (f.type.inner_decl() + " " + info.internal_prefix + f.ident + ";");
           any = true;
        }

        if (any) println ("");
        any = false;

        // methods
        Enumeration methods = s.getMethods ();
        while (methods.hasMoreElements ())
        {
           MethodInfo m = (MethodInfo) methods.nextElement ();
           send_additional_method (m);
           any = true;
        }

        if (any) println ("");
        any = false;

        String pref = info.internal_prefix;

        // next pointers
        Enumeration anchors = s.getAnchors ();
        while (anchors.hasMoreElements ())
        {
           SequenceInfo t = (SequenceInfo) anchors.nextElement ();

           println (t.owner.inner_decl () +
                    " " + info.list_prefix + "up" + t.suffix + ";");

           println (t.element.inner_decl () +
                    " " + info.list_prefix + "prev" + t.suffix + ";");

           println (t.element.inner_decl () +
                    " " + info.list_prefix + "next" + t.suffix + ";");

           println ("void " + info.list_prefix + "insertBefore" + t.suffix +
                    " (" + t.element.inner_decl () + " item);");

           println ("void " + info.list_prefix + "insertAfter" + t.suffix +
                    " (" + t.element.inner_decl () + " item);");

           println ("void " + info.list_prefix + "remove" + t.suffix + " ();");

           any = true;
        }

        // first, last, add
        Enumeration sequences = s.getSequences ();
        while (sequences.hasMoreElements ())
        {
           SequenceInfo t = (SequenceInfo) sequences.nextElement ();

           println (t.element.inner_decl () + " " + info.list_prefix + "first" + t.suffix + ";");
           println (t.element.inner_decl () + " " + info.list_prefix + "last" + t.suffix + ";");

           println ("void " + info.list_prefix + "insertFirst" + t.suffix +
                    " (" + t.element.inner_decl () + " item);");

           println ("void " + info.list_prefix + "insertLast" + t.suffix +
                    " (" + t.element.inner_decl () + " item);");

           println ("void " + info.list_prefix + "add" + t.suffix +
                    " (" + t.element.inner_decl () + " item)" +
                    " { " + info.list_prefix + "insertLast" + t.suffix + " (item); }" );

           if (t.element.custom_list_replace)
              println ("void " + info.list_prefix + "add_list" + t.suffix +
                       " (" + t.element.inner_decl () + " item);");

           any = true;
        }

        if (any) println ("");

        if (s.note_count != 0)
        {
           if (info.locations)
              declare_variable (info.location_type, pref + "location",
                                s.note_count, info.fixed_notes);

           if (info.notes)
              declare_variable (info.note_type, pref + "start_note",
                                s.note_count, info.fixed_notes);

           if (info.notes && ! info.only_start_notes)
              declare_variable (info.note_type, pref + "stop_note",
                                s.note_count, info.fixed_notes);

           println ("");
        }

        // first location
        boolean first_location = false;
        if (s.note_count != 0) // some locations for this class
        {
           first_location = true;
           StructInfo t = s.from;
           while (t != null)
           {
              if (t.note_count != 0) // no locations for supper class
                 first_location = false;
              t = t.from;
           }
        }

        // constructor
        println (s.type_name () + " ();");

        // constructor with parameters
        int cnt = s.getParamCount ();
        s.create_func = (cnt <= StructInfo.param_limit && cnt != 0);

        if (s.create_func)
        {
           if (cnt == 1) print ("explicit ");
           print (s.type_name () + " (");
           for (int i = 0; i < cnt; i++)
           {
              if (i > 0) print (", ");
              FieldInfo f = s.getParam (i);
              print (f.type.inner_decl() + " " + info.param_prefix + f.ident);
           }
           println (");");
        }

        // destructor
        if (s.getSequences ().hasMoreElements ())
           println ("~" + s.type_name () + " ();");

        // methods
        println ("");
        println ("virtual string " + pref + "getTypeName () { return " + '"' + s.ident + '"' + "; }");

        if (info.copy_method)
        {
           println ("virtual " + s.type_name () + " * " +
                    info.internal_prefix + "copy" + " ();");

           println ("virtual void " +
                    info.internal_prefix + "copy_attributes " +
                    "(" + s.type_name () + " * " + info.param_prefix + "result" + ");");
        }

        if (first_location && info.locations)
        {
           s.get_first_loc = true;
           String location_type = info.location_type;
           println ("virtual " + location_type + " " + pref + "getFirstLocation (); ");
        }

        if (s.ident.equals (info.unary_expr_type))
           send_func ("isUnary", info.unary_expr_type, info.unary_expr_selector);

        if (s.ident.equals (info.binary_expr_type))
           send_func ("isBinary", info.binary_expr_type, info.binary_expr_selector);

        // produce
        if (info.tree_code)
        {
           if (s.custom_produce)
              println ("virtual string " + pref + "produce (" + info.writer_decl + ");");
              /* produce_attributes and produce_sequence
                 are generated even if s.custom_produce == true */

           if (s.custom_subst)
           {
              println ("virtual void " + pref + "find_subst (" + info.writer_decl + ");");
              println ("virtual string " + pref + "produce_subst (" + info.writer_decl + ");");
           }

           if (s.custom_list_subst)
              println ("virtual string " + pref + "produce_list_subst " +
                       "(" + info.writer_decl + ", string & item_name);");

           if (s.getFieldCount () != 0 && ! s.custom_no_produce)
              println ("virtual void " + pref + "produce_attributes " +
                       "(" + info.writer_decl + ");");

           if (s.hasSequences () && ! s.custom_no_produce)
               println ("virtual void " + pref + "produce_sequence " +
                        "(" + info.writer_decl + ");");

           if (s.custom_produce)
              println ("virtual bool " + pref + "isCustomized () { return true; }");

           if (cnt > StructInfo.param_limit &&
               (s.from == null ||
                s.from.getParamCount () <= StructInfo.param_limit))
              println ("virtual bool " + pref + "isSimple () { return false; }");

           if (s.hasSequences () )
           {
              print ("virtual bool " + pref + "haveSubitems () ");
              print ("{ return ");

              int local_cnt = 0;
              Enumeration local_sequences = s.getSequences ();
              while (local_sequences.hasMoreElements ())
              {
                 SequenceInfo t = (SequenceInfo) local_sequences.nextElement ();
                 local_cnt ++;
                 if (local_cnt > 1) print (" || ");
                 print (info.list_prefix + "first" + t.suffix + " != NULL");
              }

              println ("; }");
           }

           if (s.getFieldCount () != 0 || s.hasSequences ())
              println ("virtual void " + pref + "recursive_find_subst (" + info.writer_decl + "); ");

           if (s.custom_subst || s.custom_replace)
           {
              println (s.decl () + " " + pref + "replace (" + info.parser_class + " * parser); ");
              // println ("string " + pref + "ident_replace (string s); ");
           }
        }

        indent --;
        indent --;
        println ("};");
        println ("");
     }
  }

  private void send_structure_definitions (boolean all, boolean only_private)
  {
     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof StructInfo)
        {
           StructInfo s = (StructInfo) t;
           if (all || s.custom_private == only_private)
              send_structure (s);
        }
     }
  }

  private void send_structure_names ()
  {
     if (info.antlr_mode)
     {
        println ("class " + info.lexer_name + ";");

        if (! info.lexer_name.equals (info.lexer_class))
        {
           println ("typedef " + info.lexer_name + " " + info.lexer_class + ";");
           // ANTLR does not support different file and class names
           // name of grammar file is not important
           // generated files are named same as lexer and parser classes
        }
        println ("");

        println ("class " + info.parser_name + ";");
        if (! info.parser_name.equals (info.parser_class))
           println ("typedef " + info.parser_name + " " + info.parser_class + ";");
        println ("");
     }

     if (! info.antlr_mode)
     {
        println ("class " + info.parser_class + ";");
        println ("");
     }

     println ("class " + info.product_class + ";");
     println ("");

     if (! empty (info.replace_type))
     {
        println ("class " + info.replace_type + ";");
        println ("");
     }

     println ("class " + info.basic_type + ";");
     println ("");

     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof StructInfo)
        {
           StructInfo s = (StructInfo) t;
           println ("class " + s.type_name () + ";");
        }
     }
     println ("");
  }

  private void send_function_headers ()
  {
     Enumeration seq = info.getRules ();
     while (seq.hasMoreElements ())
     {
        RuleInfo r = (RuleInfo) seq.nextElement ();
        RuleNode n = r.node;
        if (n.custom_internal_send)
        {
           println ("void " + info.send_internal_prefix + n.name +
                    " (" + n.type.inner_decl() + " " + info.param_prefix + "item" + ");");
        }
        if (! n.custom_no_send)
        {
           println ("void " + info.send_func_prefix + n.name +
                    " (" + n.type.inner_decl() + " " + info.param_prefix + "item" + ");");
        }
     }
  }

  private void send_open_close_function_headers ()
  {
     boolean any = false;
     Enumeration seq = info.getRules ();
     while (seq.hasMoreElements ())
     {
        RuleInfo r = (RuleInfo) seq.nextElement ();
        RuleNode n = r.node;
        if (n.custom_open_close_send)
        {
           if (!any)
           {
              println ("");
              any = true;
           }

           println ("void " + info.open_func_prefix + n.name +
                    " (" + n.type.inner_decl() + " " + info.param_prefix + "item" + ");");

           println ("void " + info.close_func_prefix + n.name +
                    " (" + n.type.inner_decl() + " " + info.param_prefix + "item" + ");");
        }
     }
  }

  private void send_utility_function_headers ()
  {
     if (info.expr_method)
     {
        String func = info.send_func_prefix + info.binary_expr_func;
        String item  = info.param_prefix + "item";

        println ("");
        println ("void " + func +
                 " (" + info.type_prefix + info.expr_type + " * " + item + ");");
     }

     println ("");
     println ("// external utilities");
     // println ("");

     String pref = info.util_func_prefix;

     if (empty (info.product_super_class))
     {
        println ("void " + info.internal_prefix + "send (string s);");

        if (info.locations)
           println ("void " + info.internal_prefix + "sendLocation (" + info.location_type + " i);");

        if (info.notes)
        {
           println ("void " + info.internal_prefix + "sendStartNote (string s);");
           if (! info.only_start_notes)
              println ("void " + info.internal_prefix + "sendStopNote (string s);");
        }
        println ("");

        println ("void " + pref + "style_space ();");
        println ("void " + pref + "style_no_space ();");
        println ("void " + pref + "style_new_line ();");
        println ("void " + pref + "style_empty_line ();");
        println ("void " + pref + "style_no_empty_line ();");
        println ("void " + pref + "style_indent ();");
        println ("void " + pref + "style_unindent ();");
        println ("");
     }

     if (info.tree_code)
     {
        println ("void " + pref + "insert_item (" + info.basic_type + " * item);");
        println ("");
     }

     if (info.tree_code)
     {
        println ("void " + pref + "store_bool   (const string field, bool value);");
        println ("void " + pref + "store_int    (const string field, int value);");
        println ("void " + pref + "store_string (const string field, const string value);");

        println ("void " + pref + "store_enum   (const string field, const string value, int index);");
        println ("void " + pref + "store_ident  (const string field, const string value);");
        println ("void " + pref + "store_item   (const string field, " + info.basic_type + "* item);");
        println ("");
     }

     if (info.locations)
        println ("void " + info.internal_prefix + "store_location (" + info.location_type + " i);");
  }

  private void send_product_class ()
  {
     print ("class " + info.product_class);
     if (! empty (info.product_super_class))
        print (" : public " + info.product_super_class);
     println ("");

     println ("{");
     indent ++;

     println ("public:");
     indent ++;
     send_function_headers ();
     send_open_close_function_headers ();
     send_utility_function_headers ();
     indent --;

     // add declarations from extensions
     print_extension ("product_class");

     // add declarations from file
     if (! empty (info.product_input))
        copy (info.inp_dir, info.product_input, "");

     indent --;
     println ("};");
     println ("");
  }

  public void sendHeader ()
  {
     send_product_class ();
  }

  public void sendCommonHeader ()
  {
     print_extension ("common_header");

     send_enumerations ();
     send_structure_names ();

     if (info.enable_public_header)
     {
        // no structures
     }
     else if (info.enable_private_header)
     {
        send_basic_type ();
        send_structure_definitions (false, false); // public structures
     }
     else
     {
        send_basic_type ();
        send_structure_definitions (true, false); // all structures
     }

     print_extension ("common_tail");
  }

  public void sendPublicHeader ()
  {
     send_basic_type ();

     if (info.enable_private_header)
     {
        send_structure_definitions (false, false); // public structures
     }
     else
     {
        send_structure_definitions (true, false); // all structures
     }
  }

  public void sendPrivateHeader ()
  {
     send_structure_definitions (false, true); // private structures
  }

  /* METHOD IMPLEMENTATIONS */

  private void send_init (String name, String value)
  {
     println (",");
     print (name + " (" + value + ")");
  }

  private void init_variable (String name,
                              int count,
                              String value,
                              boolean fixed)
  {
     if (fixed)
        for (int i = 1; i <= count; i++)
           send_init (name + "_" + i, value);
     else
     {
        String txt = "";
        /*
        for (int i = 1; i <= count; i++)
        {
           if (i > 1) txt = txt + ", ";
           txt = txt + value;
        }
        */
        send_init (name, txt);
     }
  }

  private void init_array (String name,
                           int count,
                           String value)
  {
     String inx = info.variable_prefix + "i";
     println ("for (int " + inx + " = 0; " + inx + " < " + count + "; " + inx + "++)");
     indent ++;
     println (name + " [" + inx + "] = " + value + ";");
     indent --;
  }

  private void send_cont (StructInfo s)
  {
     String pref = info.internal_prefix;

     Enumeration anchors = s.getAnchors ();
     while (anchors.hasMoreElements ())
     {
        SequenceInfo t = (SequenceInfo) anchors.nextElement ();
        send_init (info.list_prefix + "up" + t.suffix, "NULL");
        send_init (info.list_prefix + "prev" + t.suffix, "NULL");
        send_init (info.list_prefix + "next" + t.suffix, "NULL");
     }

     Enumeration sequences = s.getSequences ();
     while (sequences.hasMoreElements ())
     {
        SequenceInfo t = (SequenceInfo) sequences.nextElement ();
        send_init (info.list_prefix + "first" + t.suffix, "NULL");
        send_init (info.list_prefix + "last" + t.suffix, "NULL");
     }

     if (s.note_count != 0)
     {
        if (info.locations)
           init_variable (pref + "location",   s.note_count, "0", info.fixed_notes);

        if (info.notes)
           init_variable (pref + "start_note", s.note_count, "",  info.fixed_notes);

        if (info.notes && ! info.only_start_notes)
           init_variable (pref + "stop_note",  s.note_count, "",  info.fixed_notes);
     }

     println ("");
     indent --;

     println ("{");
     indent ++;

     if (s.note_count != 0 && ! info.fixed_notes) // fixed_notes are initialized as normal scalar variables
     {
        if (! empty (info.location_init) && info.locations)
           init_array (pref + "location", s.note_count, info.location_init);

        if (! empty (info.note_init) && info.notes)
        {
           init_array (pref + "start_note", s.note_count, info.note_init);

           if (! info.only_start_notes)
              init_array (pref + "stop_note", s.note_count, info.note_init);
        }
     }

     indent --;
     println ("}");
     println ("");
  }

  private void send_insert_first_method (SequenceInfo s)
  {
     String suffix = s.suffix;

     String first = info.list_prefix + "first" + suffix;
     String last = info.list_prefix + "last" + suffix;

     String up = info.list_prefix + "up" + suffix;
     String prev = info.list_prefix + "prev" + suffix;
     String next = info.list_prefix + "next" + suffix;

     String item = info.param_prefix + "item";

     println ("void " + s.owner.type_name () + "::" + info.list_prefix + "insertFirst" + suffix +
              " (" + s.element.inner_decl () + " " + item + ")");
     println ("{");
     indent ++;

     println (item + "->" + up +   " = this;");
     println (item + "->" + prev + " = NULL;");
     println (item + "->" + next + " = " + first + ";");
     println ("");

     println ("if (" + first + " != NULL)");
     indent ++;
     println (first + "->" + prev + " = " + item + ";");
     indent --;
     println ("else");
     indent ++;
     println (last + " = " + item + ";");
     indent --;
     println ("");

     println (first + " = " + item + ";");

     indent --;
     println ("}");
     println ("");
  }

  private void send_insert_last_method (SequenceInfo s)
  {
     String suffix = s.suffix;

     String first = info.list_prefix + "first" + suffix;
     String last = info.list_prefix + "last" + suffix;

     String up = info.list_prefix + "up" + suffix;
     String prev = info.list_prefix + "prev" + suffix;
     String next = info.list_prefix + "next" + suffix;

     String item = info.param_prefix + "item";

     println ("void " + s.owner.type_name () + "::" + info.list_prefix + "insertLast" + suffix +
              " (" + s.element.inner_decl () + " " + item + ")");
     println ("{");
     indent ++;

     println (item + "->" + up   + " = this;");
     println (item + "->" + prev + " = " + last + ";");
     println (item + "->" + next + " = NULL;");
     println ("");

     println ("if (" + last + " != NULL)");
     indent ++;
     println (last + "->" + next + " = " + item + ";");
     indent --;
     println ("else");
     indent ++;
     println (first + " = " + item + ";");
     indent --;
     println ("");

     println (last + " = " + item + ";");

     indent --;
     println ("}");
     println ("");
  }

  private void send_insert_before_method (SequenceInfo s)
  {
     String suffix = s.suffix;

     String first = info.list_prefix + "first" + suffix;
     String last = info.list_prefix + "last" + suffix;

     String up = info.list_prefix + "up" + suffix;
     String prev = info.list_prefix + "prev" + suffix;
     String next = info.list_prefix + "next" + suffix;

     String item = info.param_prefix + "item";

     println ("void " + s.element.type_name () + "::" + info.list_prefix + "insertBefore" + suffix +
              " (" + s.element.inner_decl () + " " + item + ")");
     println ("{");
     indent ++;

     println ("assert (" + up   + " != NULL);");
     println ("");

     println (item + "->" + up   + " = " + up + ";");
     println (item + "->" + prev + " = " + prev + ";");
     println (item + "->" + next + " = this;");
     println ("");

     println ("if (" + prev + " != NULL)");
     indent ++;
     println (prev + "->" + next + " = " + item + ";");
     indent --;
     println ("else");
     indent ++;
     println (up + "->" + first + " = " + item + ";");
     indent --;
     println ("");

     println (prev + " = " + item + ";");

     indent --;
     println ("}");
     println ("");
  }

  private void send_insert_after_method (SequenceInfo s)
  {
     String suffix = s.suffix;

     String first = info.list_prefix + "first" + suffix;
     String last = info.list_prefix + "last" + suffix;

     String up = info.list_prefix + "up" + suffix;
     String prev = info.list_prefix + "prev" + suffix;
     String next = info.list_prefix + "next" + suffix;

     String item = info.param_prefix + "item";

     println ("void " + s.element.type_name () + "::" + info.list_prefix + "insertAfter" + suffix +
              " (" + s.element.inner_decl () + " " + item + ")");
     println ("{");
     indent ++;

     println ("assert (" + up   + " != NULL);");
     println ("");

     println (item + "->" + up   + " = " + up + ";");
     println (item + "->" + prev + " = this;");
     println (item + "->" + next + " = " + next + ";");
     println ("");

     println ("if (" + next + " != NULL)");
     indent ++;
     println (next + "->" + prev + " = " + item + ";");
     indent --;
     println ("else");
     indent ++;
     println (up + "->" + last + " = " + item + ";");
     indent --;
     println ("");

     println (next + " = " + item + ";");

     indent --;
     println ("}");
     println ("");
  }

  private void send_remove_method (SequenceInfo s)
  {
     String suffix = s.suffix;

     String first = info.list_prefix + "first" + suffix;
     String last = info.list_prefix + "last" + suffix;

     String up = info.list_prefix + "up" + suffix;
     String prev = info.list_prefix + "prev" + suffix;
     String next = info.list_prefix + "next" + suffix;

     println ("void " + s.element.type_name () + "::" + info.list_prefix + "remove" + suffix + " ()");
     println ("{");
     indent ++;

     println ("assert (" + up   + " != NULL);");
     println ("");

     println ("if (" + prev + " != NULL)");
     indent ++;
     println (prev + "->" + next + " = " + next + ";");
     indent --;
     println ("else");
     indent ++;
     println (up + "->" + first + " = " + next + ";");
     indent --;
     println ("");

     println ("if (" + next + " != NULL)");
     indent ++;
     println (next + "->" + prev + " = " + prev + ";");
     indent --;
     println ("else");
     indent ++;
     println (up + "->" + last + " = " + prev + ";");
     indent --;
     println ("");

     println (up   + " = NULL;");
     println (prev + " = NULL;");
     println (next + " = NULL;");

     indent --;
     println ("}");
     println ("");
  }

  private void send_destructor_code (SequenceInfo s)
  {
     String suffix = s.suffix;

     String first = info.list_prefix + "first" + suffix;
     String next = info.list_prefix + "next" + suffix;

     String remove = info.list_prefix + "remove" + suffix;

     String item = info.param_prefix + "item";
     String temp = info.param_prefix + "temp";

     println (s.element.inner_decl () + " " + item + " = " + first + ";");

     println ("while (" + item + " != NULL)");
     println ("{");
     indent ++;

     println (s.element.inner_decl () + " " + temp + " = " + item + "->" + next + ";");
     println (item + "->" + remove + " ();");
     println ("delete " + item + ";");
     println (item + " = " + temp + ";");

     indent --;
     println ("}");
  }

  private void send_constructor (StructInfo s)
  {
     println (s.type_name () + "::" + s.type_name () + " () :");
     indent ++;

     if (s.from == null)
        print (info.basic_type + " ()");
     else
        print (s.from.type_name () + " ()");

     Enumeration fields = s.getFields ();
     while (fields.hasMoreElements ())
     {
        String txt = "";
        FieldInfo f = (FieldInfo) fields.nextElement ();

        if (! f.is_const)
        {
           if (f.type instanceof StructInfo) txt = "NULL";
           else if (f.type instanceof EnumInfo) txt = f.type.type_name () + " (0)";
           else if (f.type == info.bool_type) txt = "false";
           else if (f.type == info.int_type) txt = "0";
           else if (f.type == info.string_type) txt = "";
           else if (f.type == info.ident_type) txt = "";
           else if (f.type.is_pointer) txt = "NULL";
           else error ("Internal error");

           send_init (info.field_prefix + f.ident, txt);
        }
     }

     Enumeration places = s.getPlaces ();
     while (places.hasMoreElements ())
     {
        String txt = "";
        FieldInfo f = (FieldInfo) places.nextElement ();

        if (f.type == info.int_type) txt = "0";
        else if (f.type == info.string_type) txt = "";
        else error ("Internal error");

        send_init (info.internal_prefix + f.ident, txt);
     }

     send_cont (s);
  }

  private void send_constructor_with_params (StructInfo s)
  {
     print (s.type_name () + "::" + s.type_name () + " (");
     int cnt = s.getParamCount ();
     for (int i = 0; i < cnt; i++)
     {
        if (i > 0) print (", ");
        FieldInfo f = s.getParam (i);
        print (f.type.inner_decl() + " " + info.param_prefix + f.ident);
     }
     println (") :");
     indent ++;

     // parent
     int super_cnt = 0;
     if (s.from == null)
        print (info.basic_type + " ()");
     else
     {
        print (s.from.type_name ());
        print (" (");
        super_cnt = s.from.getParamCount ();
        for (int i = 0; i < super_cnt; i++)
        {
           if (i > 0) print (", ");
           FieldInfo f = s.from.getParam (i);
           print (info.param_prefix + f.ident);
        }
        print (")");
     }

     // own
     for (int i = super_cnt; i < cnt; i++)
     {
        FieldInfo f = s.getParam (i);
        send_init (info.field_prefix + f.ident, info.param_prefix + f.ident);
     }

     send_cont (s);
  }

  private void send_destructor (StructInfo s)
  {
     int count = 0;
     Enumeration sequences = s.getSequences ();
     while (sequences.hasMoreElements ())
     {
        sequences.nextElement ();
        count ++;
     }

     if (count != 0)
     {
        println (s.type_name () + "::~" + s.type_name () + " ()");
        println ("{");
        indent ++;

        sequences = s.getSequences ();
        while (sequences.hasMoreElements ())
        {
           SequenceInfo t = (SequenceInfo) sequences.nextElement ();
           if (count != 1)
           {
              println ("{");
              indent ++;
           }
           send_destructor_code (t);
           if (count != 1)
           {
              indent --;
              println ("}");
           }
        }

        indent --;
        println ("}");
        println ("");
     }
  }

  private void send_get_first_location_method (StructInfo s)
  {
     String pref = info.internal_prefix;
     String location_type = info.location_type;

     println (location_type + " " + s.type_name () + "::"  + pref + "getFirstLocation ()");
     println ("{");
     indent ++;

     String name = pref + "location";;
     int count_minus_one = s.note_count - 1;

     if (info.fixed_notes)
     {
        println ("return " + name + "_1;"); // !?
     }
     else
     {
        println ("int i = 0;");
        println ("while (i < " + count_minus_one + " && " + name + " [i] != 0)");
        indent ++;
        println ("i ++;");
        indent --;
        println ("return " + name + " [i];");
     }

     indent --;
     println ("}");
     println ("");
  }

  private void send_basic_copy_method ()
  {
     String result = info.variable_prefix + "result";

     println (info.basic_type + " * " +
              info.basic_type + "::" +
              info.internal_prefix + "copy" + " ()");

     println ("{");
     indent ++;

     println ("return new " + info.basic_type + ";");

     indent --;
     println ("}");
     println ("");
  }

  private void send_copy_method (StructInfo s)
  {
     String result = info.variable_prefix + "result";

     println (s.type_name () + " * " +
              s.type_name () + "::" +
              info.internal_prefix + "copy" + " ()");

     println ("{");
     indent ++;

     println (s.type_name () + " * " + result + " = new " + s.type_name () + ";");
     println (info.internal_prefix + "copy_attributes (" + result + ");");
     println ("return " + result + ";");

     indent --;
     println ("}");
     println ("");
  }

  private void copy_variable (String result,
                              String name,
                              int count,
                              boolean fixed)
  {
     if (fixed)
     {
        for (int i = 1; i <= count; i++)
        {
           String field = name + '_' + i;
           println (result + "->" + field + " = " + field + ";");
        }
     }
     else
     {
        String inx = info.variable_prefix + "i";
        println ("for (int " + inx + " = 0; " + inx + " < " + count + "; " + inx + "++)");
        indent ++;
        println (result + "->" + name + " [" + inx + "] = " + name + " [" + inx + "];");
        indent --;
     }
  }

  private void send_copy_attributes_method (StructInfo s)
  {
     String result = info.param_prefix + "result";

     println ("void " +
              s.type_name () + "::" +
              info.internal_prefix + "copy_attributes" +
              " (" + s.type_name () + " * " + result + ")");

     println ("{");
     indent ++;

     boolean any = false;

     // super class
     if (s.from != null)
     {
        any = true;
        println (s.from.type_name () + "::" +
                 info.internal_prefix + "copy_attributes (" + result + ");");
     }

     // fields
     Enumeration fields = s.getFields ();
     while (fields.hasMoreElements ())
     {
        FieldInfo f = (FieldInfo) fields.nextElement ();
        String n = info.field_prefix + f.ident;

        if (! f.is_const)
        {
           any = true;

           if (f.type instanceof StructInfo)
           {
              println ("if (" +  n + " != NULL)");
              indent ++;
              println (result + "->"  + n + " = " +
                       n + "->" + info.internal_prefix + "copy" + " ();");
              indent --;
           }
           else
           {
              println (result + "->"  + n + " = " + n +  ";");
           }
        }
     }

     // places
     Enumeration places = s.getPlaces ();
     while (places.hasMoreElements ())
     {
        FieldInfo f = (FieldInfo) places.nextElement ();
        String n = info.internal_prefix + f.ident;
        println (result + "->"  + n + " = " + n +  ";");
     }

     // sequences
     int cnt = 0;
     Enumeration sequences = s.getSequences ();
     while (sequences.hasMoreElements ())
     {
        SequenceInfo t = (SequenceInfo) sequences.nextElement ();
        cnt ++;
        if (any)
           println ("");
        any = true;

        String subitem = info.variable_prefix + "subitem";
        if (cnt > 1) subitem = subitem + cnt;

        println (t.element.inner_decl() + " " + subitem +
                 " = " + info.list_prefix + "first" + t.suffix + ";");

        println ("while (" + subitem + " != NULL)");
        println ("{");
        indent ++;

        println (result + "->" +
                 info.list_prefix + "insertLast" + t.suffix + " (" +
                 subitem + "->" + info.internal_prefix + "copy ()" +
                 ");");

        println (subitem + " = " + subitem + "->" + info.list_prefix + "next" + t.suffix + ";");

        indent --;
        println ("}");
     }

     if (s.note_count > 0)
     {
        if (any);
           println ("");

        if (info.locations)
           copy_variable (result, info.internal_prefix + "location",
                          s.note_count, info.fixed_notes);

        if (info.notes)
           copy_variable (result, info.internal_prefix + "start_note",
                          s.note_count, info.fixed_notes);

        if (info.notes && ! info.only_start_notes)
           copy_variable (result, info.internal_prefix + "stop_note",
                          s.note_count, info.fixed_notes);
     }

     indent --;
     println ("}");
     println ("");
  }

  private void send_rename_method (StructInfo s)
  {
     String result = info.param_prefix + "result";

     println ("void " +
              s.type_name () + "::" +
              info.internal_prefix + "rename" +
              " ( )");

     println ("{");
     indent ++;

     boolean any = false;

     // super class
     if (s.from != null)
     {
        any = true;
        println (s.from.type_name () + "::" +
                 info.internal_prefix + "rename" + " ();");
     }

     // fields
     Enumeration fields = s.getFields ();
     while (fields.hasMoreElements ())
     {
        FieldInfo f = (FieldInfo) fields.nextElement ();

        if (f.type == info.ident_type && ! f.is_const)
        {
           any = true;
           println ( info.internal_prefix + "ident_rename" + " (" + f.ident + ");");
        }
     }

     // sequences
     int cnt = 0;
     Enumeration sequences = s.getSequences ();
     while (sequences.hasMoreElements ())
     {
        SequenceInfo t = (SequenceInfo) sequences.nextElement ();
        cnt ++;
        if (any)
           println ("");
        any = true;

        String subitem = info.variable_prefix + "subitem";
        if (cnt > 1) subitem = subitem + cnt;

        println (t.element.inner_decl() + " " + subitem +
                 " = " + info.list_prefix + "first" + t.suffix + ";");

        println ("while (" + subitem + " != NULL)");
        println ("{");
        indent ++;

        println (subitem + "->" + info.internal_prefix + "rename" + " ();" );

        println (subitem + " = " + subitem + "->" + info.list_prefix + "next" + t.suffix + ";");

        indent --;
        println ("}");
     }

     indent --;
     println ("}");
     println ("");
  }

  private void send_methods (StructInfo s)
  {
     // add methods
     Enumeration sequences = s.getSequences ();
     while (sequences.hasMoreElements ())
     {
        SequenceInfo t = (SequenceInfo) sequences.nextElement ();
        send_insert_first_method (t);
        send_insert_last_method (t);
        send_insert_before_method (t);
        send_insert_after_method (t);
        send_remove_method (t);
     }

     // getFirstLocation method
     if (s.get_first_loc)
        send_get_first_location_method (s);

     if (info.copy_method)
     {
        send_copy_method (s);
        send_copy_attributes_method (s);
     }

     if (info.rename_method)
        send_rename_method (s);

     // constructor
     send_constructor (s);

     // constructor with parameters
     if (s.create_func)
        send_constructor_with_params (s);

     // destructor
     send_destructor (s);
  }

  public void sendCode ()
  {
     if (info.copy_method)
        send_basic_copy_method ();

     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof StructInfo)
        {
           StructInfo s = (StructInfo) t;
           send_methods (s);
        }
     }
  }

}
