
import java.util.Enumeration;

class GrammarOutput extends Tool
{

  /* location and notes */

  private void processLocation (int inx, String target_id)
  {
     if (inx != 0 && info.locations)
     {
        if (empty (target_id))
           target_id = rule.variable_name +
                       "->" +
                       info.LocationName (inx);

        printCode ();
        println (target_id + " = " +
                 product_ns + info.internal_prefix + "getLocation ();");
     }
  }

  private void processNote (int inx, String target_id, boolean start)
  {
     if (inx != 0 && info.notes && (start || ! info.only_start_notes))
     {
        String txt, proc;

        if (start)
        {
           txt = info.StartNoteName (inx);
           proc = "getStartNote";
        }
        else
        {
           txt = info.StopNoteName (inx);
           proc = "getStopNote";
        }

        if (empty (target_id))
           target_id = rule.variable_name + "->" + txt;

        printCode ();
        println (target_id + " = " +
                 product_ns + info.internal_prefix + proc + " ();");
     }
  }

  private void location_and_start_note (BasicNode item,
                                        String location_id,
                                        String start_note_id)
  {
     processLocation (item.note_index, location_id);
     processNote (item.note_index, start_note_id, true);
     info.start_note_dupl_cnt ++;
  }

  private void stop_note (BasicNode item,
                          String stop_note_id)
  {
     processNote (item.note_index, stop_note_id, false);
     info.stop_note_dupl_cnt ++;
  }

  private void store_location_and_notes (BasicNode item,
                                         String result_id,
                                         String location_id,
                                         String start_note_id,
                                         String stop_note_id)
  {
     if (info.notes || info.locations)
     {
        printCode ();

        if (info.locations)
           println (result_id + "->" +
                    info.LocationName (item.note_index) + " = " +
                    location_id + ";");

        if (info.notes)
           println (result_id + "->" +
                    info.StartNoteName (item.note_index) + " = " +
                    start_note_id + ";");

        if (info.notes && ! info.only_start_notes)
           println (result_id + "->" +
                    info.StopNoteName (item.note_index) + " = " +
                    stop_note_id + ";");
     }
  }

  /* rules */

  public void openRule (String name)
  {
    super.openRule (name);
    printGrammar ();
    println (info.rule_prefix + name);
    indent ++;
  }

  private void ruleVariables ()
  {
    Enumeration seq = rule.getVariables ();
    if (seq.hasMoreElements ())
    {
       // without printCode
       println ("{");
       indent ++;

       while (seq.hasMoreElements ())
       {
          VarInfo item = (VarInfo) seq.nextElement ();
          println (item.type + " " + item.ident + ";");
       }

       indent --;
       println ("}");
    }
  }

  public void ruleParam (int mode,
                         String type_name)
  {
    super.ruleParam (mode, type_name);

    printGrammar ();

    if (rule.store_type != null)
    {
       println ("[ " + product_ns + rule.store_type.type_name () + " * " +
                info.variable_prefix + "store ]");
    }

    if (mode != MODIFY_MODE)
       print ("returns ");

    println ("[ " + product_ns + info.type_prefix + type_name + " * " +
             info.variable_prefix + "result ]");
    ruleVariables ();

    printGrammar ();
    println (":");

    if (mode == NEW_MODE)
    {
       indent ++;
       printCode ();
       println (info.variable_prefix + "result = " +
                "new " + product_ns + info.type_prefix + type_name + " ();");
       indent --;
    }

    if (rule.store_type != null)
    {
       indent ++;
       printCode ();
       println (info.variable_prefix + "result->" +
                info.field_prefix + rule.store_name + " = " +
                info.variable_prefix + "store;");
       indent --;
    }
  }

  public void closeRule ()
  {
    super.closeRule ();

    if (rule.struct_type != null &&
       (rule.struct_type.custom_subst || rule.struct_type.custom_replace))
    {
       printCode ();
       println ("if (current_stencil != NULL)");
       indent ++;
       println ("result = result->" + info.internal_prefix + "replace (this);");
       indent --;
    }
    else if (rule.type == info.ident_type && ! empty (info.replace_type))
    {
       printCode ();
       println ("if (current_stencil != NULL)");
       indent ++;
       println ("result = " + info.internal_prefix + "ident_replace (result);");
       indent --;
    }

    printGrammar ();
    println (";");
    indent --;
    println ("");
  }

  /* binary */

  private String binary_branch;
  private int binary_item_count;

  public void binaryBranch (String branch)
  {
     super.binaryBranch (branch);

     printGrammar ();
     println ("returns [ " + product_ns + info.type_prefix + info.expr_type + " * " +
              info.variable_prefix + "result ]");
     ruleVariables ();

     printGrammar ();
     println (":");
     indent ++;
     println (info.variable_prefix + "result=" + info.rule_prefix + branch);

     println ("(");
     binary_branch = branch;
     binary_item_count = 0;
  }

  public void binaryItem (String key, String selector_value,
                          String middle_branch, String middle_key)
  {
     super.binaryItem (key, selector_value, middle_branch, middle_key);

     binary_item_count ++;
     if (binary_item_count != 1)
     {
        printGrammar ();
        println ("|");
     }
     indent ++;

     location_and_start_note (info.binary_node,
                              info.temp_prefix + "location", info.temp_prefix + "start_note");
     put_terminal (key);
     println ("");
     stop_note (info.binary_node, info.temp_prefix + "stop_note");

     String v = info.temp_prefix + "branch";
     String t = info.temp_prefix + "right";

     printCode ();
     println (v + " = new " + product_ns + info.type_prefix + info.binary_expr_type + " ();");
     println (v + "->" + info.field_prefix + info.binary_expr_selector + " = " +
              product_ns + selector_value + ";");
     println (v + "->" + info.field_prefix + info.binary_expr_left +
              " = " + info.variable_prefix + "result;");

     if (middle_branch != "")
     {
        String m = info.temp_prefix + "middle";

        printGrammar ();
        println (m + " = " +  info.rule_prefix + middle_branch + " ");

        printCode ();
        println (v + "->" + info.field_prefix + info.binary_expr_middle + " = " + m + ";");

        location_and_start_note (info.binary_node.another,
                                 info.temp_prefix + "another_location", info.temp_prefix + "another_start_note");
        put_terminal (middle_key);
        println ("");
        stop_note (info.binary_node.another, info.temp_prefix + "another_stop_note");
     }

     printGrammar ();
     println (t + " = " +  info.rule_prefix + binary_branch + " ");

     printCode ();
     println (v + "->" + info.field_prefix + info.binary_expr_right + " = " + t + ";");

     println (info.variable_prefix + "result = " + v + ";");
     store_location_and_notes (info.binary_node,
                               info.variable_prefix + "result",
                               info.temp_prefix + "location",
                               info.temp_prefix + "start_note",
                               info.temp_prefix + "stop_note");

     indent --;
  }

  public void closeBinary ()
  {
     super.closeBinary ();
     printGrammar ();
     println (")*");
     indent --;
  }

  /* unary */

  public void unary (String key, String selector_value, String branch)
  {
     super.unary (key, selector_value, branch);

     printGrammar ();
     println ("returns [ " + product_ns + info.type_prefix + info.unary_expr_type + " * " +
              info.variable_prefix + "result ]");
     ruleVariables ();

     printGrammar ();
     println (":");
     indent ++;
     printCode ();
     println (info.variable_prefix + "result = " +
              "new " + product_ns + info.type_prefix + info.unary_expr_type + " ();");

     location_and_start_note (info.unary_node, "", "");
     put_terminal (key);
     println ("");
     stop_note (info.unary_node, "");

     printCode ();
     println (info.variable_prefix + "result->" + info.field_prefix + info.unary_expr_selector + " = " +
              product_ns + selector_value + ";");

     printGrammar ();
     println (info.temp_prefix + "param=" + info.rule_prefix + branch + " ");

     printCode ();
     println (info.variable_prefix + "result->" + info.field_prefix + info.unary_expr_param +
              "=" + info.temp_prefix + "param;");

     indent --;
  }

  /* primitive */

  private int primitive_item_count;

  public void openPrimitive ()
  {
     super.openPrimitive ();

     printGrammar ();
     println ("returns [ string " + info.variable_prefix + "result ]");
     println (":");

     primitive_item_count = 0;
  }

  public void primitiveItem (String branch)
  {
     super.primitiveItem (branch);

     printGrammar ();

     primitive_item_count ++;
     if (primitive_item_count != 1) print ("| ");
     indent ++;

     String id = info.temp_prefix + primitive_item_count;
     println (id + ":" + branch);
     printCode ();
     println (info.variable_prefix + "result=" + id + "->getText();");

     indent --;
  }

  /* block */

  public void openBlock ()
  {
    super.openBlock ();

    block.parenthesis =
       block.above != null ||
       rule.mode == NEW_MODE && block.alt_count > 1 ||
       (rule.struct_type != null && (rule.struct_type.custom_subst || rule.struct_type.custom_replace)) ||
       (rule.type == info.ident_type && ! empty (info.replace_type));

    if (block.parenthesis)
    {
       printGrammar ();
       if (inside_predicate ())
          print ("( ");
       else
          println ("(");
    }
  }

  public void closeBlock ()
  {
    super.closeBlock ();

    if (block.parenthesis)
    {
       printGrammar ();
       print (")");
       if (block.mark != ' ' && block.mark != '>')
          print (""+block.mark);
       if (! inside_predicate ())
          println ("");
    }
  }

  /* alternative */

  public void openAlternative ()
  {
    super.openAlternative ();

    printGrammar ();
    if (alt.index != 1) println ("|");
    indent ++;
  }

  public void closeAlternative ()
  {
    super.closeAlternative ();
    indent --;
  }

  /* terminal */

  private void put_terminal (String value)
  {
    printGrammar ();
    if (isIdentifier (value))
       print ('\"' + value + '\"');
    else
       print (info.findLiteral (value));
  }

  public void terminal (String value)
  {
     super.terminal (value);
     if (inside_predicate ())
     {
        put_terminal (value);
        print (" ");
     }
     else
     {
        location_and_start_note (current_terminal, "", "");
        put_terminal (value);
        println ("");
        stop_note (current_terminal, "");
     }
  }

  /* nonterminal */

  public void nonterminal (int kind, String label, String name)
  {
     super.nonterminal (kind, label, name);

     if (inside_predicate ())
     {
        printGrammar ();
        print (info.rule_prefix + name + " ");
     }
     else
     {
        RuleNode r = info.findRule (name); // sub-rule
        TypeInfo t = r.type;
        String prefix = info.rule_prefix;

        if (r.mode == PRIMITIVE_MODE)
           location_and_start_note (current, "", "");

        if (kind == LABELED)
        {
           String v = current.variable;
           printGrammar ();
           println (v + "=" + prefix + name);
           printCode ();
           println (rule.variable_name + "->" +
                    info.field_prefix + label + "=" + v +";");
        }
        else if (kind == INSERT || kind == ADD)
        {
           String v = current.variable;
           boolean list_repl = current.sequence.element.custom_list_replace;
           String list_txt = list_repl ? "_list" : "";
           printGrammar ();
           println (v + "=" + prefix + name);
           printCode ();
           println (rule.variable_name +
                    "->" + info.list_prefix + "add" + list_txt + current.sequence.suffix +
                    " (" + v + ");");
        }
        else if (kind == CALL)
        {
           printGrammar ();
           println (prefix + name + " [ " + rule.variable_name + " ]");
        }
        else if (kind == RESULT || kind == SELECT)
        {
           printGrammar ();
           println (rule.variable_name + "=" + prefix + name);
        }
        else if (kind == CHOOSE1)
        {
           printGrammar ();
           if (rule.choose_compatible_type)
              println (info.variable_prefix + "result=" + prefix + name);
           else
              println (info.variable_prefix + "tmp=" + prefix + name);
        }
        else if (kind == CHOOSE2)
        {
           printGrammar ();
           if (rule.choose_compatible_type)
              println (info.variable_prefix + "result=" + prefix + name +
                       "[" + info.variable_prefix + "result]");
           else
              println (info.variable_prefix + "result=" + prefix + name +
                       "[" + info.variable_prefix + "tmp]");
        }
        else error ("Internal error");

        if (r.mode == PRIMITIVE_MODE)
           stop_note (current, "");
     }
  }

  /* semantic predicate */

  public void semanticPredicate (String name, boolean no_params)
  {
     super.semanticPredicate (name, no_params);

     printGrammar ();
     if (no_params)
        println ("{ " + name + "() }?");
     else
        println ("{ " + name + " }?");
  }

  /* syntactic predicate */

  public void openPredicate ()
  {
     super.openPredicate ();
  }

  public void closePredicate ()
  {
     super.closePredicate ();
     printGrammar ();
     println (" =>");
  }

  /* set option */

  public void setOption (String name, String value)
  {
     super.setOption (name, value);

     String ns = "";
     if (! value.equals ("true") && ! value.equals ("false"))
        ns = product_ns;

     printCode ();
     println (rule.variable_name + "->" + info.field_prefix + name + " = " +
              ns + value + ";");
  }

  /* inc option */

  public void incOption (String name)
  {
     super.incOption (name);

     printCode ();
     println (rule.variable_name + "->" + info.field_prefix + name + " ++;");
  }

  /* silent option */

  public void silentOption ()
  {
     super.silentOption ();
     printGrammar ();
     println ("options { warnWhenFollowAmbig = false; }:");
  }

  /* invoke option */

  public void invokeOption (String name)
  {
     super.invokeOption (name);

     printCode ();
     println (rule.variable_name + "->" + name + " (this);");
  }

  /* execute option */

  public void executeOption (String name, boolean no_params)
  {
     super.executeOption (name, no_params);

     if (! info.ignore_execute_options)
     {
        printCode ();
        if (no_params)
           println (name + " ();");
        else
           println (name + " (" + rule.variable_name + ");");
     }
  }

  /* redudant option */

  public void redudantOption ()
  {
     super.redudantOption ();
  }

  /* send list of literals */

  public void sendLiterals ()
  {
     Enumeration seq = info.getLiterals ();
     while (seq.hasMoreElements ())
     {
        String s;
        LiteralItem item = (LiteralItem) seq.nextElement ();

        if (item.text.length () == 1)
           s = '\'' + item.text + '\'';
        else
           s = '\"' + item.text + '\"';

        println (item.ident);
        println ("  options { paraphrase=\"\'" + item.text + "\'\"; }");
        println ("  : " + s + " ;");
        println ("");
     }
  }

  /* send list of tokens */

  public void sendTokens ()
  {
     Enumeration seq = info.getTokens ();

     if (seq.hasMoreElements ())
     {
        println ("tokens");
        println ("{");
        indent ++;

        while (seq.hasMoreElements ())
        {
           DictItem item = (DictItem) seq.nextElement ();
           if (! info.isExternalToken (item.ident))
              println ('\"' + item.ident + '\"' + ";");
        }

        indent --;
        println ("}");
        println ("");
     }
  }

}
