
class ListOutput extends Tool
{

  private String cast (String type, String expr)
  {
    // return "dynamic_cast <" + type + "*> (" + expr + ")";
    return expr + "->" + info.conv_func_prefix + type + " ()";
  }

  private String is_type (String type, String expr)
  {
    return cast (type, expr) + " != NULL";
  }

  private void check (String expr)
  {
    println ("assert (" + expr + ");");
  }

  private void check (String expr, String comment)
  {
    println ("assert (" + expr + "); // " + comment);
  }

  private void sendLocation (int inx, String item_id)
  {
     if (inx != 0 && info.locations)
     {
        println (info.internal_prefix + "sendLocation (" + item_id + "->" +
                 info.LocationName (inx) + ");" );
     }
  }

  private void sendNote (int inx, String item_id, boolean start)
  {
     if (inx != 0 && info.notes && (start || ! info.only_start_notes))
     {
        String txt, proc;

        if (start)
        {
           txt = info.StartNoteName (inx);
           proc = "sendStartNote";
        }
        else
        {
           txt = info.StopNoteName (inx);
           proc = "sendStopNote";
        }

        println (info.internal_prefix + proc +
                 " (" + item_id + "->" + txt + ");" );
     }
  }

  private void location_and_start_note (BasicNode item, String item_id)
  {
     sendNote (item.note_index, item_id, true);
     sendLocation (item.note_index, item_id);
     info.start_note_list_cnt ++;
  }

  private void stop_note (BasicNode item, String item_id)
  {
     sendNote (item.note_index, item_id, false);
     info.stop_note_list_cnt ++;
  }

  /* rules */

  public void openRule (String name)
  {
     super.openRule (name);

     String item = info.param_prefix + "item";

     if (rule.custom_send || rule.custom_no_send) println ("/*");

     println ("void " +
              info.product_class + "::" +
              (rule.custom_internal_send ? info.send_internal_prefix : info.send_func_prefix) +
              name +
              " (" + rule.type.inner_decl() + " " + item + ")");
     println ("{");
     indent ++;

     if (rule.struct_type != null)
        check (item + " != NULL");

     if (rule.custom_open_close_send)
        println (info.open_func_prefix + name + " (" + item + ");");

     if (rule.loop)
     {
        println (rule.loop_ref.element.inner_decl () + " " +
                 info.param_prefix + "subitem" +
                 " = " + info.param_prefix + "item" +
                 "->" + info.list_prefix + "first" + rule.loop_ref.suffix + ";");
     }
  }

  public void ruleParam (int mode,
                         String type_name)
  {
     super.ruleParam (mode, type_name);
  }

  public void closeRule ()
  {
     super.closeRule ();

     if (rule.loop)
        check (info.param_prefix + "subitem" + " == NULL");

     if (rule.custom_open_close_send)
        println (info.close_func_prefix +
                 rule.name +
                 " (" + info.param_prefix + "item);");

     indent --;
     println ("}");
     if (rule.custom_send || rule.custom_no_send) println ("*/");
     println ("");
  }

  /* binary */

  private RuleNode binary_branch;
  private int      binary_count;

  public void binaryBranch (String branch)
  {
     super.binaryBranch (branch);
     binary_branch = info.findRule (branch);
     binary_count = 0;

     if (info.expr_method)
     {
         String func = info.send_func_prefix + info.binary_expr_func;
         String item  = info.param_prefix + "item";
         println (func + " (" + item + ");");
     }
  }

  public void binaryItem (String key, String selector_value,
                          String middle_branch, String middle_key)
  {
     super.binaryItem (key, selector_value, middle_branch, middle_key);

     if (! info.expr_method)
     {
        binary_count ++;
        if (binary_count > 1) print ("else ");

        String item  = info.param_prefix + "item";
        String item1 = info.param_prefix + "item1";

        println ("if (" + item + "->" + info.internal_prefix + "isBinary (" + selector_value + "))");
        println ("{");
        indent ++;

        println (info.type_prefix + info.binary_expr_type + " * " + item1 + " = " +
                 cast (info.type_prefix + info.binary_expr_type, item) + ";" );

        check (item1 + " != NULL");
        put (rule, rule.type, item1 + "->" + info.field_prefix + info.binary_expr_left);

        location_and_start_note (info.binary_node, item1);
        put_terminal (key);
        stop_note (info.binary_node, item1);

        if (! empty (middle_branch))
        {
           RuleNode b = info.findRule (middle_branch);
           put (b, rule.type, item1 + "->" + info.field_prefix + info.binary_expr_middle);

           location_and_start_note (info.binary_node.another, item1);
           put_terminal (middle_key);
           stop_note (info.binary_node.another, item1);
        }

        put (binary_branch, rule.type, item1 + "->" + info.field_prefix + info.binary_expr_right);

        indent --;
        println ("}");
     }
  }

  public void closeBinary ()
  {
     super.closeBinary ();

     if (! info.expr_method)
     {
        print ("else ");
        put (binary_branch, rule.type, info.param_prefix + "item");
     }
  }

  private String QuoteStr (String s) { return "\"" + s + "\""; }

  public void binary_support () // directly called from Main.java
  {
     if (info.expr_method)
     {
        String func = info.send_func_prefix + info.binary_expr_func;
        String item  = info.param_prefix + "item";
        String item1 = info.param_prefix + "item1";
        String item2 = info.param_prefix + "item2";

        StructInfo expr_type = info.findStruct (info.expr_type);
        StructInfo bin_type  = info.findStruct (info.binary_expr_type);

        RuleNode sub_rule = null;
        RuleNode r = info.first_rule;
        while (r != null)
        {
           if (r.mode == BINARY_MODE && r.binary_level == 1 && sub_rule == null)
              sub_rule = r;
           r = r.next_rule;
        }
        if (sub_rule == null)
           error ("Cannot find basic binary expression");

        // function declaration

        println ("void " + info.product_class + "::" + func +
                 " (" + info.type_prefix + info.expr_type + " * " + item + ")");
        println ("{");
        indent ++;

        println (info.type_prefix + info.binary_expr_type + " * " + item1 + " = " +
                 cast (info.type_prefix + info.binary_expr_type, item) + ";" );

        println ("if (" + item1 + " != NULL)");
        println ("{");
        indent ++;

        // println ("send (" + QuoteStr ("/*(*/") + ");");
        println (func + " (" + item1 + "->" + info.field_prefix + info.binary_expr_left + ");");

        // binary op

        println ("switch (" + item1 + "->" +  info.field_prefix + info.binary_expr_selector + ")");
        println ("{");
        indent ++;

        r = info.first_rule;
        while (r != null)
        {
           if (r.mode == BINARY_MODE)
           {
              BinaryNode node = r.first_binary_item;
              while (node != null)
              {
                  println ("case " + node.selector_value + ":");
                  indent ++;

                  location_and_start_note (info.binary_node, item1);
                  put_terminal (node.key);
                  stop_note (info.binary_node, item1);

                  // ternary expression
                  if (! empty (node.middle_branch))
                  {
                     RuleNode b = info.findRule (node.middle_branch);
                     put (b, bin_type, item1 + "->" + info.field_prefix + info.binary_expr_middle);

                     location_and_start_note (info.binary_node.another, item1);
                     put_terminal (node.middle_key);
                     stop_note (info.binary_node.another, item1);
                  }

                  println ("break;");
                  indent --;

                 node = node.next_binary_item;
              }
           }
           r = r.next_rule;
        }
        println ("default:");
        indent ++;
        println ("assert (false);");
        indent --;
        indent --;
        println ("}"); // end of switch

        println (func + " (" + item1 + "->" + info.field_prefix + info.binary_expr_right + ");");
        // println ("send (" + QuoteStr ("/*)*/") + ");");

        indent --;
        println ("}");

        println ("else");
        println ("{");
        indent ++;

        RuleNode s = info.findRule (sub_rule.binary_sub_rule);
        put (s, expr_type, item);

        indent --;
        println ("}"); // end of if

        indent --;
        println ("}"); // end of function
        println ("");
     }
  }

  /* unary */

  public void unary (String key, String selector_value, String branch)
  {
     super.unary (key, selector_value, branch);

     String item  = info.param_prefix + "item";

     check (item + "->" + info.field_prefix + info.unary_expr_selector + " == " + selector_value);

     location_and_start_note (info.unary_node, item);
     put_terminal (key);
     stop_note (info.unary_node, item);

     RuleNode r = info.findRule (branch);
     put (r, rule.type, item + "->" + info.field_prefix + info.unary_expr_param);
  }

  /* primitive */

  public void openPrimitive ()
  {
     super.openPrimitive ();
     println (info.internal_prefix + "send (" + info.param_prefix + "item);");
  }

  /* block */

  public void openBlock ()
  {
     super.openBlock ();

     if (! inside_predicate () && ! inside_redudant ())
     {
        if (! is_select_block () && ! is_choose_block ())
           if (block.mark == '+' || block.mark == '*')
              if (block.loop)
              {
                 if (block.alt_count > 1) error ("Too complicated loop");
                 println ("while (" + info.param_prefix + "subitem" + " != NULL)");
                 block.braces = true;
                 block.subitem_loop = true; /* no condition for inner alternative */
              }

        if (block.braces)
        {
           println ("{");
           indent ++;
        }
     }
  }

  public void closeBlock ()
  {
     super.closeBlock ();

     if (! inside_predicate () && ! inside_redudant ())
     {
        if (block.braces)
        {
           indent --;
           println ("}");
        }

        if (is_select_block ())
        {
           println ("else assert (false);");
        }
        else if (is_choose_block ())
        {
           if (block.mark != ' ')
           {
              print ("else ");
              put (rule.keep_rule,
                   rule.type,
                   info.param_prefix + "item");
           }
        }
        else if (block.mark == ' ' && block.alt_count != 1)
        {
           println ("else assert (false);");
        }

     }
  }

  /* alternative */

  private void condition ()
  {
     int count = 0;

     int len = alt.ordKeys.size ();
     for (int i=0; i<len; i++)
     {
        DictItem2 item = (DictItem2) alt.ordKeys.get (i);
        count ++;
        if (count > 1) print (" && ");
        print (info.param_prefix + "item" + "->" + info.field_prefix + item.ident + " == " + item.value);
     }

     len = alt.incKeys.size ();
     for (int i=0; i<len; i++)
     {
        String id = (String) alt.incKeys.get (i);
        count ++;
        if (count > 1) print (" && ");
        print (info.param_prefix + "item" + "->" + info.field_prefix + id + " != 0");
     }

     len = alt.refKeys.size ();
     for (int i=0; i<len; i++)
     {
        String id = (String) alt.refKeys.get (i);
        count ++;
        if (count > 1) print (" && ");
        print (info.param_prefix + "item" + "->" + info.field_prefix + id + " != NULL");
     }

     len = alt.strKeys.size ();
     for (int i=0; i<len; i++)
     {
        String id = (String) alt.strKeys.get (i);
        count ++;
        if (count > 1) print (" && ");
        print ("! " + info.param_prefix + "item" + "->" + info.field_prefix + id + ".empty ()");
     }

     if (count == 0)
        if (alt.index == block.alt_count && block.mark == ' ')
        {
           print ("true"); // !?
        }
        else
        {
           if (! info.anonymous_fields) // !?
              error ("No condition for this alternative");
        }
  }

  public void openAlternative ()
  {
     super.openAlternative ();

     if (! inside_predicate () && ! inside_redudant ())
     {
        if (! block.subitem_loop)
           if (block.mark != ' ' || block.alt_count != 1)
           {
              if (alt.index != 1)
                 if (is_select_block () ||
                     is_choose_block () ||
                     block.mark == ' ' ||
                     block.mark == '?') // !? !!
                 {
                    print ("else ");
                 }

              print ("if (");

              String item  = info.param_prefix + "item";

              if (is_select_block ())
              {
                 if (alt.selection_rule.mode == UNARY_MODE)
                    print (item + "->" + info.internal_prefix + "isUnary" +
                           " (" +  alt.selection_rule.unary_selector + ")"); // !?
                 else
                    print (is_type (alt.selection_rule.type.type_name (), item));
              }
              else if (is_choose_block ())
              {
                 print (is_type (alt.selection_rule.type.type_name (), item));
              }
              else if (alt.loop)
                 print (info.param_prefix + "subitem" + " != NULL");
              else
                 condition ();

              println (")");

              alt.braces = true;
           }

        if (alt.braces)
        {
           println ("{");
           indent ++;
        }

        for (int inx = 0 ; inx < alt.incKeys.size (); inx ++)
        {
           String id = (String) alt.incKeys.get (inx);
           String item  = info.param_prefix + "item";
           String counter  = info.param_prefix + "counter_" + id;
           println ("for (int " + counter + " = 0; "+
                     counter + " < " + item + "->" + info.field_prefix + id + "; " +
                     counter + "++)");
           println ("{");
           indent ++;
        }
     }
  }

  public void closeAlternative ()
  {
     super.closeAlternative ();

     if (! inside_predicate () && ! inside_redudant ())
     {
        if (alt.braces)
        {
           indent --;
           println ("}");
        }

        for (int inx = 0 ; inx < alt.incKeys.size (); inx ++)
        {
           indent --;
           println ("}");
        }
     }
  }

  public void additionalOption (String name)
  {
     super.additionalOption (name);

     RuleNode subrule = info.findRule (name);
     println ("// additional");
     println ("else if (" +
              is_type (subrule.type.type_name (), info.param_prefix + "item") +
              ")");
     println ("{");
     indent ++;
     put (subrule, rule.type, info.param_prefix + "item");
     indent --;
     println ("}");
  }

  public void redudantOption ()
  {
     super.redudantOption ();
  }

  /* terminals */

  private void put_terminal (String value)
  {
     println (info.internal_prefix + "send (" + '\"' + value + '\"' +  ");");
  }

  public void terminal (String name)
  {
     super.terminal (name);
     if (! inside_predicate () && ! inside_redudant ())
     {
        location_and_start_note (current_terminal, info.param_prefix + "item");
        put_terminal (name);
        stop_note (current_terminal, info.param_prefix + "item");
     }
  }

  private void put (RuleNode subrule,
                    TypeInfo expr_type,
                    String expr)
  {
     if (subrule.custom_no_send)
     {
        check ("false", "no_send");
     }
     else
     {
        String txt = expr;
        if (! expr_type.is_derived_from (subrule.type))
        {
           txt = cast (subrule.type.type_name (), txt);
        }
        println (info.send_func_prefix + subrule.name + " (" + txt + ");");
     }
  }

  /* nonterminals */

  public void nonterminal (int kind, String label, String name)
  {
     super.nonterminal (kind, label, name);

     if (! inside_predicate () && ! inside_redudant ())
     {
        RuleNode r = info.findRule (name); // sub-rule

        // comments and source position
        if (r.mode == PRIMITIVE_MODE)
           location_and_start_note (current, info.param_prefix + "item");

        if (kind == CHOOSE1)
        {
           // already done by Decl
           // rule.keep_rule = r;
        }
        else if (kind == LABELED)
        {
           put (r, current.field.type,
                   info.param_prefix + "item" + "->" + info.field_prefix + label);
        }
        else if (kind == INSERT || kind == ADD)
        {
           String id = info.param_prefix + "subitem";
           put (r, current.sequence.element, id);
           println (id + " = " + id + "->" +
                    info.list_prefix + "next" + current.sequence.suffix + ";");
        }
        else if (kind == CALL || kind == RESULT || kind == SELECT)
        {
           put (r, rule.type, info.param_prefix + "item");
        }
        else if (kind == CHOOSE2)
        {
           String type_name = r.type.type_name ();
           String item = info.param_prefix + "item";
           String item1 = info.param_prefix + "item1";
           String stored_field = info.field_prefix + r.store_name;

           println (type_name + " * " + item1 + " = " +
                    cast (type_name, item) + ";" );

           if (block.mark == ' ')
              check (item1 + " != NULL");

           if (block.mark == '+' || block.mark == '*')
              put (rule,
                   r.store_type,
                   item1 + "->" + stored_field);
           else
              put (rule.keep_rule,
                   r.store_type,
                   item1 + "->" + stored_field);

           put (r, r.type, item1);
        }
        else error ("Internal error");

        // comments and source position
        if (r.mode == PRIMITIVE_MODE)
           stop_note (current, info.param_prefix + "item");
     }
  }

  /* options */

  public void setOption (String name, String value)
  {
     super.setOption (name, value);
     check (info.param_prefix + "item" + "->" + info.field_prefix + name + " == " + value);
  }

  public void incOption (String name)
  {
     super.incOption (name);
     check (info.param_prefix + "item" + "->" + info.field_prefix + name + " != 0");
  }

  public void styleOption (int style)
  {
     super.styleOption (style);

     switch (style)
     {
        case SPACE:           println (info.util_func_prefix + "style_space ();");           break;
        case NO_SPACE:        println (info.util_func_prefix + "style_no_space ();");        break;
        case NEW_LINE:        println (info.util_func_prefix + "style_new_line ();");        break;
        case EMPTY_LINE:      println (info.util_func_prefix + "style_empty_line ();");      break;
        case NO_EMPTY_LINE:   println (info.util_func_prefix + "style_no_empty_line ();");   break;
        case INDENT:          println (info.util_func_prefix + "style_indent ();");          break;
        case UNINDENT:        println (info.util_func_prefix + "style_unindent ();");        break;
     }
  }

}
