
import java.util.Enumeration;
import java.util.TreeMap;

class ParserInfo extends Common
{
  /* Configuration */

  public String common_header = "";  // output file name prefix for common header (.hpp)
  public String common_name = "";  // output file name prefix for common code (.cpp)
  public String public_header = "";  // output file name prefix for public header (.hpp)
  public String private_header = ""; // output file name prefix for private header (.hpp)

  public String symbol_header = ""; // output file name prefix for header with symbol enumeration (.hpp)

  public String lexer_name;  // output file name prefix (.gpp, .hpp, .cpp)
  public String parser_name; // output file name prefix (.gpp, .hpp, .cpp)

  public String product_name = ""; // output file name prefix (.hpp, .cpp)

  public String lexer_input; // file with main lexer code
  public String product_input; // file with additional writer declarations

  public String product_namespace = ""; // namespace identifier for generated code
  public boolean automatic_namespace = false; // use OPEN_NAMESPACE and CLOSE_NAMESPACE macros

  public String lexer_class = ""; // identifier for generated lexer class
  public String lexer_super_class = ""; // lexer base class

  public String parser_class = ""; // identifier for generated parser class
  public String parser_super_class = ""; // parser base class

  public String product_class = ""; // identifier for generated class
  public String product_super_class = ""; // product base class

  public boolean tree_code = false; // enable / disable 'produce_...' methods

  public boolean expr_method = false; // true => faster expression processing
  public boolean copy_method = false; // true => 'copy' methods
  public boolean rename_method = false; // true => 'rename' methods

  public boolean notes = false; // true => store spaces and comments
  public boolean only_start_notes = false; // do not store comments after tokens

  public boolean locations = false; // true => store line numbers

  public boolean ignore_notes = false; // true => do not store notes, only declare data structures
  public boolean ignore_locations = false; // true => do not store locations, only declare data structures

  public String  note_type = "string"; // type identifier for comments
  public String  location_type = "int"; // type identifier for locations
  public String  mark_type = "int"; // type identifier for mark / rewind

  public String  note_init = ""; // initialization value for notes
  public String  location_init = ""; // initialization value for locations

  public boolean anonymous_fields = false; // enable fields without names
  public boolean ignore_send_options = false;
  public boolean ignore_output_options = false;
  public boolean ignore_execute_options = false;
  public boolean ignore_product_constructor = false; // true => do not generate product class constructor ( only for gram )
  public boolean ignore_tag_marks = false; // true => ignore '<' separators ( only for grammar.g )

  public boolean antlr_mode = false; // true => ANTLR parser implemntation

  public String subst_type = ""; // class identifier for substitution items
  public String replace_type = ""; // class identifier for substitution items used during parsing

  /* identifier prefixes */

  public String rule_prefix = ""; // rule names (in parser class)
  public String type_prefix = ""; // type names

  public String field_prefix     = ""; // user defined fields (in structures)
  public String list_prefix      = ""; // fields and routines for sequences (in structures)
  public String internal_prefix  = ""; // internal fields and routines (in structures)

  public String send_func_prefix      = "send_";  // send text output functions
  public String send_internal_prefix  = "internal_send_";  // original send text output functions
  public String open_func_prefix      = "open_";  // open functions (<custom open_close_send> )
  public String close_func_prefix     = "close_"; // open functions (<custom open_close_send> )

  public String conv_func_prefix  = "conv_";    // enumeration type conversion functions
                                                // class conversion methods

  public String util_func_prefix  = "";         // support functions

  public String param_prefix      = "p_";       // parameters
  public String temp_prefix       = "tmp_";     // temporary variables
  public String local_prefix      = "local_";   // !? not used, local varibles (assign in nonterminals)
  public String variable_prefix   = "";         // variables: item, item1, subitem, result
  public String literal_prefix    = "LITERAL_"; // literal names

  /* symbol type */

  public String symbol_type       = "SYMBOL";
  public String conv_type         = "int";
  public String no_symbol         = "NONE";

  public String primitive_prefix  = "";
  public String delimiter_prefix  = "";
  public String keyword_prefix    = "LITERAL_";

  /* binary and unary expressions */

  public String basic_type = "_TBasic";
  public String expr_type = "_TExpr";

  public String binary_expr_type     = "_TBinaryExpr";
  public String binary_expr_left     = "_left";
  public String binary_expr_right    = "_right";
  public String binary_expr_middle   = "_middle";
  public String binary_expr_selector = "_selector";
  public String binary_expr_func     = "_binary_expression";

  public String unary_expr_type     = "_TUnaryExpr";
  public String unary_expr_param    = "_param";
  public String unary_expr_selector = "_selector";

  /* variables */

  public String writer_param;
  public String writer_decl;

  public boolean private_rules = false; // place declaration into public or private header

  public String inp_dir = ""; // copy of command line parameter

  /* headers */

  public boolean enable_public_header = false; // enable generation of public header
  public boolean enable_private_header = false; // enable generation of private header

  public boolean enable_lexer_output = false; // enable generation of lexer

  public boolean enable_symbol_header = false; // enable generation of header with symbol enumeration

  /* location and notes */

  public static BasicNode unary_node = new BasicNode ();
  public static BasicNode binary_node = new BasicNode ();

  public int note_decl_cnt;
  public int start_note_dupl_cnt;
  public int start_note_list_cnt;
  public int stop_note_dupl_cnt;
  public int stop_note_list_cnt;

  public /* static final */ boolean fixed_notes = false; // true => variables, not array
  public /* static final */ boolean distinct_notes = false; // dictinct identifiers

  public String LocationName (int inx)
  {
     String txt = internal_prefix + "location";
     if (fixed_notes)
        return txt + '_' + inx;
     else
        return txt + " [" + (inx-1) + "]";
  }

  public String StartNoteName (int inx)
  {
     String txt = internal_prefix + "start_note";

     if (fixed_notes)
        return txt + '_' + inx;
     else
        return txt + " [" + (inx-1) + "]";
  }

  public String StopNoteName (int inx)
  {
     String txt = internal_prefix + "stop_note";

     if (fixed_notes)
        return txt + '_' + inx;
     else
        return txt + " [" + (inx-1) + "]";
  }

  /* grammar options */

  public DictTable grammar_repository = new DictTable ();
  public DictTable lexer_repository = new DictTable ();
  public DictTable parser_repository = new DictTable ();

  public DictTable extension_repository = new DictTable (); // additional declarations

  /* types */

  private DictTable type_list = new DictTable ();

  public TypeInfo getType (String name)
  {
    return (TypeInfo) type_list.get (name);
  }

  public TypeInfo findType (String name)
  {
     TypeInfo t = getType (name);
     if (t == null)
        err ("Undefined type: " + name);
     return t;

  }

  private void createType (TypeInfo type)
  {
    if (getType (type.ident) != null)
       err ("Type already defined: " + type.ident);

    type_list.store (type);
  }

  public Enumeration getTypes ()
  {
    return type_list.elements ();
  }

  /* simple types */

  public final TypeInfo string_type = newType ("string");
  public final TypeInfo bool_type   = newType ("bool");
  public final TypeInfo int_type    = newType ("int");

  public final TypeInfo ident_type = newAlias ("string");
  // string type for identifiers

  private TypeInfo newType (String name)
  {
    TypeInfo t = new TypeInfo ();
    t.ident = name;
    createType (t);
    return t;
  }

  private TypeInfo newAlias (String name)
  {
    TypeInfo t = new TypeInfo ();
    t.ident = name;
    return t;
  }

  public TypeInfo newTypeDef (String name, boolean pointer)
  {
    TypeInfo t = new TypeInfo ();
    t.ident = name;
    t.is_pointer = pointer;
    return t;
  }

  /* enums */

  public EnumInfo getEnum (String name)
  {
    TypeInfo t = getType (name);

    if (t != null && ! (t instanceof EnumInfo))
       err ("Enum type required: " + name);

    return (EnumInfo) t;
  }

  public EnumInfo findEnum (String name)
  {
     EnumInfo t = getEnum (name);
     if (t == null)
        err ("Undefined type: " + name);
     return t;
  }

  public EnumInfo createEnum (String name)
  {
    EnumInfo typ = new EnumInfo ();
    typ.ident = name;
    createType (typ);
    return typ;
  }

  public EnumInfo lookupEnum (String name)
  {
     EnumInfo t = getEnum (name);
     if (t == null)
        t = createEnum (name);
     return t;
  }

  /* basic type methods */

  private DictTable basic_type_methods = new DictTable ();

  public void addBasicTypeMethod (MethodInfo m)
  {
    basic_type_methods.store (m);
  }

  public boolean isBasicTypeMethod (String id)
  {
    return basic_type_methods.contains (id);
  }

  public Enumeration getBasicTypeMethods ()
  {
    return basic_type_methods.elements ();
  }

  /* structures */

  public StructInfo getStruct (String name)
  {
    TypeInfo t = getType (name);

    if (t != null && ! (t instanceof StructInfo))
       err ("Struct type required: " + name);

    return (StructInfo) t;
  }

  public StructInfo findStruct (String name)
  {
     StructInfo t = getStruct (name);
     if (t == null)
        err ("Undefined type: " + name);
     return t;
  }

  public StructInfo createStruct (String name)
  {
    StructInfo t = new StructInfo ();
    t.ident = name;
    createType (t);

    return t;
  }

  public StructInfo lookupStruct (String name)
  {
     StructInfo t = getStruct (name);
     if (t == null)
        t = createStruct (name);
     return t;
  }

  /* literals - delimiters */

  private LiteralTable literal_list  = new LiteralTable ();

  public String getLiteral (String text)
  {
    LiteralItem item = literal_list.searchText (text);
    if (item == null)
       return null;
    else
       return item.ident;
  }

  public String findLiteral (String text)
  {
    String s = getLiteral (text);
    if (empty (s))
       err ("Undefined literal: " + text);
    return s;
  }

  public void storeLiteral (String ident, String text)
  {
    if (empty (ident))
       err ("Literal without name: " + ident);

    if (empty (text))
       err ("Literal without text representation: " + ident);

    if (literal_list.containsIdent (ident))
       err ("Literal identifier already defined: " + ident);

    if (literal_list.containsText (text))
       err ("Literal with same text  representation already defined: " + text);

    LiteralItem item = new LiteralItem ();
    item.ident = ident;
    item.text = text;
    literal_list.store (item);
  }

  public void checkLiteral (String text)
  {
    if (! literal_list.containsText (text))
    {
       String name = info.literal_prefix + literal_list.size ();
       storeLiteral (name, text);
    }
  }

  public Enumeration getLiterals ()
  {
    return literal_list.elements ();
  }

  /* external tokens */

  private DictTable external_token_list = new DictTable ();

  public boolean isExternalToken (String name)
  {
    return external_token_list.contains (name);
  }

  public void storeExternalToken (String name)
  {
      DictItem item = new DictItem ();
      item.ident = name;
      external_token_list.store (item);
  }

  /* tokens - keywords */

  private DictTable token_list = new DictTable ();

  public boolean containsToken (String name)
  {
    return token_list.contains (name);
  }

  public void checkToken (String name)
  {
    if (! token_list.contains (name))
    {
       DictItem item = new DictItem ();
       item.ident = name;
       token_list.store (item);
    }
  }

  public Enumeration getTokens ()
  {
    return token_list.elements ();
  }

  /* symbols - tokens and literals */

  private SymbolTable symbol_list = new SymbolTable ();

  public GramSymbol createSymbol (String ident, String text, String note)
  {
     GramSymbol symbol = symbol_list.searchText (text);

     if (symbol == null)
     {
        symbol = new GramSymbol ();

        symbol.ident = ident;
        symbol.text = text;
        symbol.note = note;

        symbol_list.store (symbol);
     }

     return symbol;
  }

  public GramSymbol findSymbol (String text)
  {
     GramSymbol symbol = symbol_list.searchText (text);

     if (symbol == null)
        err ("Unknown symbol " + text);

     return symbol;
  }

  public GramSymbol getSymbol (int inx)
  {
      GramSymbol symbol = symbol_list.searchIndex (inx);

      if (symbol == null)
        err ("Unknown symbol index: " + inx);

      return symbol;
  }

  public int getSymbolCount ()
  {
     return symbol_list.getSymbolCount ();
  }

  public Enumeration getSymbols ()
  {
    return symbol_list.elements ();
  }

  /* symbol sets */

  private TreeMap symbol_set_dict = new TreeMap ();
  private int symbol_set_count = 0;

  public String searchSymbolSet (String key)
  {
     String s = (String) symbol_set_dict.get (key);
     if (s == null)
        s = "";
     return s;
  }

  public String addSymbolSet (String key)
  {
    symbol_set_count ++;
    String name = info.internal_prefix + "set_" + symbol_set_count;
    symbol_set_dict.put (key, name);
    return name;
  }

  public GramSet allocateSet ()
  {
     GramSet result = new GramSet ();
     result.allocate (symbol_list.getSymbolCount ());
     return result;
  }

  /* rules */

  private DictTable rule_list = new DictTable ();

  public RuleNode getRule (String name)
  {
    RuleInfo item = (RuleInfo) rule_list.get (name);
    if (item == null)
       return null;
    else
       return item.node;
  }

  public RuleNode findRule (String name)
  {
     RuleNode r = getRule (name);
     if (r == null)
        err ("Undefined rule: " + name);
     return r;
  }

  public void storeRule (RuleNode rule, String name)
  {
    if (getRule (name) != null)
       err ("Redefined rule: " + name);

    RuleInfo item = new RuleInfo ();
    item.ident = name;
    item.node = rule;
    rule_list.store (item);
  }

  public Enumeration getRules ()
  {
    return rule_list.elements ();
  }

  /* list of rules */

  public RuleNode first_rule;
  public RuleNode last_rule;


  public void insertRule (RuleNode item)
  {
     if (first_rule == null)
        first_rule = item;
     else
        last_rule.next_rule = item;

     last_rule = item;
     item.next_rule = null;
  }
}
