
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import antlr.collections.AST;

class ParserOutput extends Common
{

  /* -------------------------------------------------------------------- */

  private int cond_count = 0;

  private String current_variable = "item"; // !?

  /* -------------------------------------------------------------------- */

  private void gram_error (String txt) { error (txt); }
  private void gram_warning (String txt) { warning (txt); }

  private String QuoteChr (char c) { return "\'" + c + "\'"; }
  private String QuoteStr (String s) { return "\"" + s + "\""; }

  private String StripStr (String s)
  {
     if (! empty (s))
     {
        char c = s.charAt (0);
        if (c == '\'' || c == '\"')
           s = s.substring (1, s.length()-1);
     }
     return s;
  }

  /* -------------------------------------------------------------------- */

  private String location ()
  {
     return "???";
  }

  private String description ()
  {
     return "???";
  }

  public void error (String s)
  {
     s = location () + "EG Error " + description () + ": " + s;
     err (s);
  }

  public void warning (String s)
  {
     s = location () + "EG Warning " + description () + ": " + s;
     printErrInfo (s);
  }

  /***************************** SYMBOL TYPE ******************************/

  public void initSymbols ()
  {
     info.createSymbol (info.no_symbol, "", "end of source");

     // primitive blocks

     Enumeration seq = info.getRules ();
     while (seq.hasMoreElements ())
     {
        RuleInfo item = (RuleInfo) seq.nextElement ();
        RuleNode r = item.node;
        if (r.mode == PRIMITIVE_MODE)
        {
           PrimitiveNode sub = r.first_primitive_item;
           while (sub != null)
           {
              String id = info.primitive_prefix + sub.branch;
              info.createSymbol (id, "primitive " + sub.branch, r.name); // !?
              sub = sub.next_primitive_item;
           }
        }
     }

     // delimiters

     seq = info.getLiterals ();
     while (seq.hasMoreElements ())
     {
        LiteralItem item = (LiteralItem) seq.nextElement ();
        String id = info.delimiter_prefix + item.ident;
        info.createSymbol (id, item.text, "");
     }

     // keywords

     seq = info.getTokens ();
     while (seq.hasMoreElements ())
     {
        DictItem item = (DictItem) seq.nextElement ();

        String id = info.keyword_prefix + item.ident;
        if (info.keyword_prefix == "")
           id = item.ident.toUpperCase (); // !?

        info.createSymbol (id, item.ident, "");
     }
  }

  public void declareSymbols ()
  {
     String symbol_type = info.symbol_type;

     // symbol type
     println ("enum " + symbol_type);
     println ("{");
     indent ++;

     Enumeration seq = info.getSymbols ();
     boolean first = true;
     while (seq.hasMoreElements ())
     {
        if (! first)
           println (",");

        GramSymbol item = (GramSymbol) seq.nextElement ();
        print (item.ident);

        first = false;
     }
     if (! first)
        println ("");  // end of line

     indent --;
     println ("};");
     println ("");
  }

  private void convertSymbols ()
  {
     String conv = info.internal_prefix + "conv";
     String conv_type = info.conv_type;
     String param = info.param_prefix + "param";
     String table = info.variable_prefix + "table";
     String num = Integer.toString (info.getSymbolCount ());

     String cls = info.enable_lexer_output ? info.lexer_class : info.parser_class;

     println ("string " +
              cls + "::" + conv +
              " (" + conv_type + " " + param + ")");

     println ("{");
     indent ++;
     println ("static const string " + table + " [" + num + "] = { ");
     indent ++;

     Enumeration seq = info.getSymbols ();
     boolean first = true;
     while (seq.hasMoreElements ())
     {
        if (! first)
           println (",");

        GramSymbol item = (GramSymbol) seq.nextElement ();

        String value = item.text;
        if (! empty (item.note))
           value = item.note.replace ('_', ' ');

        print (QuoteStr (value));

        first = false;
     }
     if (! first)
        println ("");  // end of line

     indent --;
     println ("};");
     println ("return " + table + " [" + param + "];");
     indent --;
     println ("}");
     println ("");
  }

  /****************************** DELIMITERS ******************************/

  private void store_delimiter (TextItem root, String id, String txt, int level)
  {
     int len = txt.length ();
     char c = txt.charAt (level); // current character

     // find position
     TextItem p = root.getFirst ();
     while (p != null && p.key < c)
        p = p.getNext ();

     TextItem n = null;  // target for new item

     if (p != null && p.key == c) // item with letter found
     {
        n = p;
     }
     else // letter not present, create new item
     {
        n = new TextItem ();
        n.key = c;

        if (p == null)
           root.insertLast (n);
        else
           p.insertBefore (n);
     }

     if (level+1 == len)
     {
        // store value at current level
        if (n.used)
           gram_error ("Identifier already defined: " + txt);
        n.used = true;
        n.complete = true;
        n.value = txt;
        n.name = id;
     }
     else
     {
        // continue with next level
        store_delimiter (n, id, txt, level+1);
     }
  }

  private void produce_delimiters (TextItem item, int level)
  {
     String id = info.delimiter_prefix + item.name;
     String stop = info.internal_prefix + "stop";

     String TokenSym = info.internal_prefix + "TokenSym";
     String InpCh    = info.internal_prefix + "InpCh";
     String NextCh   = info.internal_prefix + "NextCh";

     if (item.getFirst () == null)
     {
         println ("return " + id + ";");
     }
     else
     {
        if (level == 0)
           println ("switch (" + TokenSym + ")"); // TokenSym - already processed character
        else
           println ("switch (" + InpCh + ")"); // InpCh - new character

        println ("{");
        indent ++;

        TextItem p = item.getFirst ();
        while (p != null)
        {
           String letter = QuoteChr (p.key);

           println ("case " + letter + ":");
           indent ++;
           if (level != 0) // level == 0 => character already processed
               println (NextCh + " ();");
               // do not use GetCh () ;
               // otherwise character is not stored in input_stream

           produce_delimiters (p, level+1);
           // println ("break;");
           indent --;

           p = p.getNext ();
        }

        if (item.used)
        {
           println ("default:");
           indent ++;
           // NO NextCh ();
           println ("return " + id + ";");
           indent --;
        }

        if (level == 0)
        {
           println ("default:");
           indent ++;
           // NO NextCh ();
           println (stop + " ();");
           indent --;
        }

        indent --;
        println ("}");
     }
  }

  private void codeLiterals ()
  {
     TextItem root = new TextItem ();

     Enumeration seq = info.getLiterals ();
     while (seq.hasMoreElements ())
     {
        LiteralItem item = (LiteralItem) seq.nextElement ();
        store_delimiter (root, item.ident, item.text, 0);
     }

     String type = info.symbol_type;
     String delimiter = info.internal_prefix + "delimiter";

     String cls = info.enable_lexer_output ? info.lexer_class : info.parser_class;

     println (type + " " + cls + "::" + delimiter + " ()");
     println ("{");
     indent ++;

     produce_delimiters (root, 0);

     // println ("assert (false);"); // not necessary
     indent --;
     println ("}");
     println ("");
  }

  /******************************* KEYWORDS *******************************/

  private void storeKeyword (TextItem root, String id, String txt, int level)
  {
     int len = txt.length ();
     char c = txt.charAt (level); // current character

     // find position
     TextItem p = root.getFirst ();
     while (p != null && p.key < c)
        p = p.getNext ();

     TextItem  n = null;  // target for new item

     if (p != null && p.key == c) // item with letter found
     {
        if (p.used) // target item is used for text value
        {
           if (! p.complete) // target item can be extended with next level
           {
              storeKeyword (p, p.name, p.value, level+1); // add item at next level
              p.used = false; // mark target item as not used for text value
              p.complete = false;
              p.value = "";
              p.name = "";
           }
        }
        n = p;
     }
     else // letter not present, create new item
     {
        n = new TextItem ();
        n.key = c;

        if (p == null)
           root.insertLast (n);
        else
           p.insertBefore (n);
     }

     if (level+1 == len)
     {
        // store value at current level
        if (n.used)
           gram_error ("Identifier already defined: " + txt);
        n.used = true;
        n.complete = true;
        n.name = id;
        n.value = txt;
     }
     else if (! n.used && n.getFirst () == null)
     {
        n.used = true;
        n.complete = false;
        n.name = id;
        n.value = txt;
     }
     else
     {
        // continue with next level
        storeKeyword (n, id, txt, level+1);
     }
  };

  private void produceKeywords (TextItem item, int level)
  {
     String id = item.name;
     String id_as_str = QuoteStr (item.value);

     int level_plus_one = level + 1;

     String param = info.param_prefix + "param";
     String len = info.variable_prefix + "len";

     if (item.getFirst () == null)
     {
        // assert (item.used);
        if (item.complete)
        {
           println ("if (" + len + " == " + level + ")");
           indent ++;
           println ("return " + id + ";");
           indent --;
        }
        else if (item.value.length () - 1 == level)
        {
           String letter = QuoteChr (item.value.charAt (level));
           println ("if (" + len + " == " + level_plus_one + " && " + param + " [" + level + " ] == " + letter + ")");
           indent ++;
           println ("return " + id + ";");
           indent --;
        }
        else
        {
           println ("if (" + param + " == " + id_as_str + ")");
           indent ++;
           println ("return " + id + ";");
           indent --;
        }
     }
     else
     {
        boolean step = false; // indentation

        if (item.complete)
        {
           println ("if (" + len + " == " + level + ")");
           indent ++;
           println ("return " + id + ";");
           indent --;
           println ("else");
           step = true;
        }
        else if (item.used)
        {
           // impossible
           // assert (false);
        }
        else if (level == 0)
        {
           // nothing
        }
        else
        {
           println ("if (" + len + " >= " + level_plus_one + ")");
           step = true;
        }

        if (step)
           indent ++;

        TextItem p = item.getFirst ();

        if (p.getNext () == null && p.getFirst () == null)
        {
           // optimalization for one item

           // assert (p.used);

           if (p.complete)
           {
              String letter = QuoteChr (p.key);
              String sub = p.name;

              println ("if (" + len + " == " + level_plus_one + " && " + param + " [" + level + "] == " + letter + ")");
              indent ++;
              println ("return " + sub + ";");
              indent --;
           }
           else
           {
              String sub = p.name;
              String sub_as_str = QuoteStr (p.value);

              println ("if (" + param + " == " + sub_as_str + ")");
              indent ++;
              println ("return " + sub + ";");
              indent --;
           }
        }
        else
        {
           println ("switch (" + param + " [" + level + "])");
           println ("{");
           indent ++;

           while (p != null)
           {
              String letter = QuoteChr (p.key);
              println ("case " + letter + ":");
              indent ++;

              produceKeywords (p, level+1);

              println ("break;");
              indent --;

              p = p.getNext ();
           }

           indent --;
           println ("}");
        }

        if (step)
           indent --;
     }
  }

  private void codeTokens ()
  {
     TextItem root = new TextItem ();

     Enumeration seq = info.getTokens ();
     while (seq.hasMoreElements ())
     {
        DictItem item = (DictItem) seq.nextElement ();
        storeKeyword (root, info.keyword_prefix + item.ident, item.ident, 0);
     }

     String type = info.symbol_type;
     String keyword = info.internal_prefix + "keyword";
     String param = info.param_prefix + "param";
     String len = info.variable_prefix + "len";
     String no_symbol = info.no_symbol;

     String cls = info.enable_lexer_output ? info.lexer_class : info.parser_class;

     println (type + " " +
              cls + "::" + keyword +
              " (string " + param + ")");
     println ("{");
     indent ++;
     println ("int " + len + " = " + param + ".length ();");
     println ("assert (" + len + " > 0);");

     produceKeywords (root, 0);

     println ("return " + no_symbol + ";");
     indent --;
     println ("}");
     println ("");
  }

  /***************************** SYMBOL SETS ******************************/

  private String ascii_key (GramSet s)
  {
     // create String (7 items per byte)
     // used as key in dictionary of allocated set variables

     final int step = 7; // used bits per byte
     final byte space = ' '; // representation of zero
     int len = (s.count + step - 1) / step; // number of bytes

     byte [] tab = new byte [len]; // allocate array
     for (int i = 0; i < len; i++)
        tab [i] = space; // empty text
     // tab [len] = 0;

     int pos = 0; // byte position
     int bit = 0; // bit position
     for (int inx = 0; inx < s.count; inx++)
     {
        if (s.table [inx]) // data from TGramSet
           tab [pos] += (1 << bit); // add bit to temporary array

        if (bit < step-1)
        {
           bit ++;
        }
        else
        {
           bit = 0;
           pos ++;
        }
     }

     return new String (tab);
  }

  /* -------------------------------------------------------------------- */

  private String condition_set (GramSet s)
  {
     // declare byte array for symbol set

     // assert (s != null);

     String key = ascii_key (s);
     String result = "";

     // check if variable already exists
     result = info.searchSymbolSet (key);
     if (empty (result))
     {
        result = info.addSymbolSet (key); // store new variable name  into dictionary

        int save_indent = indent;
        switchTo (parserDeclPack); // store declaration to another stream

        // comment

        String txt = result + " : ";
        int cnt = txt.length ();
        boolean start = true;
        for (int inx = 0; inx < s.count; inx ++)
           if (s.table [inx])
           {
              String id = info.getSymbol(inx).ident;
              // String id = "set_" + inx; // !?

              if (! start)
              {
                 txt += ",";
                 cnt ++;

                 if (cnt + id.length () > 64)
                 {
                    // new comment line
                    final char lf  = '\n';
                    txt += lf;
                    txt += "   ";
                    cnt = 3;
                 }
                 else
                 {
                    txt += " ";
                    cnt ++;
                 }

              }
              start = false;

              txt += id;
              cnt += id.length ();
           }

        // create temporary array (8 items per byte) (for generated code)

        int len = (s.count + 7) / 8; // number of bytes
        short [] tab = new short [len]; // allocate
        for (int i = 0; i < len; i ++) // clear
           tab[i] = 0;

        int pos = 0; // byte position
        int bit = 0; // bit position
        for (int inx = 0; inx < s.count; inx ++)
        {
           if (s.table [inx]) // data from GramSet
              tab [pos] += (1 << bit); // add bit to temporary array

           if (bit < 7)
              bit ++;
           else
           {
              bit = 0;
              pos ++;
           }
        }

        // create list of values

        println ("/* " + txt + " */");
        println ("");
        println ("const unsigned char " + result + " [" + len + "] = {");
        // println ("{");
        indent ++;

        for (int i = 0; i < len; i ++)
        {
           print ("" + tab [i]);
           if (i < len-1)
              print (",");
           println ("");
        }

        indent --;
        println ("};");
        println ("");

        switchTo (parserImplPack); // back to original stream
        indent = save_indent;
     }

     return result;
  }


  /* -------------------------------------------------------------------- */

  private String and_expr (String a, String b)
  {
     String result = "";

     if (empty (a))
        result = b;
     else if (empty (b))
        result = a;
     else
        result = a + " && " + b;

     return result;
  }

  private String or_expr (String a, String b)
  {
     String result = "";

     if (empty (a))
        result = b;
     else if (empty (b))
        result = a;
     else
        result = a + " || " + b;

     return result;
  }

  private String parenthesis_expr (String a)
  {
     if (! empty (a))
        a = "(" + a + ")";
     return a;
  }

  /* -------------------------------------------------------------------- */

  private String plain_condition_expr (GramSet s)
  {
     final int limit = 3; // maximum number of items for inline testing
     String result = "";

     if (! s.epsilon)
     {
        if (s.item_count () <= limit)
        {
           // compare with single items
           for (int i = 0; i < s.count; i++)
              if (s.table [i])
              {
                 String symbol = info.internal_prefix + "symbol";
                 String name = info.getSymbol(i).ident;

                 String tmp_expr = symbol + " (" + name + ")";
                 result = or_expr (result, tmp_expr);
              }
        }
        else
        {
            // use byte array
            String id = condition_set (s); // symbol set identifier
            String tmp_expr = info.internal_prefix + "test (" + id + ")";
            result = or_expr (result, tmp_expr);

            // !? !!
            // print (" /* " + id + " */ ");
        }
     }
     return result;
  }

  private String condition_expr (GramSet s)
  {
     String result = plain_condition_expr (s);

     if (result == "") // !?
     {
        result = "unknown";
     }

     return result;
  }

  /* ---------------------------------------------------------------------- */

  private void concat_subrule_symbols (String subrule_name, GramSet result)
  {
     RuleNode r = info.findRule (subrule_name);

     if (! r.evaluating)
     {
        if (r.symbol_set == null)
           rule_symbols (r);

        result.concatenate (r.symbol_set);
     }
  }

  private void concat_term_symbol (String term_name, GramSet result)
  {
     term_name = StripStr (term_name);
     GramSymbol t = info.findSymbol (term_name);
     boolean ignore = t.text.equals ("<") && info.ignore_tag_marks;

     if (! ignore)
     {
        if (t == null)
           println ("UNKNOWN (1)" + term_name);

        if (t != null) // !?
           result.concatenate_symbol (t.index);
     }
  }

  private void concat_primitive_symbol (String prim_name, GramSet result)
  {
     GramSymbol t = info.findSymbol ("primitive " + prim_name); // !?

     if (t == null)
        println ("UNKNOWN (2) " + prim_name);

     if (t != null) // !?
        result.concatenate_symbol (t.index);
  }

  /* ---------------------------------------------------------------------- */

  private GramSet alt_symbols (AltNode seq)
  {
     GramSet result = info.allocateSet ();

     seq.directed = seq.silent;

     SourceNode item = seq.first_item;
     while (item != null && result.epsilon)
     {
        if (item instanceof TerminalNode)
        {
           // terminal
           TerminalNode term = (TerminalNode) item;
           // if (! inside_predicate ())
           {
              concat_term_symbol (term.term_name, result);
           }
        }
        else if (item instanceof NontermNode)
        {
           // nonterminal
           NontermNode nonterm = (NontermNode) item;
           concat_subrule_symbols (nonterm.subrule, result);
        }
        else if (item instanceof BlockNode)
        {
           BlockNode blk = (BlockNode) item;

           if (blk.syntactic_predicate)
           {
              // syntactic predicate
              seq.directed = true;
           }
           else
           {
              // ebnf
              GramSet tmp = block_symbols (blk);
              if (blk.mark == '?' || blk.mark == '*')
                 tmp.epsilon = true;

              result.concatenate (tmp);
           }
        }
        else if (item instanceof SemanticNode)
        {
           // semantic predicate
           seq.directed = true;
        }
        /*
        else if (item.conv_GrmTryOption () != null)
        {
           // also semantic predicate
           seq.directed = true;
        }
        */

        item = item.next_item;
     }

     // !? seq.empty = result.epsilon;

     return result;
  }

  private GramSet block_symbols (BlockNode blk)
  {
     GramSet result = info.allocateSet ();
     result.epsilon = (blk.first_alt == null);

     GramSet check = info.allocateSet ();
     check.epsilon = false;

     AltNode seq = blk.first_alt;
     while (seq != null)
     {
        GramSet tmp = alt_symbols (seq);

        if (! seq.directed)
        {
           String note = check.compatible (tmp);
           if (! empty (note))
              gram_warning ("GRAMMAR CONFLICT: " + note);

           check.unite (tmp);
        }

        result.unite (tmp);

        seq = seq.next_alt;
     }

     return result;
  }

  /* ---------------------------------------------------------------------- */

  private GramSet primitive_block_symbols (RuleNode r)
  {
     GramSet result = info.allocateSet ();

     PrimitiveNode item = r.first_primitive_item;
     result.epsilon = (item == null);

     while (item != null)
     {
        GramSet tmp = info.allocateSet ();
        concat_primitive_symbol (item.branch, tmp);
        result.unite (tmp);
        item = item.next_primitive_item;
     }

     return result;
  }

  private GramSet binary_block_symbols (RuleNode r)
  {
     GramSet result = info.allocateSet ();

     concat_subrule_symbols (r.binary_sub_rule, result);

     return result;
  }

  private GramSet unary_block_symbols (RuleNode r)
  {
     GramSet result = info.allocateSet ();

     concat_term_symbol (r.unary_key, result);
     // NO concat_subrule_symbols (blk.sub_rule, result);

     return result;
  }

  /* -------------------------------------------------------------------- */

  private void rule_symbols (RuleNode r)
  {
     r.evaluating = true;

     r.symbol_set = info.allocateSet ();

     if (r.normal_block != null)
     {
        r.symbol_set = block_symbols (r.normal_block);
     }

     else if (r.mode == BINARY_MODE)
     {
        r.symbol_set = binary_block_symbols (r);
     }
     else if (r.mode == UNARY_MODE)
     {
        r.symbol_set = unary_block_symbols (r);
     }
     else if  (r.mode == PRIMITIVE_MODE)
     {
        r.symbol_set = primitive_block_symbols (r);
     }

     r.evaluating = false;
  }

  /************************* SYNTACTIC PREDICATES *************************/

  private boolean simple_item (SourceNode item)
  {
     boolean ok = false;

     if (item instanceof TerminalNode)
     {
        // terminal
        ok = true;
     }
     else if (item instanceof NontermNode)
     {
        // nonterminal
        NontermNode nonterm = (NontermNode) item;
        RuleNode r = info.findRule (nonterm.subrule); // sub-rule
        if (r.mode == PRIMITIVE_MODE)
           ok = true;
     }

     return ok;
  }

  private boolean superfluous_syntactic_predicate (BlockNode blk)
  {
     // check predicate expression
     // test length of alternatives

     boolean result = true;

     AltNode seq = blk.first_alt;

     if (seq == null)
        gram_error ("Empty syntactic predicate");

     while (seq != null)
     {
        SourceNode item = seq.first_item;

        if (item == null)
           gram_error ("Empty alternative in syntactic predicate");

        if (item.next_item != null)
           result = false; // two or more items in alternative

        while (item != null)
        {
           if (! simple_item (item))
              result = false;

           item = item.next_item;
        }

        seq = seq.next_alt;
     }

     return result;
  }

  private boolean simple_syntactic_predicate (BlockNode blk)
  {
     boolean result = true;

     AltNode seq = blk.first_alt;

     if (seq == null)
        gram_error ("Empty syntactic predicate");

     while (seq != null)
     {
        SourceNode item = seq.first_item;

        if (item == null)
           gram_error ("Empty alternative in syntactic predicate");

        while (item != null)
        {
           if (! simple_item (item))
              result = false;

           item = item.next_item;
        }

        seq = seq.next_alt;
     }

     return result;
  }

  private String syntactic_predicate_expr (SourceNode item)
  {
     GramSet result = info.allocateSet ();

     if (item instanceof TerminalNode)
     {
        // terminal
        TerminalNode term = (TerminalNode) item;
        concat_term_symbol (term.term_name, result);
     }
     else if (item instanceof NontermNode)
     {
        // nonterminal
        NontermNode nonterm = (NontermNode) item;
        concat_subrule_symbols (nonterm.subrule, result);
     }

     return condition_expr (result);
  }

  private void syntactic_predicate_item (SourceNode item)
  {
     String expr = syntactic_predicate_expr (item);

     println ("if (" + expr + ")");

     if (item.next_item == null)
     {
        indent ++;
        String answer = info.variable_prefix + "answer";
        println (answer + " = true;");
        indent --;
     }
     else
     {
        println ("{");
        indent ++;

        String next = info.internal_prefix + "next";
        println (next + " ();");

        syntactic_predicate_item (item.next_item);

        indent --;
        println ("}");
     }
  }

  private void syntactic_predicate_alt (AltNode seq)
  {
     SourceNode item = seq.first_item;
     if (item == null)
        gram_error ("Empty alternative in syntactic predicate");

     syntactic_predicate_item (item);
  }

  private String syntactic_predicate_condition (BlockNode blk, boolean defn)
  {
     if (blk.condition_inx == 0)
     {
        cond_count ++;
        blk.condition_inx = cond_count;
     }

     String func = info.internal_prefix + "condition_" + blk.condition_inx;

     if (defn)
     {
        String answer = info.variable_prefix + "answer";
        String inx = info.variable_prefix + "inx";
        String mark_type = info.mark_type;
        String mark = info.internal_prefix + "mark";
        String rewind = info.internal_prefix + "rewind";
        String save_trace = info.internal_prefix + "save_trace";
        String restore_trace = info.internal_prefix + "restore_trace";

        println ("bool " + info.parser_class + "::" + func + " ()");
        println ("{");
        indent ++;

        // create code

        if (simple_syntactic_predicate (blk))
        {
           // without try statement

           println ("bool " + answer + " = false;");
           println (mark_type + " " + inx + " = " + mark + " ();");
           println ("");

           AltNode seq = blk.first_alt;

           if (seq == null)
              gram_error ("Empty syntactic predicate");

           while (seq != null)
           {
              syntactic_predicate_alt (seq);
              seq = seq.next_alt;
           }

           println ("");
           println (rewind + " (" + inx + ");");
        }
        else
        {
           // with try statement

           println ("bool " + answer + " = true;");
           println (mark_type + " " + inx + " = " + mark + " ();");
           println (save_trace + " ();");
           println ("");

           println ("try");
           println ("{");
           indent ++;

           // inner code

           pred_block (blk);

           // finish try statement

           indent --;
           println ("}");

           println ("catch (IOException & e)");
           println ("{");
           indent ++;
           println (answer + " = false;");
           indent --;
           println ("}");

           println ("");
           println (restore_trace + " ();");
           println (rewind + " (" + inx + ");");
        }

        // end of function

        println ("return " + answer + ";");

        indent --;
        println ("}");
        println ("");
     }

     // create expression

     return func + " ()";
  }

  private void pred_block (BlockNode blk)
  {
     AltNode seq = blk.first_alt;
     pred_alternative (seq);

     if (seq.next_alt != null)
        error ("Predicate too complicated");
  }

  private void pred_alternative (AltNode seq)
  {
     SourceNode item = seq.first_item;
     while (item != null)
     {
        if (item instanceof TerminalNode)
           pred_terminal ((TerminalNode) item);

        else if (item instanceof NontermNode)
           pred_nonterm ((NontermNode) item);

        else
           error ("Predicate too complicated");

        item = item.next_item;
     }
  }

  private void pred_terminal (TerminalNode node)
  {
     put_terminal (node.term_name);
  }

  private void pred_nonterm (NontermNode node)
  {
     RuleNode r = info.findRule (node.subrule); // sub-rule
     String parse = info.rule_prefix + r.name;
     println (parse + " ();");
  }

  /* ---------------------------------------------------------------------- */

  private String semantic_predicate_condition (SemanticNode predicate)
  {
     String result = "";

     String func = predicate.name;

     if (predicate.no_params)
        result = func + " ()";
     else
        result = func + " (" + current_variable + ")";

     return result;
  }

  /* ---------------------------------------------------------------------- */

  private String additional_condition (AltNode seq, boolean defn)
  {
     String result = "";

     SourceNode item = seq.first_item;
     while (item != null)
     {
        String tmp_expr = "";

        if (item instanceof BlockNode)
        {
           BlockNode blk = (BlockNode) item;
           if (blk.syntactic_predicate)
           {
              if (item != seq.first_item)
                 error ("Syntactic predicate not supported here");

              if (! superfluous_syntactic_predicate (blk)) // !?
                 tmp_expr = syntactic_predicate_condition (blk, defn);
           }
        }
        else if (item instanceof SemanticNode)
        {
           SemanticNode predicate = (SemanticNode) item;

           if (item != seq.first_item)
              error ("Semantic predicate not supported here");

           tmp_expr = semantic_predicate_condition (predicate);
       }
       /*
       else
       {
          GrmTryOption * option = item.conv_GrmTryOption ();
          if (option != null)
          {
             if (item.prev != null)
                error ("<try> option not supported here");
             tmp_expr = try_option_condition (option);
          }
       }
       */

        result = and_expr (result, tmp_expr);
        item = item.next_item;
     }

     return result;
  }

  /* -------------------------------------------------------------------- */

  /* declare / define syntactic predicate functions */

  private void cond_block (BlockNode blk, boolean defn)
  {
     AltNode seq = blk.first_alt;
     while (seq != null)
     {
        additional_condition (seq, defn);

        // recursion
        SourceNode item = seq.first_item;
        while (item != null)
        {
           if (item instanceof BlockNode)
              cond_block ((BlockNode) item, defn);

           item = item.next_item;
        }
        seq = seq.next_alt;
     }
  }

  private void cond_rule (RuleNode node, boolean defn)
  {
     if (node.normal_block != null)
        cond_block (node.normal_block, defn);
  }

  /************************* LOCATIONS and NOTES **************************/

  private void code_location (int inx, String item_id)
  {
     if (inx != 0 && info.locations && ! info.ignore_locations)
     {
        String getLocation = info.internal_prefix + "getLocation";
        String subitem_id = info.LocationName (inx);

        println (item_id + "->" + subitem_id + " = " + getLocation + " ();");
     }
  }

  private void code_note (int inx, String item_id, boolean start)
  {
     if (inx != 0 &&
         info.notes &&
         (start || ! info.only_start_notes) &&
         ! info.ignore_notes)
     {
        String proc_id, subitem_id;

        if (start)
        {
           proc_id = info.internal_prefix + "getStartNote";
           subitem_id = info.StartNoteName (inx);
        }
        else
        {
           proc_id = info.internal_prefix + "getStopNote";
           subitem_id = info.StopNoteName (inx);
        }

        println (item_id + "->" + subitem_id + " = " + proc_id + " ();");
     }
  }

  private void code_location_and_start_note (BasicNode item, String item_id)
  {
     code_note (item.note_index, item_id, true);
     code_location (item.note_index, item_id);
  }

  private void code_stop_note (BasicNode item, String item_id)
  {
     code_note (item.note_index, item_id, false);
  }

  /********************** TERMINALS and NONTERMINALS **********************/

  /* terminals */

  private void put_terminal (String name)
  {
     String match = info.internal_prefix + "match";
     String param = info.findSymbol (name) . ident;
     println (match + " (" + param + ");");
  }

  private void code_terminal (TerminalNode node)
  {
     GramSymbol t = info.findSymbol (node.term_name);
     boolean ignore = t.text.equals ("<") && info.ignore_tag_marks;
     if (! ignore)
     {
        code_location_and_start_note (node, current_variable);
        put_terminal (node.term_name);
        code_stop_note (node, current_variable);
     }
  }

  /* nonterminals */

  private void code_nonterm (NontermNode node)
  {
     RuleNode r = info.findRule (node.subrule); // sub-rule
     int kind = node.kind;

     if (node.allocated)
        code_location_and_start_note (node, current_variable);

     String item = current_variable;
     String parse = info.rule_prefix + r.name;
     // println ("// " + r.name + " " + kind);

     if (kind == LABELED)
     {
        String field = info.field_prefix + node.label;
        println (item + "->" + field + " = " + parse + " ();");
     }
     else if (kind == INSERT || kind == ADD)
     {
        String insert_last = info.list_prefix + "add" + node.sequence.suffix;

        println (item + "->" + insert_last + " (" + parse + " ());");
     }
     else if (kind == RESULT || kind == SELECT)
     {
        println (item + " = " + parse + " ();");
     }
     else if (kind == CALL)
     {
        println (parse + " (" + item + ");");
     }
     else if (kind == CHOOSE1 || kind == CHOOSE2)
     {
        String temp = info.variable_prefix + "tmp";

        if (node.choose_compatible_type)
        {
           if (kind == CHOOSE1)
              println (item + " = " + parse + " ();");
           else
              println (item + " = " + parse + " (" + item + ");");
        }
        else
        {
           if (kind == CHOOSE1)
              println (r.type.decl () + " " + temp + " = " + parse + " ();"); // !?
           else
              println (item + " = " + parse + " (" + temp + ");");
        }
     }
     else if (kind == ANONYMOUS)
     {
        gram_warning ("Anonymous field not allowed here");
     }
     else
     {
        gram_error ("Internal gram_error (1)");
     }

     if (node.allocated)
        code_stop_note (node, current_variable);
  }

  /******************************** BLOCK *********************************/

  /* set option */

  private void code_set_option (SetOptionNode node)
  {
     // code for <set name=value>
     println (current_variable + "->" + info.field_prefix + node.name + " = " + node.value + ";");
  }

  /* inc option */

  private void code_inc_option (IncOptionNode node)
  {
     // code for <inc name>
     println (current_variable + "->" + info.field_prefix + node.name + " ++;");
  }

  /* invoke option */

  private void code_invoke_option (InvokeOptionNode node)
  {
     // code for <invoke name>
     println (current_variable + "->" + node.name + " (this);");
  }

  /* execute option */

  private void code_execute_option (ExecuteOptionNode node)
  {
      if (! info.ignore_execute_options)
      {
         // code for <execute name>
         if (node.no_params)
            println (node.name + " ();");
         else
            println (node.name + " (" + current_variable + ");");
      }
  }

  /* alternative */

  private String alt_condition (AltNode seq)
  {
     GramSet s = alt_symbols (seq);
     String result = plain_condition_expr (s);

     String tmp_expr = additional_condition (seq, false);
     if (! empty (tmp_expr))
     {
        result = parenthesis_expr (result);
        result = and_expr (result, tmp_expr);
     }

     return result;
  }

  private void code_alternative (BlockNode blk, AltNode seq)
  {
     if (blk.alt_count != 1 || blk.mark == '?')
     {
        if (seq.index != 1)
           print ("else ");

        String expr = alt_condition (seq);

        boolean cond = true;

        if (blk.mark == ' ' && seq.index == blk.alt_count)
        {
           // if not last alternatve in simple block
           GramSet s = alt_symbols (seq);
           if (s.epsilon)
              cond = false;
        }

        if (cond)
           println ("if (" + expr + ")");
        else
           println (""); // end of line

        println ("{");
        indent ++;
     }

     SourceNode item = seq.first_item;
     while (item != null)
     {
        if (item instanceof TerminalNode)
           code_terminal ((TerminalNode) item);

        else if (item instanceof NontermNode)
           code_nonterm ((NontermNode) item);

        else if (item instanceof BlockNode)
        {
           BlockNode b = (BlockNode) item;
           if (! b.syntactic_predicate)
              code_block (b);
        }

        else if (item instanceof SetOptionNode)
           code_set_option ((SetOptionNode) item);

        else if (item instanceof IncOptionNode)
           code_inc_option ((IncOptionNode) item);

        else if (item instanceof ExecuteOptionNode)
           code_execute_option ((ExecuteOptionNode) item);

        else if (item instanceof InvokeOptionNode)
           code_invoke_option ((InvokeOptionNode) item);

        item = item.next_item;
     }

     if (blk.alt_count != 1 || blk.mark == '?')
     {
        indent --;
        println ("}");
     }
  }

  /* block */

  private String block_condition (BlockNode blk)
  {
     GramSet s = block_symbols (blk);
     String result = condition_expr (s);
     return result;
  }

  private void code_block (BlockNode blk)
  {
     if (blk.mark == '*')
     {
        String expr = block_condition (blk);
        println ("while (" + expr + ")");
        println ("{");
        indent ++;
     }
     else if (blk.mark == '+')
     {
        println ("do");
        println ("{");
        indent ++;
     }

     AltNode seq = blk.first_alt;
     while (seq != null)
     {
        code_alternative (blk, seq);
        seq = seq.next_alt;
     }

     if (blk.alt_count != 1)
        if (blk.mark == '+' || blk.mark == ' ')
        {
           String stop = info.internal_prefix + "stop";

           boolean cond = true;

           if (blk.mark == ' ')
           {
              // if not last alternatve in simple block
              GramSet s = alt_symbols (blk.last_alt);
              if (s.epsilon)
                 cond = false;
           }

           if (cond)
           {
              println ("else");
              // !? println ("{");
              indent ++;
              println (stop + " ();");
              indent --;
              // !? println ("}");
           }
        }

     if (blk.mark == '*')
     {
        indent --;
        println ("}");
     }
     else if (blk.mark == '+')
     {
        String expr = block_condition (blk);
        indent --;
        println ("} while (" + expr + ");");
     }
  }

  /*************************** PRIMITIVE BLOCK ****************************/

  private void code_primitive (RuleNode node)
  {
     String recall = info.util_func_prefix + "recall_" + node.name; // !?
     String ident_replace = info.internal_prefix + "ident_replace";

     if (node.type == info.ident_type && ! empty (info.replace_type))
     {
        println ("if (current_stencil != NULL)");
        indent ++;
        println ("return " + ident_replace + " (" + recall + " ()" + ");");
        indent --;
        println ("else");
        indent ++;
        println ("return " + recall + " ();");
        indent --;
     }
     else
     {
        println ("return " + recall + " ();");
     }
  }

  /***************************** UNARY BLOCK ******************************/

  private void code_unary (RuleNode node)
  {
     String key = node.unary_key;
     String selector_value = node.unary_selector;
     String sub_rule = node.unary_sub_rule;

     String result = info.variable_prefix + "result";
     String unary_type = info.type_prefix + info.unary_expr_type;
     String selector_field = info.field_prefix + info.unary_expr_selector;

     println (unary_type + " * " + result + " = new " + unary_type + ";");

     code_location_and_start_note (info.unary_node, result);
     put_terminal (key);
     code_stop_note (info.unary_node, result);

     println (result + "->" + selector_field + " = " + selector_value + ";");

     println (result + "->" + info.field_prefix + info.unary_expr_param + " = " +
              info.rule_prefix + sub_rule + " ();");

     current_variable = result;
     // NO println ("return " + result + ";");
  }

  /***************************** BINARY BLOCK *****************************/

  private void code_binary_item (RuleNode r, BinaryNode node, int inx)
  {
     String result = info.variable_prefix + "result";
     String done = info.variable_prefix + "done";
     String expr_type = info.type_prefix + info.expr_type;

     String sub_rule = info.rule_prefix + r.binary_sub_rule;
     String selector_value = node.selector_value;

     String item = info.variable_prefix + "item";
     String binary_type = info.type_prefix + info.binary_expr_type;
     String selector_field = info.field_prefix + info.binary_expr_selector;
     String left_field = info.field_prefix + info.binary_expr_left;
     String right_field = info.field_prefix + info.binary_expr_right;

     String id = StripStr (node.key);

     String symbol = info.internal_prefix + "symbol";
     String name = info.findSymbol (id) . ident;

     if (inx != 1)
        print ("else ");

     println ("if (" + symbol + " (" + name + "))");

     println ("{");
     indent ++;

     println (binary_type + " * " + item + " = new " + binary_type + ";");

     println (item + "->" + selector_field + " = " + selector_value + ";");
     println (item + "->" + left_field + " = " + result + ";");

     code_location_and_start_note (info.binary_node, item);
     String next = info.internal_prefix + "next";
     println (next + " ();");
     code_stop_note (info.binary_node, item);

     if (! empty (node.middle_branch) && ! empty (node.middle_key))
     {
        String middle_field = info.field_prefix + info.binary_expr_middle;
        String middle_rule = info.rule_prefix + node.middle_branch;

        println (item + "->" + middle_field + " = " + middle_rule + " ();");

        code_location_and_start_note (info.binary_node.another, item);

        String middle_id = StripStr (node.middle_key);
        String middle_name = info.findSymbol (middle_id) . ident;
        String match = info.internal_prefix + "match";
        println (match + " (" + middle_name + ");");

        code_stop_note (info.binary_node.another, item);
     }

     println (item + "->" + right_field + " = " + sub_rule + " ();");

     println (result + " = " + item + ";");
     indent --;
     println ("}");
  }

  private void code_binary (RuleNode node)
  {
     String result = info.variable_prefix + "result";
     String expr_type = info.type_prefix + info.expr_type;
     String sub_rule = info.rule_prefix + node.binary_sub_rule;
     String done = info.variable_prefix + "done";

     // println (expr_type + " * " + result + " = " + sub_rule + " ();");
     println (expr_type + " * " + result + ";");
     println (result + " = " + sub_rule + " ();");

     println ("bool " + done + " = false;");
     println ("while (! " + done + ")");
     println ("{");
     indent ++;

     int inx = 1;
     BinaryNode item = node.first_binary_item;
     while (item != null)
     {
        code_binary_item (node, item, inx);
        inx ++;
        item = item.next_binary_item;
     }

     println ("else");
     // println ("{");
     indent ++;
     println (done + " = true;");
     indent --;
     // println ("}");

     indent --;
     println ("}");

     current_variable = result;
     // NO println ("return " + result + ";");
  }

  /* optimized binary expressions */

  private void calculateLevel (RuleNode rule)
  {
     RuleNode r = info.findRule (rule.binary_sub_rule);
     if (r.mode == BINARY_MODE)
     {
        if (r.binary_level == 0)
           calculateLevel (r);
        rule.binary_level = r.binary_level + 1;
     }
     else
     {
        rule.binary_level = 1;
     }
  }

  public void code_expr_decl ()
  {
     String func = info.internal_prefix + info.binary_expr_func;

     String param_max_level = info.param_prefix + "max_level";
     String param_result_level = info.param_prefix + "result_level";
     String param_result_selector = info.param_prefix + "result_selector";

     String expr_type = info.type_prefix + info.expr_type;
     String selector_type = info.type_prefix + info.binary_expr_type + "_" + info.binary_expr_selector;

     println (expr_type + " * " + func +
              " (int " + param_max_level + ", " +
              "int * " + param_result_level + ", " +
              selector_type + " * " + param_result_selector + ");");
  }

  public void code_expr_method ()
  {
     String func = info.internal_prefix + info.binary_expr_func;

     String param_max_level = info.param_prefix + "max_level";
     String param_result_level = info.param_prefix + "result_level";
     String param_result_selector = info.param_prefix + "result_selector";

     String result = info.variable_prefix + "result";
     String selector = info.variable_prefix + "selector";
     String level = info.variable_prefix + "level";

     String expr_type = info.type_prefix + info.expr_type;
     String selector_type = info.type_prefix + info.binary_expr_type + "_" + info.binary_expr_selector;

     String item = info.variable_prefix + "item";
     String binary_type = info.type_prefix + info.binary_expr_type;
     String selector_field = info.field_prefix + info.binary_expr_selector;
     String left_field = info.field_prefix + info.binary_expr_left;
     String right_field = info.field_prefix + info.binary_expr_right;

     String sub_rule = "";
     RuleNode r = info.first_rule;
     while (r != null)
     {
        if (r.mode == BINARY_MODE)
        {
           calculateLevel (r);

           if (r.binary_level == 1 && sub_rule == "")
              sub_rule = info.rule_prefix + r.binary_sub_rule;
        }
        r = r.next_rule;
     }
     if (sub_rule == "")
        error ("Cannot find basic binary expression");

     // function declaration

     println (expr_type + " * " +
              info.parser_class + "::" + func +
              " (int " + param_max_level + ", " +
              "int * " + param_result_level + ", " +
              selector_type + " * " + param_result_selector + ")");
     println ("{");
     indent ++;

     // left operand and local declarations

     println (expr_type + " * " + result + " = " + sub_rule + " ();");
     println ("int " + level + ";");
     println (selector_type + " " + selector + ";");
     println ("");

     // binary op and priority

     println ("switch (" + info.internal_prefix + "LA (1))");
     println ("{");
     indent ++;

     r = info.first_rule;
     while (r != null)
     {
        if (r.mode == BINARY_MODE)
        {
           BinaryNode node = r.first_binary_item;
           while (node != null)
           {
               String id = StripStr (node.key);

               String symbol = info.internal_prefix + "symbol";
               String name = info.findSymbol (id) . ident;

               println ("case " + name + ":");
               indent ++;

               println (selector + " = " + node.selector_value + ";");
               println (level + " = " + r.binary_level + ";");

               println ("break;");
               indent --;

              // inx ++;
              node = node.next_binary_item;
           }
        }
        r = r.next_rule;
     }
     println ("default:");
     indent ++;
     println (level + " = 0;");
     indent --;
     indent --;
     println ("}");

     // loop for another operands

     println ("");
     println ("while (" + level + " != 0 && " + level + " <= " + param_max_level + ")");
     println ("{");
     indent ++;

     // construct tree item

     println (binary_type + " * " + item + " = new " + binary_type + ";");

     println (item + "->" + selector_field + " = " + selector + ";");
     println (item + "->" + left_field + " = " + result + ";");

     code_location_and_start_note (info.binary_node, item);
     String next = info.internal_prefix + "next";
     println (next + " ();");
     code_stop_note (info.binary_node, item);

     // middle operand in ternary expressions

     r = info.first_rule;
     while (r != null)
     {
        if (r.mode == BINARY_MODE)
        {
           BinaryNode node = r.first_binary_item;
           while (node != null)
           {

              if (! empty (node.middle_branch) && ! empty (node.middle_key))
              {
                 String middle_field = info.field_prefix + info.binary_expr_middle;
                 String middle_rule = info.rule_prefix + node.middle_branch;

                 println ("if (" + selector + " == " + node.selector_value + ")");
                 println ("{");
                 indent ++;

                 println (item + "->" + middle_field + " = " + middle_rule + " ();");

                 code_location_and_start_note (info.binary_node.another, item);

                 String middle_id = StripStr (node.middle_key);
                 String middle_name = info.findSymbol (middle_id) . ident;
                 String match = info.internal_prefix + "match";
                 println (match + " (" + middle_name + ");");

                 code_stop_note (info.binary_node.another, item);

                 indent --;
                 println ("}");
              }

              node = node.next_binary_item;
           }
        }
        r = r.next_rule;
     }

     // right operand

     println (item + "->" + right_field + " = " +
              func + " (" + level + "-1 , &" + level + ",  &" + selector + ");");

     // String sub_rule = info.rule_prefix + r.binary_sub_rule;
     // println (item + "->" + right_field + " = " + sub_rule + " ();");

     // end of loop

     println (result + " = " + item + ";");
     indent --;
     println ("}");

     // results

     println ("");

     println ("if (" + param_result_level + " != NULL)");
     indent ++;
     println ("* " + param_result_level + " = " + level + ";");
     indent --;

     println ("");

     println ("if (" + level + "!= 0 && " + param_result_selector + " != NULL)");
     // selector valid only if level is non zero
     indent ++;
     println ("* " + param_result_selector + " = " + selector + ";");
     indent --;

     println ("");

     println ("return " + result + ";");

     indent --;
     println ("}");
     println ("");
  }

  private void code_expr_call (RuleNode node)
  {
     String expr_type = info.type_prefix + info.expr_type;
     String result = info.variable_prefix + "result";
     String func = info.internal_prefix + info.binary_expr_func;
     int level = node.binary_level;
     println (expr_type + " * " + result + " = " + func + " (" + level + ", NULL, NULL);");
  }

  /********************************* RULE *********************************/

  private void code_rule_decl (RuleNode node, boolean separate)
  {
     String parameters = "";

     if (node.mode == MODIFY_MODE)
        current_variable = info.param_prefix + "result";
     else
        current_variable = info.variable_prefix + "result";

     String type_decl = node.type.decl ();

     if (node.mode == MODIFY_MODE)
     {
        parameters = type_decl + " " + current_variable;
     }

     if (node.store_type != null)
     {
        String store_param = info.param_prefix + "store";
        String store_type = node.store_type.decl ();

        if (! empty (parameters))
           parameters +=  ", ";

        parameters +=  store_type + " " +  store_param;
     }

     String type = "";
     if (node.mode == MODIFY_MODE)
        type = "void";
     else
        type = node.type.inner_decl ();

     print (type + " ");

     if (separate)
        print (info.parser_class + "::");

     print (info.rule_prefix);
     print (node.name);
     print (" (" + parameters + ")");
  }

  private void code_rule (RuleNode node)
  {
     code_rule_decl (node, true);
     println (""); // end of line

     println ("{");
     indent ++;

     String type_decl = node.type.decl ();

     if (node.mode == NEW_MODE)
     {
        String type_name = node.type.type_name ();
        println (type_decl + " " + current_variable + " = new " + type_name + ";");
     }

     if (node.mode != NEW_MODE &&
         node.mode != MODIFY_MODE &&
         node.mode != PRIMITIVE_MODE &&
         node.mode != UNARY_MODE &&
         node.mode != BINARY_MODE)
     {
        println (type_decl + " " + current_variable + " = NULL;");
     }

     if (node.mode == MODIFY_MODE)
     {
        println ("assert (" + current_variable + " != NULL);");
     }

     if (node.store_type != null)
     {
        String store_param = info.param_prefix + "store";
        String store_field = info.field_prefix + node.store_name;

        println (current_variable + "->" + store_field + " = " + store_param + ";");
     }

     // contents

     if (node.mode == BINARY_MODE)
     {
        if (info.expr_method)
           code_expr_call (node);
        else
           code_binary (node);
     }
     else if (node.mode == UNARY_MODE)
     {
        code_unary (node);
     }
     else if (node.mode == PRIMITIVE_MODE)
     {
        code_primitive (node);
     }
     else
     {
        code_block (node.normal_block);
     }

     // stencils

     if (! empty (info.replace_type) &&
         // ( node.mode != PRIMITIVE_MODE && node.mode != UNARY_MODE && node.mode != BINARY_MODE ) &&
         node.struct_type != null &&
         (node.struct_type.custom_subst || node.struct_type.custom_replace))
     {
        println ("if (current_stencil != NULL)");
        indent ++;
        println (current_variable + " = " + current_variable + "->" + info.internal_prefix + "replace (this);");
        indent --;
     }

     // end of function

     if (node.mode != MODIFY_MODE &&
         // node.mode != UNARY_MODE &&
         // node.mode != BINARY_MODE &&
         node.mode != PRIMITIVE_MODE )
     {
        println ("return " + current_variable + ";");
     }

     indent --;
     println ("}");
     println ("");
  }

  /***************************** CODE OUTPUT ******************************/

  public void code_output ()
  {
     print_comment ("parser");

     /*
     if (empty (info.parser_super_class))
     {
        println ("class IOException { };");
        println ("");
     }
     */

     if (info.expr_method)
        code_expr_method ();

     RuleNode r = info.first_rule;
     while (r != null)
     {
        code_rule (r); // write parser functions
        r = r.next_rule;
     }

     print_comment ("syntactic predicates");
     r = info.first_rule;
     while (r != null)
     {
        cond_rule (r, true);
        r = r.next_rule;
     }
  }

  public void lexer_code_output ()
  {
     print_comment ("conversion");
     convertSymbols ();

     print_comment ("delimiters");
     codeLiterals ();

     print_comment ("keywords");
     codeTokens ();
  }

  /**************************** HEADER OUTPUT *****************************/

  private void shared_header_output ()
  {
     String symbol_type = info.symbol_type;
     String conv_type = info.conv_type;
     String note_type = info.note_type;
     String location_type = info.location_type;

     String getLocation  = info.internal_prefix + "getLocation";
     String getStartNote = info.internal_prefix + "getStartNote";
     String getStopNote  = info.internal_prefix + "getStopNote";

     boolean lexer_methods;
     if (info.enable_lexer_output)
        lexer_methods = empty (info.lexer_super_class);
     else
        lexer_methods = empty (info.parser_super_class);

     if (lexer_methods)
     {
        String actual        = info.internal_prefix + "actual";

        String symbol        = info.internal_prefix + "symbol";
        String test          = info.internal_prefix + "test";
        String LA            = info.internal_prefix + "LA";
        String match         = info.internal_prefix + "match";
        String next          = info.internal_prefix + "next";
        String stop          = info.internal_prefix + "stop";
        String mark          = info.internal_prefix + "mark";
        String rewind        = info.internal_prefix + "rewind";
        String save_trace    = info.internal_prefix + "save_trace";
        String restore_trace = info.internal_prefix + "restore_trace";

        String mark_type = info.mark_type;

        String TokenSym      = info.internal_prefix + "TokenSym";
        String InpCh         = info.internal_prefix + "InpCh";
        String NextCh        = info.internal_prefix + "NextCh";

        String table = info.param_prefix + "table";
        String pos   = info.variable_prefix + "pos";
        String bit   = info.variable_prefix + "bit";

        println ("private:");
        indent ++;
        println (symbol_type + " " + actual + ";");
        println ("char " + TokenSym + ";");
        println ("char " + InpCh + ";");
        println ("");
        indent --;

        println ("public:");
        indent ++;
        println ("inline bool " + symbol + " (" + symbol_type + " s)"); // test current symbol
        println ("{");
        indent ++;
           println ("return " + actual + " == s;");
        indent --;
        println ("}");
        println ("");

        println ("inline bool " + test + " (const unsigned char " + table + " [])"); // test current symbol with table
        println ("{");
        indent ++;
           println ("int " + pos + " = " + actual + " / 8;");
           println ("int " + bit + " = " + actual + " % 8;");
           println ("return " + table + " [" + pos + "] & (1 << " + bit + ");");
        indent --;
        println ("}");
        println ("");

        println (symbol_type + " " + LA + " (int inx);");
        println ("");

        println ("void " + match + " (" + symbol_type + " s);"); // compare token and read next token
        println ("void " + next + " ();"); // read next token
        println ("");

        println ("void " + stop + " (string msg);");
        println ("void " + stop + " ();"); // stop on unexpected input
        println ("");

        println (mark_type + " " + mark + " ();"); // remember current position
        println ("void " + rewind + " (" + mark_type + " inx);"); // return to previous position
        println ("");

        println ("void " + save_trace + " ();");
        println ("void " + restore_trace + " ();");
        println ("");

        println ("void " + NextCh + " ();");
        println ("");

        indent --;
     }

     println ("public:");
     indent ++;

     String param = info.param_prefix + "param";
     String conv = info.internal_prefix + "conv";
     String delimiter = info.internal_prefix + "delimiter";
     String keyword = info.internal_prefix + "keyword";

     println ("string " +
              conv +
              " (" + conv_type + " " + param + ");");

     println (symbol_type + " " +
              delimiter +
              " ();");

     println (symbol_type + " " +
              keyword +
              " (string " + param + ");");

     if (info.locations)
        println (location_type + " " + getLocation + " ();");

     if (info.notes)
        println (note_type + " " + getStartNote + " ();");

     if (info.notes && !info.only_start_notes)
        println (note_type + " " + getStopNote + " ();");

     println ("");

     // primitive blocks

     Enumeration seq = info.getRules ();
     while (seq.hasMoreElements ())
     {
        RuleInfo item = (RuleInfo) seq.nextElement ();
        RuleNode r = item.node;
        if (r.mode == PRIMITIVE_MODE)
        {
           String id = r.name;
           println ("string " + info.util_func_prefix + "recall_" + id + " ();");
        }
     }

     indent --;
  }

  public void lexer_header_output ()
  {
     // parser module
     // env.parser_module.IncludeFile (incl_intf, "envio.h"); // !? %%

     // symbol type
     if (! info.enable_symbol_header)
        declareSymbols ();

     // lexer class
     print ("class ");
     print (info.lexer_class);
     if (! empty (info.lexer_super_class))
     {
        print (" : public ");
        print (info.lexer_super_class);
     }
     println ("");
     println ("{");
     indent ++;

     shared_header_output ();
     print_extension ("lexer_class");

     // end of class
     indent --;
     println ("};");
     println ("");
  }

  public void header_output ()
  {
     // parser module
     // env.parser_module.IncludeFile (incl_intf, "envio.h"); // !? %%

     // symbol type
     if (! info.enable_symbol_header && ! info.enable_lexer_output)
        declareSymbols ();

     // parser class
     print ("class ");
     print (info.parser_class);
     if (! empty (info.parser_super_class))
     {
        print (" : public ");
        print (info.parser_super_class);
     }
     else if (info.enable_lexer_output)
     {
        print (" : public ");
        print (info.lexer_class);
     }
     println ("");
     println ("{");
     indent ++;

     {
        println ("// parse");
        println ("public:");
        indent ++;

        RuleNode r = info.first_rule;
        while (r != null)
        {
           code_rule_decl (r, false);
           println (";");
           r = r.next_rule;
        }

        indent --;
        println ("");
     }

     {
        // count syntactic predicate functions
        RuleNode r = info.first_rule;
        while (r != null)
        {
           cond_rule (r, false);
           r = r.next_rule;
        }

        if (cond_count > 0)
        {
           println ("// syntactic predicates");
           println ("public:");
           indent ++;
           for (int i = 1; i <= cond_count; i++)
               println ("bool " + info.internal_prefix + "condition_" + i + " ();");
           indent --;
           println ("");
        }
     }

     if (! info.enable_lexer_output)
        shared_header_output ();

     println ("public:");
     indent ++;
     if (! empty (info.replace_type))
        println ("string " + info.internal_prefix + "ident_replace (string name);");
     if (info.expr_method)
        code_expr_decl ();
     println ("");
     indent --;

     print_extension ("parser_class");

     // end of class
     indent --;
     println ("};");
     println ("");
  }

}
