
import java.util.Enumeration;

class RuleNode extends SourceNode
{
   public String name; // rule identifier
   public int mode; // NEW_MODE, MODIFY_MODE, RETURN_MODE, ...
   public TypeInfo type; // result type
   public StructInfo struct_type; // result type for classic rules
   public String variable_name; // identifier of variable with result

   public String store_name; // identifier from store option
   public StructInfo store_type; // type from store option

   public boolean loop;          // used <add> item
   public SequenceInfo loop_ref; // sequence used in this loop

   public RuleNode keep_rule; // sub-rule from leading choose nonterminal
   public boolean choose_compatible_type; // <choose> rule with ()? or ()*

   public boolean rule_ready; // result variable is ready (only for first pass)

   public boolean custom_send; // user defined send_... function
   public boolean custom_no_send; // without send_... function
   public boolean custom_internal_send; // rename send_ ... to internal_send_...
   public boolean custom_open_close_send; // user defined open, close_... function

   /* parser */

   public boolean evaluating;
   public GramSet symbol_set;

   /* normal block */

   public BlockNode normal_block;

   /* binary block */

   public String binary_sub_rule;

   public int binary_level;

   public BinaryNode first_binary_item;
   public BinaryNode last_binary_item;

   public void insertBinaryItem (BinaryNode item)
   {
      if (first_binary_item == null)
         first_binary_item = item;
      else
         last_binary_item.next_binary_item = item;

      last_binary_item = item;
      item.next_binary_item = null;
   }

   /* unary block */

   public String unary_key; // terminal name
   public String unary_selector; // value stored into data structures
   public String unary_sub_rule;

   /* primitive block */

   public PrimitiveNode first_primitive_item; // also in AltNode
   public PrimitiveNode last_primitive_item;

   public void insertPrimitiveItem (PrimitiveNode item)
   {
      if (first_primitive_item == null)
         first_primitive_item = item;
      else
         last_primitive_item.next_primitive_item = item;

      last_primitive_item = item;
      item.next_primitive_item = null;
   }

   /* list of rules */

   public RuleNode next_rule;

   /* variables */

   private DictTable dict = new DictTable ();

   public String addVariable (String name, String type)
   {
      String s = name;
      int i = 0;
      while (dict.contains (s))
      {
         i ++;
         s = name + i;
      }

      VarInfo item = new VarInfo ();

      item.ident = s;
      item.type = type;
      dict.store (item);

      return s;
   }

   public Enumeration getVariables ()
   {
      return dict.elements ();
   }

}
