
import java.util.Enumeration;

class Scanner extends Tool
{
  /* structures, rule names, modes and types */

  private StructInfo defineStruct (String type_name)
  {
     /* find or create type */
     return info.lookupStruct (type_name);
  }

  private void defineSuper (String type_name,
                            String super_name)
  {
     /* find or create type */
     StructInfo t = info.lookupStruct (type_name);

     /* super */
     StructInfo s = info.lookupStruct (super_name);
     if (t.from == null)
     {
        t.from = s;
        s.addSubclass (t);
     }
     else
     {
        if (t.from != s)
           error ("Type " + type_name + " already has super type " + s.ident);
     }
  }

  /* structures */

  public void openStruct(String type_name)
  {
     super.openStruct (type_name);
     defineStruct (type_name);
  }

  public void addSuper (String type_name,
                        String super_name)
  {
     super.addSuper (type_name, super_name);
     defineSuper (type_name, super_name);
  }

  /* rules */

  public void publicOption ()
  {
     super.publicOption ();
     info.private_rules = false;
  }

  public void privateOption ()
  {
     super.privateOption ();
     info.private_rules = true;
  }

  public void openRule (String name)
  {
     super.openRule (name);
     // resetLineNum ();

     // create rule object
     info.storeRule (rule, name);
     rule.name = name;
  }

  public void ruleParam (int mode,
                         String type_name)
  {
     super.ruleParam (mode, type_name);

     // store rule mode
     rule.mode = mode;

     boolean new_struct = (info.getStruct (type_name) != null);

     // store rule type
     StructInfo t = defineStruct (type_name);
     rule.type = t;
     rule.struct_type = t;

     if (new_struct && info.private_rules)
        t.custom_private = true;
  }

  public void closeRule ()
  {
     super.closeRule ();
     if (rule.type == null) error ("Internal error");
  }

  public void addStore (String name, String type_name)
  {
     super.addStore (name, type_name);

     // create type (if not already created)
     rule.store_name = name;
     rule.store_type = defineStruct (type_name);
     // necessary before Decl, because Decl need store_type field
  }

  /* binary */

  public void openBinary ()
  {
     super.openBinary ();
     rule.mode = BINARY_MODE;
     rule.struct_type = defineStruct (info.expr_type);
     rule.type = rule.struct_type;
     defineStruct (info.binary_expr_type);
     defineSuper  (info.binary_expr_type, info.expr_type);
  }

  /* unary */

  public void unary (String key, String selector_value, String branch)
  {
     super.unary (key, selector_value, branch);
     rule.mode = UNARY_MODE;
     defineStruct (info.expr_type);
     rule.struct_type = defineStruct (info.unary_expr_type);
     rule.type = rule.struct_type;
     defineSuper (info.unary_expr_type, info.expr_type);
  }

  /* primitive */

  public void openPrimitive ()
  {
     super.openPrimitive ();
     rule.mode = PRIMITIVE_MODE;
     rule.type = info.string_type; // string
  }

  public void identifierOption ()
  {
     super.identifierOption ();
     rule.type = info.ident_type; // identifier
  }

  /* options */

  private String checkIdent (String value)
  {
     if (! isIdentifier (value))
        error ("Identifier expected: " + value);
     return value;
  }

  private String checkIdentOrEmpty (String value)
  {
     if (! empty (value) && value.charAt (0) == '"' && value.length () == 2)
        return "";
     else
        return checkIdent (value);
  }

  private boolean checkBoolean (String value)
  {
     if (! value.equals ("true") && ! value.equals ("false"))
        error ("Boolean constatnt expected: " + value);

     return value.equals ("true");
  }

  private String checkString (String value)
  {
     if (empty (value) || value.charAt (0) != '"')
        error ("String expected: " + value);

     value = value.substring (1, value.length()-1);

     /*
     // search for escapes
     boolean done = false;
     while (! done)
     {
        int inx = value.indexOf ('\\');
        if (inx >= 0)
           value = value.substring (0, inx) + value.substring (inx+1);
        else
           done = true;
     }
     */

     return value;
  }

  public void option (String name, String value)
  {
     super.option (name, value);

     /* file names */

     if (name.equals ("common_header"))
        info.common_header = checkIdent (value);

     else if (name.equals ("common_name"))
        info.common_name = checkIdent (value);

     else if (name.equals ("public_header"))
        info.public_header = checkIdent (value);

     else if (name.equals ("private_header"))
        info.private_header = checkIdent (value);

     else if (name.equals ("symbol_header"))
        info.symbol_header = checkIdent (value);

     else if (name.equals ("lexer_name"))
        info.lexer_name = checkIdent (value);

     else if (name.equals ("parser_name"))
        info.parser_name = checkIdent (value);

     else if (name.equals ("product_name"))
        info.product_name = checkIdent (value);

     else if (name.equals ("lexer_input"))
        info.lexer_input = checkString (value);

     else if (name.equals ("product_input"))
        info.product_input = checkString (value);

     /* namespace and class names */

     else if (name.equals ("product_namespace"))
     {
        info.product_namespace = checkIdent (value);
        grammarOption ("namespace", info.product_namespace + "::");
     }

     else if (name.equals ("automatic_namespace"))
     {
        info.automatic_namespace = checkBoolean (value);
     }

     else if (name.equals ("lexer_class"))
        info.lexer_class = checkIdent (value);

     else if (name.equals ("lexer_super_class"))
        info.lexer_super_class = checkIdent (value);

     else if (name.equals ("parser_class"))
        info.parser_class = checkIdent (value);

     else if (name.equals ("parser_super_class"))
        info.parser_super_class = checkIdent (value);

     else if (name.equals ("product_class"))
        info.product_class = checkIdent (value);

     else if (name.equals ("product_super_class"))
        info.product_super_class = checkIdent (value);

     /* options */

     else if (name.equals ("tree_code"))
        info.tree_code = checkBoolean (value);

     else if (name.equals ("expr_method"))
        info.expr_method = checkBoolean (value);

     else if (name.equals ("copy_method"))
        info.copy_method = checkBoolean (value);

     else if (name.equals ("rename_method"))
        info.rename_method = checkBoolean (value);

     else if (name.equals ("notes"))
        info.notes = checkBoolean (value);

     else if (name.equals ("only_start_notes"))
        info.only_start_notes = checkBoolean (value);

     else if (name.equals ("fixed_notes"))
        info.fixed_notes = checkBoolean (value);

     else if (name.equals ("distinct_notes"))
        info.distinct_notes = checkBoolean (value);

     else if (name.equals ("locations"))
        info.locations = checkBoolean (value);

     else if (name.equals ("ignore_notes"))
        info.ignore_notes = checkBoolean (value);

     else if (name.equals ("ignore_locations"))
        info.ignore_locations = checkBoolean (value);

     else if (name.equals ("note_type"))
        info.note_type = checkIdent (value);

     else if (name.equals ("location_type"))
        info.location_type = checkIdent (value);

     else if (name.equals ("mark_type"))
        info.mark_type = checkIdent (value);

     else if (name.equals ("note_init"))
        info.note_init = checkString  (value);

     else if (name.equals ("location_init"))
        info.location_init = checkString  (value);

     else if (name.equals ("anonymous_fields"))
        info.anonymous_fields = checkBoolean (value);

     else if (name.equals ("ignore_send_options"))
        info.ignore_send_options = checkBoolean (value);

     else if (name.equals ("ignore_output_options"))
        info.ignore_output_options = checkBoolean (value);

     else if (name.equals ("ignore_execute_options"))
        info.ignore_execute_options = checkBoolean (value);

     else if (name.equals ("ignore_product_constructor"))
        info.ignore_product_constructor = checkBoolean (value);

     else if (name.equals ("ignore_tag_marks"))
        info.ignore_tag_marks = checkBoolean (value);

     else if (name.equals ("antlr_mode"))
        info.antlr_mode = checkBoolean (value);

     else if (name.equals ("subst_type"))
        info.subst_type = checkIdent (value);

     else if (name.equals ("replace_type"))
        info.replace_type = checkIdent (value);

     /* identifier prefixes */

     else if (name.equals ("rule_prefix"))
        info.rule_prefix = checkIdent (value);

     else if (name.equals ("type_prefix"))
        info.type_prefix = checkIdent (value);

     else if (name.equals ("field_prefix"))
        info.field_prefix = checkIdent (value);

     else if (name.equals ("list_prefix"))
        info.list_prefix = checkIdent (value);

     else if (name.equals ("internal_prefix"))
        info.internal_prefix = checkIdent (value);

     else if (name.equals ("send_func_prefix"))
        info.send_func_prefix = checkIdent (value);

     else if (name.equals ("send_internal_prefix"))
        info.send_internal_prefix = checkIdent (value);

     else if (name.equals ("open_func_prefix"))
        info.open_func_prefix = checkIdent (value);

     else if (name.equals ("close_func_prefix"))
        info.close_func_prefix = checkIdent (value);

     else if (name.equals ("conv_func_prefix"))
        info.conv_func_prefix = checkIdent (value);

     else if (name.equals ("util_func_prefix"))
        info.util_func_prefix = checkIdent (value);

     else if (name.equals ("param_prefix"))
        info.param_prefix = checkIdent (value);

     else if (name.equals ("temp_prefix"))
        info.temp_prefix = checkIdent (value);

     else if (name.equals ("local_prefix"))
        info.local_prefix = checkIdent (value);

     else if (name.equals ("variable_prefix"))
        info.variable_prefix = checkIdent (value);

     else if (name.equals ("literal_prefix"))
        info.literal_prefix = checkIdent (value);

     /* symbol type */

     else if (name.equals ("symbol_type"))
        info.symbol_type = checkIdent (value);

     else if (name.equals ("conv_type"))
        info.conv_type = checkIdent (value);

     else if (name.equals ("no_symbol"))
        info.no_symbol = checkIdent (value);

     else if (name.equals ("primitive_prefix"))
        info.primitive_prefix = checkIdent (value);

     else if (name.equals ("delimiter_prefix"))
        info.delimiter_prefix = checkIdent (value);

     else if (name.equals ("keyword_prefix"))
        info.keyword_prefix = checkIdent (value);

     /* binary and unary expressions */

     else if (name.equals ("basic_type"))
        info.basic_type = checkIdent (value);

     else if (name.equals ("expr_type"))
        info.expr_type = checkIdent (value);

     else if (name.equals ("binary_expr_type"))
        info.binary_expr_type = checkIdent (value);

     else if (name.equals ("binary_expr_left"))
        info.binary_expr_left = checkIdent (value);

     else if (name.equals ("binary_expr_right"))
        info.binary_expr_right = checkIdent (value);

     else if (name.equals ("binary_expr_middle"))
        info.binary_expr_middle = checkIdent (value);

     else if (name.equals ("binary_expr_selector"))
        info.binary_expr_selector = checkIdent (value);

     else if (name.equals ("binary_expr_func"))
        info.binary_expr_func = checkIdent (value);

     else if (name.equals ("unary_expr_type"))
        info.unary_expr_type = checkIdent (value);

     else if (name.equals ("unary_expr_param"))
        info.unary_expr_param = checkIdent (value);

     else if (name.equals ("unary_expr_selector"))
        info.unary_expr_selector = checkIdent (value);

     else error ("Unknown option " + name);
  }

  /* custom */

  public void addCustom (String key)
  {
     super.addCustom (key);

     if (key.equals ("send"))
     {
        // modify rule
        if (! info.ignore_send_options)
           rule.custom_send = true;
     }
     else if (key.equals ("no_send"))
     {
        if (! info.ignore_send_options)
           rule.custom_no_send = true;
     }
     else if (key.equals ("internal_send"))
     {
        if (! info.ignore_send_options)
           rule.custom_internal_send = true;
     }
     else if (key.equals ("open_close_send"))
     {
        if (! info.ignore_send_options)
           rule.custom_open_close_send = true;
     }
     else if (key.equals ("output"))
     {
        if (! info.ignore_output_options)
           rule.custom_send = true;
        // "output" is similar to "send"
        // "send" is controlled by ignore_send_options
        // "output" is controlled by ignore_output_options
     }
     else if (key.equals ("internal_output"))
     {
        if (! info.ignore_output_options)
           rule.custom_internal_send = true;
     }
     else if (key.equals ("conv") ||
              key.equals ("subst") ||
              key.equals ("list_subst") ||
              key.equals ("replace") ||
              key.equals ("list_replace") ||
              key.equals ("produce") ||
              key.equals ("no_produce") ||
              key.equals ("private") )
     {
        if (! info.ignore_send_options)
        {
           // modify structure type
           StructInfo t = rule.struct_type;
           if (t == null) error ("Invalid custom key: " + key);

           if (key.equals ("conv"))
              t.custom_conv = true;
           else if (key.equals ("subst"))
              t.custom_subst = true;
           else if (key.equals ("list_subst"))
              t.custom_list_subst = true;
           else if (key.equals ("replace"))
              t.custom_replace = true;
           else if (key.equals ("list_replace"))
              t.custom_list_replace = true;
           else if (key.equals ("produce"))
              t.custom_produce = true;
           else if (key.equals ("no_produce"))
             t.custom_no_produce = true;
           else if (key.equals ("private"))
             t.custom_private = true;
         }
     }
     else error ("Unknown custom key: " + key);
  }

  public void extension (String name, String value)
  {
     super.extension (name, value);

     if (value.startsWith ("@@"))
        value = value.substring (2);

     if (value.endsWith ("@@"))
        value = value.substring (0, value.length () - 2);

    if (! name.equals ("lexer_include") &&
        ! name.equals ("lexer_header") &&
        ! name.equals ("lexer_class") &&

        ! name.equals ("parser_include") &&
        ! name.equals ("parser_header") &&
        ! name.equals ("parser_class") &&

        ! name.equals ("product_include") &&
        ! name.equals ("product_header") &&
        ! name.equals ("product_class") &&

        ! name.equals ("common_include") &&
        ! name.equals ("common_header") &&
        ! name.equals ("common_class") &&
        ! name.equals ("common_tail") )

       error ("Unknown extension key: " + name);

     DictItem2 item = new DictItem2 ();
     item.ident = name;
     item.value = value;
     info.extension_repository.store (item);
  }

  public void onInit ()
  {
     super.onInit ();
  }

  public void onFinish ()
  {
     super.onFinish ();
  }
}
