
import java.util.Vector;
import java.util.Enumeration;

class StructInfo extends TypeInfo
{
  public StructInfo from; // super class

  public int note_count; // number of variables for comments

  // public StructInfo sequence; // type of elemets in sequence
  // public boolean sequence_element; // true => this type is element of some sequence

  public boolean custom_conv;         // true => create conversion function in basic type
  public boolean custom_produce;      // true => user defined produce method
  public boolean custom_no_produce;   // true => without produce_attributes, produce_sequence
  public boolean custom_subst;        // true => user defined find_subst and produce_subst methods
  public boolean custom_list_subst;   // true => user defined produce_list_subst method
  public boolean custom_replace;      // true => user defined replace method
  public boolean custom_list_replace; // true => user defined list_replace method
  public boolean custom_private;      // true => place class declaration into private header file

  public boolean declared; // true => declaration already written
  public boolean get_first_loc; // true => class has getFirstLocation method

  public boolean create_func; // true => constructor with parameters
  public static final int param_limit = 3; // max. number of parameters

  /* fields */

  private DictTable fields = new DictTable ();

  public FieldInfo getField (String field_name)
  {
    return (FieldInfo) fields.get (field_name);
  }

  public FieldInfo findField (String field_name)
  {
     FieldInfo f = getField (field_name);
     if (f == null)
        err ("Undefined field: " + field_name);
     return f;
  }

  public FieldInfo createField (String field_name)
  {
    if (getField (field_name) != null)
       err ("Field already defined: " + field_name);

    FieldInfo fld = new FieldInfo ();
    fld.ident = field_name;

    fields.store (fld);
    return fld;
  }

  public FieldInfo lookupField (String field_name)
  {
     FieldInfo f = getField (field_name);

     if (f == null && from != null)
        f = from.lookupField (field_name);

     return f;
  }

  public Enumeration getFields ()
  {
    return fields.elements ();
  }

  public int getFieldCount ()
  {
    return fields.size ();
  }

  /* sequences */

  private DictTable sequences = new DictTable ();

  public SequenceInfo getSequence (String sequence_name)
  {
    return (SequenceInfo) sequences.get (sequence_name);
  }

  /*
  public SequenceInfo findSequenece (String sequences_name)
  {
     SequenceInfo f = getSequence (sequence_name);
     if (f == null)
        err ("Undefined sequence: " + sequence_name);
     return f;
  }
  */

  public SequenceInfo createSequence (String sequence_name)
  {
    if (getSequence (sequence_name) != null)
       err ("Sequence already defined: " + sequence_name);

    SequenceInfo fld = new SequenceInfo ();
    fld.ident = sequence_name;

    sequences.store (fld);
    return fld;
  }

  public SequenceInfo lookupSequence (String sequence_name)
  {
     SequenceInfo f = getSequence (sequence_name);

     if (f == null && from != null)
        f = from.lookupSequence (sequence_name);

     return f;
  }

  public Enumeration getSequences ()
  {
    return sequences.elements ();
  }

  public boolean hasSequences ()
  {
    return ! sequences.empty ();
  }

  /* sequence anchors */

  private DictTable anchors = new DictTable ();

  public void addAnchor (SequenceInfo sequence)
  {
    anchors.store (sequence);
  }

  public Enumeration getAnchors ()
  {
    return anchors.elements ();
  }

  public boolean hasAnchors ()
  {
    return ! anchors.empty ();
  }

  public boolean hasAnchor (String id)
  {
    return anchors.contains (id);
  }

  /* subclasses */

  private Vector subclasses = new Vector ();

  public void addSubclass (StructInfo type)
  {
    subclasses.add (type);
  }

  public Enumeration getSubclasses ()
  {
    return subclasses.elements ();
  }

  public boolean hasSubclasses ()
  {
    return ! subclasses.isEmpty ();
  }

  public boolean is_derived_from (TypeInfo type)
  {
     StructInfo s = this;
     while (s != type && s != null)
        s = s.from;
     return (s == type);
  }

  /* parameters */

  private Vector params = new Vector ();

  public void addParam (FieldInfo f)
  {
    params.add (f);
  }

  public void copyParams (StructInfo s)
  {
    params.addAll (s.params);
  }

  public FieldInfo getParam (int inx)
  {
    return (FieldInfo) params.get (inx);
  }

  public int getParamCount ()
  {
    return params.size ();
  }

  /* methods */

  private DictTable methods = new DictTable ();

  public void addMethod (MethodInfo m)
  {
    methods.store (m);
  }

  public boolean hasMethod (String id)
  {
    return methods.contains (id);
  }

  public Enumeration getMethods ()
  {
    return methods.elements ();
  }

  /* places */

  private DictTable places = new DictTable ();

  public FieldInfo createPlace (String name)
  {
    FieldInfo fld = new FieldInfo ();
    fld.ident = name;

    places.store (fld);
    return fld;
  }

  public Enumeration getPlaces ()
  {
    return places.elements ();
  }

  public int getPlaceCount ()
  {
    return places.size ();
  }

}
