
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

class TreeOutput extends Tool
{

  private void check (String expr)
  {
    println ("assert (" + expr + ");");
  }

  /* enumerations */

  private String decl_conv (TypeInfo t)
  {
     return "string " + info.conv_func_prefix + t.ident +
            " (" + t.inner_decl () + " " + info.param_prefix + "item" + ")";
  }

  private void conv_enumeration (EnumInfo e)
  {
     println (decl_conv (e));
     println ("{");
     indent ++;

     println ("switch (" + info.param_prefix+ "item)");
     println ("{");
     indent ++;

     Enumeration items = e.getItems ();
     while (items.hasMoreElements ())
     {
        DictItem item = (DictItem) items.nextElement ();

        println ("case " + item.ident + ":");
        indent ++;
        println ("return " + '"' + item.ident + '"' + "; ");
        println ("break;");
        indent --;
     }

     println ("default: ");
     indent ++;
     check ("false");
     indent --;

     indent --;
     println ("}");

     indent --;
     println ("}");
     println ("");
  }

  /* structure fields */

  private void produce_attributes (StructInfo s)
  {
     if (s.custom_no_produce) println ("/*");

     println ("void " + s.type_name () + "::" +
              info.internal_prefix + "produce_attributes " +
              "(" + info.writer_decl + ")");
     println ("{");
     indent ++;

     if (s.note_count > 0 && info.locations)
     {
        String loc;
        if (info.fixed_notes)
           loc = info.internal_prefix + "location_1";
        else
           loc = info.internal_prefix + "location [0]";


        println (info.writer_param + "->" +
                 info.internal_prefix + "store_location (" + loc + ");");
     }

     // super class
     if (s.from != null)
        println (s.from.type_name () + "::" +
                 info.internal_prefix + "produce_attributes (" + info.writer_param + ");");

     // fields
     Enumeration fields = s.getFields ();
     while (fields.hasMoreElements ())
     {
        FieldInfo f = (FieldInfo) fields.nextElement ();

        print (info.writer_param);
        print ("->");
        print (info.util_func_prefix);
        if (f.type instanceof EnumInfo)
        {
           /*
           print ("if (");
           print ("this->" + info.field_prefix + f.ident);
           print (" != ");
           print (f.type.type_name () + " (0)");
           println (") ");
           indent ++;
           print ("store_enum");
           indent --;
           */
           print ("store_enum");
        }
        else if (f.type instanceof StructInfo)
        {
           print ("store_item");
        }
        else if (f.type == info.ident_type)
        {
           print ("store_ident");
        }
        else
        {
           print ("store_" + f.type.ident);
        }

        print (" (");

        print ('"' + f.ident + '"');
        print (", ");

        if (f.type instanceof EnumInfo)
        {
           print (info.conv_func_prefix + f.type.ident +
                  " (this->" + info.field_prefix + f.ident + ")");
           print (", ");
           print ("this->" + info.field_prefix + f.ident);
        }
        else
        {
           print ("this->" + info.field_prefix + f.ident);
        }

        println (");");
     }

     indent --;
     println ("}");
     if (s.custom_no_produce) println ("*/");
     println ("");

  }

  /* sequences */

  private void produce_sequence (SequenceInfo t, int cnt)
  {
     if (cnt > 1) println ("");

     String subitem = info.variable_prefix + "subitem";
     if (cnt > 1) subitem = subitem + cnt;

     println (t.element.inner_decl() + " " + subitem +
              " = this->" + info.list_prefix + "first" + t.suffix + ";");

     println ("while (" + subitem + " != NULL)");
     println ("{");
     indent ++;

     println (info.writer_param + "->" +
              info.util_func_prefix + "insert_item (" + subitem + ");");
     println (subitem + " = " + subitem + "->" + info.list_prefix + "next" + t.suffix + ";");

     indent --;
     println ("}");
  }

  private void produce_sequences (StructInfo s)
  {
     if (s.custom_no_produce) println ("/*");

     println ("void " + s.type_name () + "::" +
              info.internal_prefix + "produce_sequence " +
              "(" + info.writer_decl + ")");
     println ("{");
     indent ++;

     int cnt = 0;
     Enumeration sequences = s.getSequences ();
     while (sequences.hasMoreElements ())
     {
        SequenceInfo t = (SequenceInfo) sequences.nextElement ();
        cnt ++;
        produce_sequence (t, cnt);
     }

     indent --;
     println ("}");
     if (s.custom_no_produce) println ("*/");
     println ("");
  }

  /* substitutions */

  private void recursive_find_subst (StructInfo s)
  {
     println ("void " + s.type_name () + "::" +
              info.internal_prefix + "recursive_find_subst (" +
              info.writer_decl + ") ");
     println ("{");
     indent ++;

     // fields

     Enumeration fields = s.getFields ();
     while (fields.hasMoreElements ())
     {
        FieldInfo f = (FieldInfo) fields.nextElement ();

        // !? only real fields
        // if (f.is_real)
        {
           if (f.type instanceof StructInfo)
           {
              println ("if (" + info.field_prefix + f.ident + " != NULL)");
              indent ++;
              println (info.field_prefix + f.ident + "->" +
                       info.internal_prefix + "recursive_find_subst (" + info.writer_param + ");");
              indent --;
           }
        }
     }

     // sequences

     int cnt = 0;
     Enumeration sequences = s.getSequences ();
     while (sequences.hasMoreElements ())
     {
        sequences.nextElement ();
        cnt ++;
     }

     sequences = s.getSequences ();
     while (sequences.hasMoreElements ())
     {
        SequenceInfo t = (SequenceInfo) sequences.nextElement ();

        if (cnt > 1)
        {
           println ("{");
           indent ++;
        }

        println (info.type_prefix + t.element.ident + " * item = " + info.list_prefix + "first;");
        println ("while (item != NULL)");
        println ("{");
           indent ++;
           println ("item->" + info.internal_prefix + "recursive_find_subst (" + info.writer_param + ");");
           println ("item = item->" + info.list_prefix + "next;");
           indent --;
        println ("}");

        if (cnt > 1)
        {
           indent --;
           println ("}");
        }
     }

     // after subitems
     // println (info.internal_prefix + "find_subst (" + info.writer_param + ");");

     // super class
     String super_type = info.basic_type;
     if (s.from != null)
        super_type = s.from.type_name ();

     println (super_type + "::" +
              info.internal_prefix + "recursive_find_subst (" + info.writer_param + ");");

     indent --;
     println ("}");
     println ("");
  }

  /* local headers */

  public void local_headers ()
  {
     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof EnumInfo)
        {
           println (decl_conv (t) + ";");
        }
     }
     println ("");
  }

  /* code */

  public void sendCode ()
  {
     print_comment ("conv decl");

     local_headers ();

     print_comment ("subst");

     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof StructInfo)
        {
           StructInfo s = (StructInfo) t;

           if (s.getFieldCount () != 0 || s.hasSequences ())
              recursive_find_subst (s);
        }
     }

     print_comment ("tree");

     seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof StructInfo)
        {
           StructInfo s = (StructInfo) t;

           if (s.getFieldCount () != 0)
              produce_attributes (s);

           if (s.hasSequences ())
              produce_sequences (s);
        }
     }

     print_comment ("conv");

     seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof EnumInfo)
        {
           EnumInfo e = (EnumInfo) t;
           conv_enumeration (e);
        }
     }
  }

}
