
options
{
  language=Java;
}

class EGLexer extends Lexer;

options
{
  k=2;
  exportVocab=EG1;
  charVocabulary = '\3'..'\377';
  testLiterals=false;
  // defaultErrorHandler=true;
  // do not use default error handler in ANTLR 2.7.5 lexer
}

tokens
{
   "struct";
   "enum";
   "literals";
   "external_tokens";
   "options";
   "grammar_options";
   "lexer_options";
   "parser_options";

   "extension";

   "parser";
   "public";
   "private";

   "new";
   "modify";
   "return";
   "select";
   "choose";
   "binary";
   "unary";
   "primitive";

   "store";
   "custom";
   "identifier";

   "add";
   "call";
   "result";

   "set";
   "inc";
   "silent";
   "additional";
   "invoke";
   "execute";
   "redudant";
   "location";

   "space";
   "no_space";
   "new_line";
   "empty_line";
   "no_empty_line";
   "indent";
   "unindent";
}

{
   boolean inside_block;
   /* keywords are disabled inside { ... } */

   boolean inside_class;
   /* keywords are disabled after "class" */

   boolean inside_option;
   /* keywords are temporary enabled at the beginng of < ... > option */

   public int testLiteralsTable (int ttype)
   {
      if ((! inside_class && ! inside_block) || inside_option)
      {
         ttype = super.testLiteralsTable (ttype);

         if (ttype == LITERAL_parser)
            inside_class = true;
            // disable keywords after "class" keyword

         if (inside_option)
            inside_option = false;
            // disable keywords after first identifier in < ... > option
      }
      return ttype;
   }

}

OPEN_OPTION  : '<' { inside_option = true;  };
CLOSE_OPTION : '>' { inside_option = false; };

EXT_OPEN_OPTION  : "<:" { inside_option = true;  };
EXT_CLOSE_OPTION : ":>" { inside_option = false; };

LCURLY : '{' { inside_block = true; };
RCURLY : '}' { inside_block = false; };

WS  :
    (
      options { generateAmbigWarnings=false; }  :

        ' '
      | '\t'
      | '\r' '\n' {newline();}
      | '\r'    {newline();}
      | '\n'    {newline();}
    )
    { $setType(Token.SKIP); }
    ;

COMMENT :
  ( SL_COMMENT | ML_COMMENT) {$setType(Token.SKIP);}
  ;

protected
SL_COMMENT :
  "//"
  ( ~('\n'|'\r') )*
  (
    options { generateAmbigWarnings=false; } :

      '\r' '\n'
    | '\r'
    | '\n'
  )
  { newline(); }
  ;

protected
ML_COMMENT :
  "/*"
  (
    options
    {
      greedy=false;  // make it exit upon "*/"
      generateAmbigWarnings=false; // shut off newline errors
    } :

      '\r' '\n' { newline(); }
    | '\r'      { newline(); }
    | '\n'      { newline(); }
    | ~('\n'|'\r')
  )*
  "*/"
  ;

LPAREN: '(' ;

RPAREN: ')' ;

LBRACKET: '[' ;

RBRACKET: ']' ;

QUESTION :  '?' ;

STAR: '*' ;

PLUS: '+' ;

AT: '@' ;

INSERT: "<-" ;

IMPL: "=>" ;

ASSIGN: '=' ;

OR: '|' ;

SEMI: ';' ;

COMMA: ',' ;

COLON: ':' ;

CHAR_LITERAL:
  '\'' ( ~ '\'' )* '\''
  ;

STRING_LITERAL:
  '"' ( ~ '"' )* '"'
  ;

NUMBER  :
  ('0'..'9')+
  ;

NAME
  options { testLiterals = true; }
  :
  ('a'..'z' | 'A'..'Z' | '_')
  ( // scarf as many letters/numbers as you can
    options { warnWhenFollowAmbig=false; } :
    'a'..'z'|'A'..'Z'|'_'|'0'..'9'
  )*
  ;

CODE :
  "@@"
  (
    options
    {
      greedy=false;  // make it exit upon "@@"
      generateAmbigWarnings=false; // shut off newline errors
    } :

      '\r' '\n' { newline(); }
    | '\r'      { newline(); }
    | '\n'      { newline(); }
    | ~('\n'|'\r')
  )*
  "@@"
  ;
