
options
{
  language=Java;
}

class EGParser extends Parser;
options
{
  k=2;
  importVocab=EG1;
  exportVocab=EG2;
  defaultErrorHandler=false;
  // do not use default error handler in ANTLR 2.7.5 lexer
  buildAST=true;
}

tokens
{
  STRUCT_DEF;
  FIELD_DEF;
  SEQUENCE_DEF;
  POINTER_DEF;

  ENUM_DEF;

  LITERALS_DEF;
  LITERAL_DEF;

  EXTERNAL_TOKENS_DEF;
  EXTERNAL_TOKEN_DEF;

  OPTIONS_DEF;
  GRAMMAR_OPTIONS_DEF;
  LEXER_OPTIONS_DEF;
  PARSER_OPTIONS_DEF;
  OPTION_DEF;
  EXTENSION_DEF;

  RULES_DEF;
  PUBLIC_DEF;
  PRIVATE_DEF;

  RULE_DEF;
  PARAM_DEF;
  SUPER_DEF;

  BINARY_DEF;
  BINARY_ITEM_DEF;
  UNARY_DEF;
  PRIMITIVE_DEF;
  PRIMITIVE_ITEM_DEF;
  IDENTIFIER_DEF;

  BLOCK_DEF;
  EBNF_DEF;
  SEMANTIC_PREDICATE_DEF;
  PARENTHESIS_DEF;
  SYNTACTIC_PREDICATE_DEF;
  ALTERNATIVE_DEF;
  ALTERNATIVE_ITEM_DEF;
  TERMINAL_DEF;
  NONTERMINAL_DEF;

  CUSTOM_DEF;
  STORE_DEF;

  LABEL_DEF;
  INSERT_DEF;
  ADD_DEF;
  ASSIGN_DEF;
  CALL_DEF;
  RESULT_DEF;

  SET_DEF;
  INC_DEF;
  SILENT_DEF;
  ADDITIONAL_DEF;

  INVOKE_DEF;
  EXECUTE_DEF;
  REDUDANT_DEF;
  LOCATION_DEF;
  STYLE_DEF;
}

/* ---------------------------------------------------------------------- */

grammar
  :
    (
      structDef |
      enumDef |
      literalsDef |
      externalTokensDef |
      optionsDef |
      grammarOptionsDef |
      lexerOptionsDef |
      parserOptionsDef |
      extensionDef
    )*

    rules

    EOF !
  ;

/* ---------------------------------------------------------------------- */

id
  :
    NAME
  ;

string_value
  :
    CHAR_LITERAL | STRING_LITERAL
  ;

simple_value
  :
    // NAME | NUMBER | CHAR_LITERAL | STRING_LITERAL
    NAME
  ;

/* ---------------------------------------------------------------------- */

structDef
  :
    "struct" !
    param

    LCURLY !
    ( fieldDef | sequenceDef )*
    RCURLY !

    ( SEMI ! )?

    { ## = #( #[STRUCT_DEF, "STRUCT_DEF"], ##); }
  ;

param
  :
    id
    ( superDef )?
    { ## = #( #[PARAM_DEF, "PARAM_DEF"], ##); }
  ;

superDef
  :
    COLON! id
    { ## = #( #[SUPER_DEF, "SUPER_DEF"], ##); }
  ;

fieldDef
  :
    id ( pointerDef )? id SEMI!
    { ## = #( #[FIELD_DEF, "FIELD_DEF"], ##); }
  ;

pointerDef
  :
    STAR!
    { ## = #( #[POINTER_DEF, "POINTER_DEF"], ##); }
  ;

sequenceDef
  :
    id LBRACKET! RBRACKET! id SEMI!
    { ## = #( #[SEQUENCE_DEF, "SEQUENCE_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

enumDef
  :
    "enum" !
    id
    LCURLY !
    ( id ( COMMA ! id )* )?
    RCURLY !
    ( SEMI ! )?

    { ## = #( #[ENUM_DEF, "ENUM_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

literalsDef
  :
    "literals" !
    LCURLY !
    ( literalDef )*
    RCURLY !
    ( SEMI ! )?

    { ## = #( #[LITERALS_DEF, "LITERALS_DEF"], ##); }
  ;

literalDef
  :
    id COLON! string_value ( SEMI! )?
    { ## = #( #[LITERAL_DEF, "LITERAL_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

externalTokensDef
  :
    "external_tokens" !
    LCURLY !
    ( externalTokenDef )*
    RCURLY !
    ( SEMI ! )?

    { ## = #( #[EXTERNAL_TOKENS_DEF, "EXTERNAL_TOKENS_DEF"], ##); }
  ;

externalTokenDef
  :
    string_value ( SEMI! )?
    { ## = #( #[EXTERNAL_TOKEN_DEF, "EXTERNAL_TOKEN_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

optionsDef
  :
    "options" !
    optionsBody
    { ## = #( #[OPTIONS_DEF, "OPTIONS_DEF"], ##); }
  ;

grammarOptionsDef
  :
    "grammar_options" !
    optionsBody
    { ## = #( #[GRAMMAR_OPTIONS_DEF, "GRAMMAR_OPTIONS_DEF"], ##); }
  ;

lexerOptionsDef
  :
    "lexer_options" !
    optionsBody
    { ## = #( #[LEXER_OPTIONS_DEF, "LEXER_OPTIONS_DEF"], ##); }
  ;

parserOptionsDef
  :
    "parser_options" !
    optionsBody
    { ## = #( #[PARSER_OPTIONS_DEF, "PARSER_OPTIONS_DEF"], ##); }
  ;

extensionDef
  :
    "extension" !
    id
    LCURLY !
    CODE
    RCURLY !
    ( SEMI ! )?
    { ## = #( #[EXTENSION_DEF, "EXTENSION_DEF"], ##); }
  ;

optionsBody
  :
    LCURLY !
    ( optionDef )*
    RCURLY !
    ( SEMI ! )?
  ;

optionDef
  :
    id
    ASSIGN !
    optionValue
    SEMI !
    { ## = #( #[OPTION_DEF, "OPTION_DEF"], ##); }
  ;

optionValue
  :
    NAME
    | NUMBER
    | STRING_LITERAL
    // | CHAR_LITERAL
  ;

/* ---------------------------------------------------------------------- */

rules
  :
    "parser" !
    ( SEMI ! )?
    ( public_option | private_option | rule )*
    { ## = #( #[RULES_DEF, "RULES_DEF"], ##); }
  ;

public_option
  :
    OPEN_OPTION !
    "public" !
    CLOSE_OPTION !
    { ## = #( #[PUBLIC_DEF, "PUBLIC_DEF"], ##); }
  ;

private_option
  :
    OPEN_OPTION !
    "private" !
    CLOSE_OPTION !
    { ## = #( #[PRIVATE_DEF, "PRIVATE_DEF"], ##); }
  ;

rule
  :
    id
    (
      normal_cont |
      binary_block |
      unary_block |
      primitive_block
    )
    { ## = #( #[RULE_DEF, "RULE_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

normal_cont
  :
    (
       OPEN_OPTION !
       ( "new" | "modify" | "return" | "select" | "choose" )?
       param
       CLOSE_OPTION !

       ( custom )*
    )?

    COLON !

    ( store )?
    block

    SEMI !
  ;

custom
  :
    OPEN_OPTION !
    "custom" !
    id
    ( COMMA! id )*
    CLOSE_OPTION !
    { ## = #( #[CUSTOM_DEF, "CUSTOM_DEF"], ##); }
  ;

store
  :
    OPEN_OPTION !
    "store" !
    id
    COLON !
    id
    CLOSE_OPTION !
    { ## = #( #[STORE_DEF, "STORE_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

binary_block
  :
    OPEN_OPTION !
    "binary" !
    CLOSE_OPTION !

    ( custom )*

    COLON !

    NAME
    binary_alternative
    ( COMMA ! binary_alternative )*

    SEMI !
    { ## = #( #[BINARY_DEF, "BINARY_DEF"], ##); }
  ;

binary_alternative
  :
    string_value
    NAME
    ( LPAREN ! NAME string_value RPAREN ! )?
    { ## = #( #[BINARY_ITEM_DEF, "BINARY_ITEM_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

unary_block
  :
    OPEN_OPTION !
    "unary" !
    CLOSE_OPTION !

    ( custom )*

    COLON !

    string_value
    NAME
    NAME

    SEMI !
    { ## = #( #[UNARY_DEF, "UNARY_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

primitive_block
  :
    OPEN_OPTION !
    "primitive" !
    CLOSE_OPTION !

    ( identifier_option )?
    ( custom )*

    COLON !

    primitive_alternative
    ( OR ! primitive_alternative )*

    SEMI !
    { ## = #( #[PRIMITIVE_DEF, "PRIMITIVE_DEF"], ##); }
  ;

primitive_alternative :
    NAME
    { ## = #( #[PRIMITIVE_ITEM_DEF, "PRIMITIVE_ITEM_DEF"], ##); }
  ;

identifier_option
  :
    OPEN_OPTION !
    "identifier" !
    CLOSE_OPTION !
    { ## = #[IDENTIFIER_DEF, "IDENTIFIER_DEF"]; }
  ;

/* ---------------------------------------------------------------------- */

block
  :
    ( silent_option )?
    alternative ( OR! alternative )*
    ( additional_option )*
    { ## = #( #[BLOCK_DEF, "BLOCK_DEF"], ##); }
  ;

alternative
  :
    ( alternative_item )*
    { ## = #( #[ALTERNATIVE_DEF, "ALTERNATIVE_DEF"], ##); }
  ;

alternative_item
  :
    (
      // options { warnWhenFollowAmbig = false; }: // !? correct additional_option
      set_option |
      inc_option |
      invoke_option |
      execute_option |
      redudant_option |
      location_option |
      style_option |
      ebnf |
      semantic_predicate |
      terminal |
      nonterminal
    )
    { ## = #( #[ALTERNATIVE_ITEM_DEF, "ALTERNATIVE_ITEM_DEF"], ##); }
  ;

ebnf
  :
    LPAREN !
    b:block
    RPAREN !

    (
       m:ebnf_mark !
       {
         if (#m != null)
         {
            char c = #m.getText().charAt(0);
            if (c == '@')
               c = '+';
            ((BlockNode)#b).mark = c;
         }
       }
       { ## = #( #[EBNF_DEF, "EBNF_DEF"], ##); }
    |
       IMPL
       { ((BlockNode)#b).mark = '>'; }
       { ((BlockNode)#b).syntactic_predicate = true; }
       { ## = #( #[SYNTACTIC_PREDICATE_DEF, "SYNTACTIC_PREDICATE_DEF"], ##); }
    )
  ;

ebnf_mark
  :
    ( QUESTION | STAR | PLUS | AT )?
  ;

semantic_predicate
  :
    LCURLY !
    id
    ( parenthesis )?
    RCURLY !
    QUESTION !
    { ## = #( #[SEMANTIC_PREDICATE_DEF, "SEMANTIC_PREDICATE_DEF"], ##); }
  ;

parenthesis
  :
    LPAREN !
    RPAREN !
    { ## = #( #[PARENTHESIS_DEF, "PARENTHESIS_DEF"], ##); }
  ;

terminal
  :
    string_value
    { ## = #( #[TERMINAL_DEF, "TERMINAL_DEF"], ##); }
  ;

nonterminal
  :
    ( field_label   |
      insert_label  |
      add_option    |
      assign_label  |
      call_option   |
      result_option )?
    NAME
    { ## = #( #[NONTERMINAL_DEF, "NONTERMINAL_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

field_label
  :
    id
    COLON !
    { ## = #( #[LABEL_DEF, "LABEL_DEF"], ##); }
  ;

insert_label
  :
    id
    INSERT !
    { ## = #( #[INSERT_DEF, "INSERT_DEF"], ##); }
  ;

add_option
  :
    OPEN_OPTION !
    "add" !
    CLOSE_OPTION !
    { ## = #( #[ADD_DEF, "ADD_DEF"], ##); }
  ;

assign_label
  :
    id
    ASSIGN!
    { ## = #( #[ASSIGN_DEF, "ASSIGN_DEF"], ##); }
  ;

call_option
  :
    OPEN_OPTION !
    "call" !
    CLOSE_OPTION !
    { ## = #( #[CALL_DEF, "CALL_DEF"], ##); }
  ;

result_option
  :
    OPEN_OPTION !
    "result" !
    CLOSE_OPTION !
    { ## = #( #[RESULT_DEF, "RESULT_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

silent_option
  :
    OPEN_OPTION !
    "silent" !
    CLOSE_OPTION !
    { ## = #( #[SILENT_DEF, "SILENT_DEF"], ##); }
  ;

additional_option
  :
    EXT_OPEN_OPTION !
    "additional" !
    id
    EXT_CLOSE_OPTION !
    { ## = #( #[ADDITIONAL_DEF, "ADDITIONAL_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */

set_option
  :
    OPEN_OPTION !
    "set" !
    id
    ASSIGN !
    simple_value
    CLOSE_OPTION !
    { ## = #( #[SET_DEF, "SET_DEF"], ##); }
  ;

inc_option
  :
    OPEN_OPTION !
    "inc" !
    id
    CLOSE_OPTION !
    { ## = #( #[INC_DEF, "INC_DEF"], ##); }
  ;

invoke_option
  :
    OPEN_OPTION !
    "invoke" !
    id
    CLOSE_OPTION !
    { ## = #( #[INVOKE_DEF, "INVOKE_DEF"], ##); }
  ;

execute_option
  :
    OPEN_OPTION !
    "execute" !
    id
    ( parenthesis )?
    CLOSE_OPTION !
    { ## = #( #[EXECUTE_DEF, "EXECUTE_DEF"], ##); }
  ;

redudant_option
  :
    OPEN_OPTION !
    "redudant" !
    CLOSE_OPTION !
    { ## = #( #[REDUDANT_DEF, "REDUDANT_DEF"], ##); }
  ;

location_option
  :
    OPEN_OPTION !
    "location" !
    id
    CLOSE_OPTION !
    { ## = #( #[LOCATION_DEF, "LOCATION_DEF"], ##); }
  ;

style_option
  :
    OPEN_OPTION !
    (
    "space" |
    "no_space" |
    "new_line" |
    "empty_line" |
    "no_empty_line" |
    "indent" |
    "unindent"
    )
    CLOSE_OPTION !
    { ## = #( #[STYLE_DEF, "STYLE_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */
