
options
{
  language="Java";
}

class EGTreeParser extends TreeParser;

options
{
  importVocab=EG2;
  exportVocab=EG3;
  defaultErrorHandler=false;
}

{
   public Tool tool;
   private String tmp_type;
   private int tmp_mode;
   private String tmp_label;
   private int tmp_kind;
   private boolean tmp_parenthesis;
   private boolean tmp_pointer;
}

/* ---------------------------------------------------------------------- */

grammar
  :
    (
      structDef |
      enumDef |
      literalsDef |
      externalTokensDef |
      optionsDef |
      grammarOptionsDef |
      lexerOptionsDef |
      parserOptionsDef |
      extensionDef
    )*
    rules
  ;

/* ---------------------------------------------------------------------- */

structDef
  :
  #( STRUCT_DEF
     param
     { tool.rememberPosition ((SourceNode) #structDef); }
     { tool.openStruct (tmp_type); } /* tmp_type from param */

     ( fieldDef | sequenceDef )*

     { tool.closeStruct (); }
  );

param
  :
  #( PARAM_DEF
     name:NAME
     { tmp_type = #name.getText (); }
     ( superDef )?
  );

superDef
  :
  #( SUPER_DEF
     name:NAME
     { tool.addSuper (tmp_type, #name.getText ()); } /* tmp_type from param */
  );

fieldDef
  :
  #( FIELD_DEF
     type:NAME
     pointerDef
     name:NAME
     { tool.structField (#name.getText (), #type.getText(), tmp_pointer); }
  );

pointerDef
  :
  { tmp_pointer = false; }
  (
     #( POINTER_DEF
        { tmp_pointer = true; }
     )
  )?
  ;

sequenceDef
  :
  #( SEQUENCE_DEF
     type:NAME
     name:NAME
     { tool.structSequence (#name.getText (), #type.getText()); }
  );

/* ---------------------------------------------------------------------- */

enumDef
  :
  #( ENUM_DEF
     name:NAME
     { tool.rememberPosition ((SourceNode) #enumDef); }
     { tool.openEnum (#name.getText()); }

     (
        fld:NAME
        { tool.enumItem (#fld.getText()); }
     )*

     { tool.closeEnum (); }
  );

/* ---------------------------------------------------------------------- */

literalsDef
  :
  #( LITERALS_DEF (literalDef)* )
  ;

literalDef
  :
  #( LITERAL_DEF
     name:NAME
     value:string_value
     { tool.literal (#name.getText (), Common.strip (#value.getText ())); }
  );

/* ---------------------------------------------------------------------- */

externalTokensDef
  :
  #( EXTERNAL_TOKENS_DEF (externalTokenDef)* )
  ;

externalTokenDef
  :
  #( EXTERNAL_TOKEN_DEF
     value:string_value
     { tool.externalToken (Common.strip (#value.getText ())); }
  );

/* ---------------------------------------------------------------------- */

optionsDef
  :
  #( OPTIONS_DEF (optionDef)* )
  ;

optionDef
  :
  #( OPTION_DEF
     name:NAME
     value:simple_value
     { tool.rememberPosition ((SourceNode) #optionDef); }
     { tool.option (#name.getText (), #value.getText ()); }
  );

string_value
  :
  STRING_LITERAL | CHAR_LITERAL
  ;

simple_value
  :
  NAME | NUMBER | CHAR_LITERAL | STRING_LITERAL
  ;

/* ---------------------------------------------------------------------- */

grammarOptionsDef
  :
  #(GRAMMAR_OPTIONS_DEF (grammarOptionDef)* )
  ;

grammarOptionDef
  :
  #( OPTION_DEF
     name:NAME
     value:simple_value
     { tool.grammarOption (#name.getText (), #value.getText ()); }
  );

/* ---------------------------------------------------------------------- */

lexerOptionsDef
  :
  #( LEXER_OPTIONS_DEF (lexerOptionDef)* )
  ;

lexerOptionDef
  :
  #( OPTION_DEF
     name:NAME
     value:simple_value
     { tool.lexerOption (#name.getText (), #value.getText ()); }
  );

/* ---------------------------------------------------------------------- */

parserOptionsDef
  :
  #( PARSER_OPTIONS_DEF (parserOptionDef)* )
  ;

parserOptionDef
  :
  #( OPTION_DEF
     name:NAME
     value:simple_value
     { tool.parserOption (#name.getText (), #value.getText ()); }
  );

/* ---------------------------------------------------------------------- */

extensionDef
  :
  #( EXTENSION_DEF
     name:NAME
     code:CODE
     { tool.rememberPosition ((SourceNode) #extensionDef); }
     { tool.extension (#name.getText (), #code.getText ()); }
  );

/* ---------------------------------------------------------------------- */

rules
  :
  #( RULES_DEF (public_option | private_option | rule)* )
  ;

public_option
  :
  #( PUBLIC_DEF
     { tool.publicOption (); }
  );

private_option
  :
  #( PRIVATE_DEF
     { tool.privateOption (); }
  );

rule
  :
  #( RULE_DEF

     name:NAME
     { tool.startRule ((RuleNode) #rule, #name.getText()); }

     (
        binary_block
     |
        unary_block
     |
        primitive_block
     |
        (
           { tmp_mode = Common.NEW_MODE; }
           (
             "new" |
             "modify" { tmp_mode = Common.MODIFY_MODE; } |
             "return" { tmp_mode = Common.RETURN_MODE; } |
             "select" { tmp_mode = Common.SELECT_MODE; } |
             "choose" { tmp_mode = Common.CHOOSE_MODE; }
           )?

           param
           { tool.onRuleParam (tmp_mode, tmp_type); }
           /* tmp_type from param */
        |
           { tool.onRuleParam (Common.NONAME_MODE, ""); }
        )

        ( custom )*
        ( store )?

        block
     )

     { tool.stopRule (); }
  );

custom
  :
  #( CUSTOM_DEF
     (
        name:NAME
        { tool.addCustom (#name.getText ()); }
     )*
  );

store
  :
  #( STORE_DEF
     name:NAME
     type_name:NAME
     { tool.addStore (#name.getText (), #type_name.getText ()); }
  );

/* ---------------------------------------------------------------------- */

binary_block
  :
  #( BINARY_DEF
     { tool.openBinary (); }
     ( custom )*
     branch:NAME
     { tool.binaryBranch (#branch.getText()); }
     ( binary_alternative )*
     { tool.closeBinary (); }
  );

binary_alternative
  :
  #( BINARY_ITEM_DEF
     key:string_value
     selector_value:NAME
     (
         middle_branch:NAME
         middle_key:string_value
         { tool.onBinaryItem ((BinaryNode) #binary_alternative,
                              Common.strip (#key.getText()),
                              #selector_value.getText(),
                              #middle_branch.getText(),
                              Common.strip (#middle_key.getText())); }
     |
         { tool.onBinaryItem ((BinaryNode) #binary_alternative,
                              Common.strip (#key.getText()),
                              #selector_value.getText(),
                              "",
                              ""); }
     )
  );

/* ---------------------------------------------------------------------- */

unary_block
  :
  #( UNARY_DEF
     ( custom )*
     key:string_value
     selector_value:NAME
     branch:NAME
     { tool.unary (Common.strip (#key.getText()),
                   #selector_value.getText(),
                   #branch.getText()); }
  );

/* ---------------------------------------------------------------------- */

primitive_block
  :
  #( PRIMITIVE_DEF
     { tool.openPrimitive (); }
     ( identifier_option )?
     ( custom )*
     ( primitive_alternative )*
     { tool.closePrimitive (); }
  );

primitive_alternative
  :
  #( PRIMITIVE_ITEM_DEF
    name:NAME
    { tool.onPrimitiveItem ((PrimitiveNode) #primitive_alternative,
                            #name.getText()); }
  );

identifier_option
  :
  #( IDENTIFIER_DEF
     { tool.identifierOption (); }
  );

/* ---------------------------------------------------------------------- */

block
  :
  #( BLOCK_DEF
     { tool.startBlock ((BlockNode) #block); }
     ( silent_option )?
     ( alternative )*
     ( additional_option )*
     { tool.stopBlock (); }
  );

alternative
  :
  #( ALTERNATIVE_DEF
     { tool.startAlternative ((AltNode) #alternative); }
     ( alternative_item )*
     { tool.stopAlternative (); }
  );

alternative_item
  :
  #( ALTERNATIVE_ITEM_DEF
     { tool.startAltItem (#alternative_item); }
     (
        set_option |
        inc_option |
        invoke_option |
        execute_option |
        redudant_option |
        location_option |
        style_option |
        ebnf |
        semantic_predicate |
        syntactic_predicate |
        terminal |
        nonterminal
     )
     { tool.stopAltItem (); }
  );

ebnf
  :
  #( EBNF_DEF
     block
  );

syntactic_predicate
  :
  #( SYNTACTIC_PREDICATE_DEF
     { tool.startSyntacticPredicate (); }
     block
     { tool.stopSyntacticPredicate (); }
  );

semantic_predicate
  :
  #( SEMANTIC_PREDICATE_DEF
     name:NAME
     parenthesis
     { tool.onSemanticPredicate ((SemanticNode) #semantic_predicate,
                                 #name.getText (),
                                 tmp_parenthesis); }
  );

parenthesis
  :
  { tmp_parenthesis = false; }
  (
     #( PARENTHESIS_DEF
        { tmp_parenthesis = true; }
     )
  )?
  ;

terminal
  :
  #( TERMINAL_DEF
     val:string_value
     { tool.onTerminal ((TerminalNode) #terminal,
                        Common.strip (#val.getText ())); }
   );

/* ---------------------------------------------------------------------- */

nonterminal
  :
  #( NONTERMINAL_DEF
     { tmp_label = ""; }

     ( element_label { tmp_kind = Common.LABELED; }   |
       insert_label  { tmp_kind = Common.INSERT; }    |
       ADD_DEF       { tmp_kind = Common.ADD; }       |
       assign_label  { tmp_kind = Common.INSERT; }    |
       CALL_DEF      { tmp_kind = Common.CALL; }      |
       RESULT_DEF    { tmp_kind = Common.RESULT; }    |
                     { tmp_kind = Common.ANONYMOUS; } )

     name:NAME

     { tool.onNonterminal ((NontermNode) #nonterminal,
                           tmp_kind, tmp_label, #name.getText ()); }
  );

element_label
  :
  #( LABEL_DEF
     name:NAME
     { tmp_label = #name.getText (); }
  );

insert_label
  :
  #( INSERT_DEF
     name:NAME
     { tmp_label = #name.getText (); }
  );

assign_label
  :
  #( ASSIGN_DEF
     name:NAME
     { tmp_label = #name.getText (); }
  );

/* ---------------------------------------------------------------------- */

silent_option
  :
  #( SILENT_DEF
     { tool.onSilentOption (); }
  );

additional_option
  :
  #( ADDITIONAL_DEF
     name:NAME
     { tool.additionalOption (#name.getText ()); }
  );

/* ---------------------------------------------------------------------- */

set_option
  :
  #( SET_DEF
     name:NAME
     value:simple_value
     { tool.onSetOption ((SetOptionNode) #set_option,
                         #name.getText (),
                         #value.getText ()); }
  );

inc_option
  :
  #( INC_DEF
     name:NAME
     { tool.onIncOption ((IncOptionNode) #inc_option,
                         #name.getText ()); }
  );

invoke_option
  :
  #( INVOKE_DEF
     name:NAME
     { tool.onInvokeOption ((InvokeOptionNode) #invoke_option,
                            #name.getText ()); }
  );

execute_option
  :
  #( EXECUTE_DEF
     name:NAME
     parenthesis
     { tool.onExecuteOption ((ExecuteOptionNode) #execute_option,
                             #name.getText (),
                             tmp_parenthesis); }
  );

redudant_option
  :
  #( REDUDANT_DEF
     { tool.onRedudantOption (); }
  );

location_option
  :
  #( LOCATION_DEF
     name:NAME
     { tool.locationOption (#name.getText ()); }
  );

style_option
  :
  #( STYLE_DEF
     { int opt; }
     (
     "space"               { opt = Common.SPACE; } |
     "no_space"            { opt = Common.NO_SPACE; } |
     "new_line"            { opt = Common.NEW_LINE; } |
     "empty_line"          { opt = Common.EMPTY_LINE; } |
     "no_empty_line"       { opt = Common.NO_EMPTY_LINE; } |
     "indent"              { opt = Common.INDENT; } |
     "unindent"            { opt = Common.UNINDENT; }
     )
     { tool.styleOption (opt); }
  );

/* ---------------------------------------------------------------------- */
