
/* collection.cc */

#include "collection.h"

#ifdef USE_OWN_DICT
   #include "own-collection.cc"

#elif defined (USE_COMPLETE_DICT)
   #include "complete-collection.cc"

#else
   #include SET_INCL
   #ifdef USE_STR_HASH
      #include "strhash.h"
   #endif
#endif

OPEN_NAMESPACE

#if ! defined (USE_OWN_DICT) && ! defined (USE_COMPLETE_DICT)

/******************************** STD HASH ********************************/

/* DictLong */

typedef SET_TYPE <long> LongSet; // use hash set if available

bool DictLong::contains (Key inx)
{
   return (*(LongSet*)m).find (inx) != (*(LongSet*)m).end ();
}

void DictLong::add (Key inx)
{
   (*(LongSet*)m).insert (inx);
}

void DictLong::remove (Key inx)
{
   (*(LongSet*)m).erase (inx);
}

void DictLong::clear ()
{
   (*(LongSet*)m).clear ();
}

DictLong::DictLong ()
{
   m = (void *) new LongSet;
}

DictLong::~DictLong ()
{
   delete (LongSet*) m;
}

/* DictString */

#ifdef USE_STR_HASH
   typedef SET_TYPE <string, StrHash> StrSet;
#else
   typedef SET_TYPE <string> StrSet;
#endif

bool DictString::contains (Key inx)
{
   return (*(StrSet*)m).find (inx) != (*(StrSet*)m).end ();
}

void DictString::add (Key inx)
{
   (*(StrSet*)m).insert (inx);
}

void DictString::remove (Key inx)
{
   (*(StrSet*)m).erase (inx);
}

void DictString::clear ()
{
   (*(StrSet*)m).clear ();
}

DictString::DictString ()
{
   m = (void *) new StrSet;
}

DictString::~DictString ()
{
   delete (StrSet*) m;
}

#endif

/* ---------------------------------------------------------------------- */

#define TEST_COLECTION

#ifdef TEST_COLECTION

class C2
{
   public:
      int x, y, z;

      C2 () :
         x (0),
         y (0),
         z (0)
      { }

      C2 (const C2 & p) :
         x (p.x),
         y (p.y),
         z (p.z)
      { }
};

template <typename Key>
void test (Key a)
{
   Collection <Key> dict;

   dict.add (a);
   if (dict.contains (a))
      dict.clear ();
}

enum color { red, blue, green, yellow };

void test_collections ()
{
   test <bool> (true);
   test <int> (1);
   test <color> (blue);
   test <string> ("abc");

   C2 c;
   c.x = 1;
   c.y = 2;
   c.z = 3;
   // NO test <C2> (c);
   test <C2 *> (&c);
}

#endif

CLOSE_NAMESPACE
