
/* collection.h */

#ifndef __PROG_COLLECTION_H__
#define __PROG_COLLECTION_H__

#include "std.h"

#ifdef USE_OWN_DICT
   #include "own-collection.h"

#elif defined (USE_COMPLETE_DICT)
   #include "complete-collection.h"

#else

OPEN_NAMESPACE

/* ---------------------------------------------------------------------- */

/* Support */

class DictLong
{
   private:
      void * m;

   public:
      typedef long Key;

      bool contains (Key inx);
      void add (Key inx);
      void remove (Key inx);
      void clear ();

      DictLong ();
      ~ DictLong ();
};

class DictString
{
   private:
      void * m;

   public:
      typedef string Key;

      bool contains (Key inx);
      void add (Key inx);
      void remove (Key inx);
      void clear ();

      DictString ();
      ~ DictString ();
};

/* ---------------------------------------------------------------------- */

/* Collection */

template <typename Key>
class Collection
{
   private:
      DictLong dict;

   public:
      bool contains (Key inx) { return dict.contains ((size_t) inx); }
      void add (Key inx)      { dict.add ((size_t) inx); }
      void remove (Key inx)   { dict.remove ((size_t) inx); }
      void clear ()           { dict.clear (); }
};

template < >
class Collection <string>
{
   private:
      typedef string Key;
      DictString dict;

   public:
      bool contains (Key inx) { return dict.contains (inx); }
      void add (Key inx)      { dict.add (inx); }
      void remove (Key inx)   { dict.remove (inx); }
      void clear ()           { dict.clear (); }
};

typedef Collection <string> StringCollection;

/* ---------------------------------------------------------------------- */

CLOSE_NAMESPACE

#endif // USE_OWN_DICT, USE_COMPLETE_DICT

#endif /* __PROG_COLLECTION_H__ */
