
/* dictionary.h */

#ifndef __PROG_DICTIONARY_H__
#define __PROG_DICTIONARY_H__

#include "std.h"

#ifdef USE_OWN_DICT
   #include "own-dictionary.h"

#elif defined (USE_COMPLETE_DICT)
   #include "complete-dictionary.h"

#else

OPEN_NAMESPACE

/* ---------------------------------------------------------------------- */

/* Support */

class DictLongToLong
{
   private:
      void * m;

   public:
      typedef long Key;
      typedef long Data;

      bool contains (Key inx);
      Data get (Key inx);
      void store (Key inx, Data value);
      void clear ();

      DictLongToLong ();
      ~ DictLongToLong ();
};

class DictLongToString
{
   private:
      void * m;

   public:
      typedef long Key;
      typedef string Data;

      bool contains (Key inx);
      Data get (Key inx);
      void store (Key inx, Data value);
      void clear ();

      DictLongToString ();
      ~ DictLongToString ();
};

class DictStringToLong
{
   private:
      void * m;

   public:
      typedef string Key;
      typedef long Data;

      bool contains (Key inx);
      Data get (Key inx);
      void store (Key inx, Data value);
      void clear ();

      DictStringToLong ();
      ~ DictStringToLong ();
};

class DictStringToString
{
   private:
      void * m;

   public:
      typedef string Key;
      typedef string Data;

      bool contains (Key inx);
      Data get (Key inx);
      void store (Key inx, Data value);
      void clear ();

      DictStringToString ();
      ~ DictStringToString ();
};

/* ---------------------------------------------------------------------- */

/* Dictionary */

template <typename T1, typename T2>
class Dictionary
{
   public:
      typedef T1 Key;
      typedef T2 Data;

   private:
      DictLongToLong dict;

   public:
      bool contains (Key inx)
         { return dict.contains ((size_t) inx); }

      Data get (Key inx)
         { return (Data) (dict.get ((size_t) inx)); }

      void store (Key inx, Data value)
         { dict.store ((size_t) inx, (size_t) value); }

      void clear ()
         { dict.clear (); }
};

template <typename T>
class Dictionary <string, T>
{
   public:
      typedef string Key;
      typedef T Data;

   private:
      DictStringToLong dict;

   public:
      bool contains (Key inx)
         { return dict.contains (inx); }

      Data get (Key inx)
         { return (Data) (dict.get (inx)); }

      void store (Key inx, Data value)
         { dict.store (inx, (size_t) value); }

      void clear ()
         { dict.clear (); }
};

template <typename T>
class Dictionary <T, string>
{
   public:
      typedef T Key;
      typedef string Data;

   private:
      DictLongToString dict;

   public:
      bool contains (Key inx)
         { return dict.contains ((size_t) inx); }

      Data get (Key inx)
         { return dict.get ((size_t) inx); }

      void store (Key inx, Data value)
         { dict.store ((size_t) inx, value); }

      void clear ()
         { dict.clear (); }
};

template < >
class Dictionary <string, string>
{
   public:
      typedef string Key;
      typedef string Data;

   private:
      DictStringToString  dict;

   public:
      bool contains (Key inx)          { return dict.contains (inx); }
      Data get (Key inx)               { return dict.get (inx); }
      void store (Key inx, Data value) { dict.store (inx, value); }
      void clear ()                    { dict.clear (); }
};

typedef Dictionary <string, string> StringDictionary;

/* ---------------------------------------------------------------------- */

CLOSE_NAMESPACE

#endif // USE_OWN_DICT, USE_COMPLETE_DICT

#endif /* __PROG_DICTIONARY_H__ */
